%% `alignment of two automata'
%% as requested by David.Carter@SpeechMachines.com
%%
%% approximate intersection, where every deletion, insertion, 
%% and substitution is assigned a penalty score of 1.
%% The result of approx_intersect/2 is a weighted recognizer
%% where the symbols are actually pairs of symbols, and/or [].
%%
%% the 'report_approx_intersect' is similar, except it reports
%% the best match to error output. Note that you only get a
%% single result, which is an arbitrary member of the set of
%% best results. It uses a special character, '-', to indicate
%% deletion/insertion, but this can be changed easily (fourth
%% argument of two_paths/2 below).
%%
%% report_approx_intersect([a,b,c..z,d,e,g,h],[a,c,c,d,e,f,g,h])
%% produces:
%%
%% weight: 2
%% abcde-gh
%% accdefgh
%%
%% whereas
%% report_approx_intersect([a..c,{a,b},c..z,d,?,g,h],[a..c,{b,c},c,d,e,?,g,h])
%%
%% weight: 1
%% $@(in([a,b,c]))bcd-$@(not_in([]))gh
%% $@(in([a,b,c]))bcde$@(not_in([]))gh
%%
%% if you don't like this predicates stuff, consider
%% report_approx_intersect_letters_only in case you only work
%% with letters, and don't want to see these funny predicate
%% names ever:
%% 
%% report_approx_intersect_letters_only([a..c,{a,b},c..z,d,?,g,h],
%%                                      [a..c,{b,c},c,d,e,?,g,h])
%% 
%% weight: 1
%% abcd-0gh
%% abcde0gh


:- multifile rx/2.
:- multifile macro/2.

rx(report_approx_intersect(A,B),Fa) :-
    rx(approx_intersect(A,B),Fa),
    fsa_minimum_path:fsa_minimum_path(Fa,Score/Path),
    two_paths(Path,PathA,PathB,'-'),
    fsa_interpreter:list_of_symbols_to_string(PathA,StrA,in),
    fsa_interpreter:list_of_symbols_to_string(PathB,StrB,in),
    format(user_error,"weight: ~w~n~s~n~s~n~n",[Score,StrA,StrB]).

macro(sig,{a..z,'A'..'Z','0'..'9'}).

rx(report_approx_intersect_letters_only(A,B),Fa) :-
    rx(approx_intersect(expand_predicates(sig* & A),
			expand_predicates(sig* & B)),Fa),
    fsa_minimum_path:fsa_minimum_path(Fa,Score/Path),
    two_paths(Path,PathA,PathB,'-'),
    fsa_interpreter:list_of_symbols_to_string(PathA,StrA,in),
    fsa_interpreter:list_of_symbols_to_string(PathB,StrB,in),
    format(user_error,"weight: ~w~n~s~n~s~n~n",[Score,StrA,StrB]).

two_paths([],[],[],_).
two_paths([S|Ss],[A|As],[B|Bs],Zero) :-
    (	S = A0/B0
    ->	(   A0 == []
	->  A = Zero
	;   A0 = A
	),
	(   B0 == []
	->  B = Zero
	;   B0 = B
	)
    ;	S=A, S=B
    ),
    two_paths(Ss,As,Bs,Zero).

rx(approx_intersect(A,B),Fa) :-
    fsa_regex:rx(A,FaA),
    fsa_regex:rx(B,FaB),
    %% for now, simply assume FaA and FaB are recognizers
    fsa_regex:all_components_reachable(
        fsa_regex_aux:approx_intersect,FaA,FaB,Fa).

prepare_reachable(approx_intersect,FaA,FaB,
	     d(FinalsA,FinalsB,TransA,TransB,JumpsA,JumpsB,M),
		  StartStates,r(M,fsa_frozen)):-
    fsa_data:fsa_symbols_decl(FaA,r(M)),
    fsa_data:fsa_symbols_decl(FaB,r(M)),
    fsa_regex:pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    fsa_data:index_jumps(FaA,JumpsA),
    fsa_data:index_jumps(FaB,JumpsB),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB).

trans_reachables(approx_intersect,Pa/Pb,d(_,_,TrA,TrB,JA,JB,Mod),Trans):-
    findall(Tr,trans_approx(Pa,Pb,TrA,TrB,JA,JB,Mod,Tr),Trans).

trans_approx(Pa,Pb,_,_,JA,_,_,jump(Pa/Pb,Qa/Pb)) :-
    fsa_array:fsa_array_access(Pa,List,[],JA),
    lists:member(Qa,List).
trans_approx(Pa,Pb,_,_,_,JB,_,jump(Pa/Pb,Pa/Qb)) :-
    fsa_array:fsa_array_access(Pb,List,[],JB),
    lists:member(Qb,List).
trans_approx(Pa,Pb,TrA0,TrB0,_,_,M,trans(Pa/Pb,Sym,Qa/Qb)) :-
    fsa_array:fsa_array_access(Pa,TrA,[],TrA0), 
    fsa_array:fsa_array_access(Pb,TrB,[],TrB0),
    lists:member(SymA-Qas,TrA),
    lists:member(SymB-Qbs,TrB),
    trans_intersect_sym(M,SymA,SymB,Sym),
    lists:member(Qa,Qas),
    lists:member(Qb,Qbs).
trans_approx(Pa,Pb,TrA0,_,_,_,_,trans(Pa/Pb,(Sym/[])/1,Qa/Pb)):-
    fsa_array:fsa_array_access(Pa,TrA,[],TrA0), 
    lists:member(Sym-Qas,TrA),
    lists:member(Qa,Qas).
trans_approx(Pa,Pb,_,TrB0,_,_,_,trans(Pa/Pb,([]/Sym)/1,Pa/Qb)):-
    fsa_array:fsa_array_access(Pb,TrB,[],TrB0), 
    lists:member(Sym-Qbs,TrB),
    lists:member(Qb,Qbs).

trans_intersect_sym(Mod,SymA,SymB,Sym/0) :-
    fsa_pred_module:conjunction(Mod,SymA,SymB,Sym0),
    fsa_pred_module:identity(Mod,Sym0,Sym).
trans_intersect_sym(_Mod,SymA,SymB,(SymA/SymB)/1).


is_final_reachable(approx_intersect,d(FinalsA,FinalsB,_,_,_,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    fsa_regex:and(Bool1,Bool2,Bool).


