%% Dale wrote:
/*
I made an improved version of automata.pl for the Examples
directory. Not that anybody was clamoring for an improved version, but
it's something I just happened to think of while doing something
entirely different. I was actually working on the morphotactic part of
Englex to make a Kimmo example for my students. I found myself
thinking in double negatives: a sequence of morphemes should not
contain a morpheme followed by something not in it's continuation
class. This naturally led to a lot of complexity. The alternative
seems obvious in retrospect, but it took me a little while to think of
it.
*/

%% (c) Dale Gerdemann 1999.
%% slightly adapted by Gertjan van Noord.
%%
%% Dale: I found a better way to construct an extended regular expression from
%% an automaton. The H&U algorithm results in a mess because of the
%% limited range of operators they allow themselves. If you use the full
%% power of the finite state calculus, then you can handle the H&U
%% (p. 35) example by:

%% Figure 2.16 Hopcroft and Ullman, page 35
macro(f216, fa({0,1},              % sigma
	       {q1,q2,q3},         % states
	       {q1},               % initials
	       {q2,q3},            % finals
	       {[q1,0,q2],         % transitions
		[q1,1,q3],
		[q2,0,q1],
		[q2,1,q3],
		[q3,{0,1},q2]
	       })
     ).

%% Similar example, but a transducer
macro(f216td,td({0,1},              % sig1
		{0,1},		    % sig2
		{q1,q2,q3},	    % states
		{q1},		    % initials
		{q2,q3},	    % finals
		{[q1,0,'|',0,q2],   % transitions
		 [q1,1,'|',[0,1],q3],
		 [q2,0,'|',0,q1],
		 [q2,1,'|',1,q3],
		 [q3,{0,1},'|',1,q2]
		})
     ).


%% Figure 2.33a Hopcroft and Ullman, page 51
macro(f233a, fa({0,1},{a,b,c},{a},{a},
	       {[a,0,a],[a,1,b],[b,1,b],[b,0,c],[c,1,b],[c,0,a]})).


%% recognizer:
macro(fa(Alphabet,States,Initials,Finals,Transitions),
      m(range( (                   %% triples Source Symbol State
		       [[States,Alphabet]*,States]     
	                       &   %% no non-transition triples
               ~ $ ([States,Alphabet,States]-Transitions)
	                       &   %% start in initial state
	                  [Initials,? *]
	                       &   %% end in final state
	                  [? *,Finals]
 	       )
	                  o        %% get rid of state names
	    [[States x [],?]*,States x []]))
     ).

%% transducer:
%%
%% Dale:
%% You see I've used '|' as a marker symbol to divide the input part of a
%% transition from the output. This, of course brings in the usual problems
%% with markers. We could treat this as we do in the paper, but I didn't
%% feel like going into any complications.
%% Gertjan:
%% Note that intersection of states and symbols should be empty;
%% i.e. write states and symbols differently.
%% Also note that | should not be a symbol or a state.
macro(td(Sig1,Sig2,States,Initials,Finals,Transitions),
          [ [[] x States,Sig1*,[] x ['|',Sig2*]]*, [] x States]
                              o
               (
                       [[States,Sig1*,'|',Sig2*]*,States]
                               &   %% no non-transition quadruples
               ~ $ ([States,Sig1*,'|',Sig2*,States]-Transitions)
                               &   %% start in initial state
                          [Initials,? *]
                               &   %% end in final state
                          [? *,Finals]
               )
                               o
            [ [States x [],[Sig1*,'|'] x [],Sig2*]*,States x []]).





%% Improved Version (added 2002 by Dale Gerdemann)

%% The macro fa/5 shown above constructs sequences of transitions
%% using a kind of double negation: does not contain a
%% non-transition. This adds unnecessary complexity to the
%% problem. The improved version fa/3 works by intersecting two fsa's, 
%% one of which ensures that every even transition is legal and one
%% ensures that every odd transition is okay. The reason why this
%% cannot be expressed in a simple fsa without intersection is that
%% transitions overlap. The destination state for a transition is the
%% source state for the next transition.
%%
%% I've only added recognizer construction. Transducers can easily be
%% handled in the same way.



%% recognizer:
macro(fa(Initials,Finals,Trans),
   pragma([States-states(Trans),
           Sigma-sigma(Trans)
	  ],
      m(range(
         ( [States,[Sigma,Trans]*,[Sigma,States]^]
                          &
           [[Trans,Sigma]*,{States,Trans}]
                          &   %% start in initial state
                    [Initials,? *]
                           &   %% end in final state
                       [? *,Finals] )
 
                           o        %% get rid of state names

	       [[States x [],?]*,States x []]
	     )))).


% Get states from transtions
macro(states(Transitions),
   range( Transitions o { [?,? : [], ? : []], [? : [], ? : [], ?] })).

% Get alphabet from transitions
macro(sigma(Transitions),
   range( Transitions o [? : [], ?, ? : []])).



%% A couple of examples from the old edition of Hopcroft and Ullman
%% Do these same examples occur in the new edition?

%% Figure 2.16 Hopcroft and Ullman, page 35

macro(f216_v2, fa({q1},               % initials
                  {q2,q3},            % finals
                  {[q1,0,q2],         % transitions
                   [q1,1,q3],
                   [q2,0,q1],
                   [q2,1,q3],
                   [q3,{0,1},q2]
                  })
        ).

%% Figure 2.33a Hopcroft and Ullman, page 51
macro(f233a_v2, fa({a},{a},
	       {[a,0,a],[a,1,b],[b,1,b],[b,0,c],[c,1,b],[c,0,a]})).

