%%           -*-Mode: prolog;-*-

%% (c) Dale Gerdemann 2001
%% Merged boolean predicates. This file combines the code in
%% Examples/Booleans/booleans.pl and 
%% Examples/RocheSchabes97/gvn.pl. The code from the latter file was
%% implemented by Gertjan van Noord and has been copied here with
%% very minor alterations (and with the bimachine code deleted). The
%% original header of gvn.pl is as follows: 
%%
%% (c) Gertjan van Noord 1999
%% inspired by dg.pl by Dale Gerdemann
%%      and by carola_fricke.pl by Carola Fricke
%% based mostly on the introduction of the 1997 Roche&Schabes book:
%% Roche and Schabes, Introduction. In: Roche and Schabes (eds),
%% Finite State Language Processing. MIT Press 1997.
%%
%% NB: does treat symbols as predicates experimentally, by using the
%% new expand_non_overlapping_predicates construct; except for the
%% bimachine stuff.

:- fsa_globals:fsa_host_prolog(Prolog),
      (   Prolog == swi
      ->  style_check([-(atom)])
      ;	  true
      ).

:- use_module(library(lists),   [ member/2 ]).
:- multifile macro/2.
:- multifile rx/2.
:- discontiguous macro/2.
:- discontiguous rx/2.

:- op(400, yfx, and).
:- op(405, yfx, or).
:- op(390, fx, not).

% Boolean values. These are based on the "Semiring of Formal Languages
% over Sigma", (Werner Kuich, Semirings and Formal Power Series, in
% Handbook of Formal Languages, vol. 1).

macro(bool,{true,false}).  % 'bool' is actually the same lg as 'true'
macro(true,[]).
macro(false,{}).

% Coercion to boolean. Any language that accepts epsilon can be
% interpreted as true. For efficiciency, however, we may want to have
% coercion into a cannonical representation for true. The use of this
% coercion operator should never be logically necessary.
% NOTE: A previous version of this library used the total language
% (? *) to represent 'true'. This approach is more difficult to
% generalize, however, since there are many languages that are neither 
% total, nor empty. Thus, in this older approach, coercion played a
% more important role.
macro(bool(A),(A & bool)).

% boolean operations
macro(A or B, {A,B}).
macro(A and B, [A,B]).
macro(not A, bool(~ A)).   % coercion just for efficiency

% conditional
macro(if(Expr,Then,Else),
   { range(Expr o (true x (? *)) o Then),
     range((not Expr) o (true x (? *)) o Else) }).

% Decidability properties for recognizers
macro(member(String,A),not empty(String & A)).
macro(empty(A), not range(A o ((? *) x true))).
macro(total(A), empty(~ A)).
macro(subset(A1,A2), empty(A1 & ~ A2)).
macro(equal(A1,A2), subset(A1,A2) and subset(A2,A1)).
macro(overlap(A1,A2), not empty(A1 & A2)).

% A language is infinite if it uses an infinite alphabet or if it is
% non-empty and has and empty set of longest strings. XFST has related 
% predicates called: "test lower/upper-bounded". 
macro(infinite(A),
   infinite_alphabet(A)
        or
   not empty(A)
        and       
   empty(longest(A))).
   
% Fa0 uses an infinite alphabet if minimize(Fa0) compiles into
% an automaton with an infinite predicate on a transition.
% Minimization is necessary to ensure that the transtion is
% (co-)reachable. 
rx(infinite_alphabet(Fa0),Fa) :-
    fsa_regex:rx(minimize(Fa0),Fa1),
    fsa_data:fsa_transitions(Fa1,Transs),
    (   member(trans(_,Pred,_),Transs),
	infinite_pred(Pred)
    ->  fsa_regex:rx('[]',Fa)
    ;   fsa_regex:rx('{}',Fa)
    ).

% Should really be definined in predicates module
infinite_pred(not_in(_)).

% Get the longest strings in A. These are the strings for which it is
% not the case that you can extend them to match the length of some
% longer string in A.
macro(longest(A), 
    A & ~ domain([? *, ([] x ?)+] o range(A o (? x ?)*))).

% The language A consists of a single string. If A is a unitary
% language, it may also be referred to as a string.
macro(unitary_lg(A),
    not empty(A)
       and
   is_functional([] x A)).
 

%% Decidability properties for transducers
macro(t_member(In,Out,T), member(Out, range(In o T))).

macro(t_empty(T),empty(T)).  % a transducer is empty if the underlying
                             %  automaton is empty

macro(t_equal(T1,T2),             % Only works for functional transducers
    equal(domain(T1),domain(T2))  % Assumes appropriate definition of
              and                 % is_functional
     is_functional({T1,T2})).  


macro(t_subset(T1,T2),             % Only works for functional transducers
    subset(domain(T1),domain(T2))  % Assumes appropriate definition of
              and                  % is_functional
     is_functional({T1,T2})).  
     
macro(t_identity(T),               % T is an identity transducer
    t_equal(identity(domain(T)), T)).


/*
macro(same_length(T),              % T denotes a same-length relation
   ...                             % a useful exercise for someone to
*/                                 % define this.


% is_equal(T1,T2)
% equality testing described by Roche/Schabes p. 40/41
% Precondition: T1 and T2 are functional transducers
% Result: true (? *) is T1 and T2 are equal, otherwise false ({}),
%    see booleans.pl for these definitions of true and false
macro(is_equal(T1,T2),
    equal(domain(T1),domain(T2))
               &
    is_functional({T1,T2})).
     


rx(is_functional(Expr),Fa):-
    fsa_regex:rx(expand_non_overlapping_predicates(Expr),Fa0),
    is_functional(Fa0,Result),
    (	Result == yes
    ->	fsa_regex:rx('[]',Fa)
    ;   fsa_regex:rx('{}',Fa)
    ).

rx(is_subsequential(Expr),Fa):-
    fsa_regex:rx(expand_non_overlapping_predicates(Expr),Fa0),
    is_subsequential(Fa0,Bool),
    (	Bool == yes
    ->	fsa_regex:rx('[]',Fa)
    ;   fsa_regex:rx('{}',Fa)
    ).

rx(efree_domain(Expr),Fa):-
    fsa_regex:rx(Expr,Fa0),   
    efree_domain(Fa0,Fa1,Cycle),
    (	Cycle == cycle
    ->	fsa_regex:rx('{}',Fa)
    ;   Fa1=Fa
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% EFREE_DOMAIN %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% this is ok for arbitrary pred modules.

%% creates equivalent automaton Fa for Fa0, except that
%% Fa is a sequence transducer without any epsilon on input
%% side. If no equivalent transducer exists, then Fa will
%% be garbage, and Cycle=cycle. Otherwise Cycle=nocycle.
%%
%% this is slow. Not a very good algorithm!!!
efree_domain(Fa0,Fa,Cycle):-
    fsa_regex:rx(sequence_transducer(m(fa(Fa0))),Fa1),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    collect_domain_jumps(Trans0,Trans1,[],Jumps0,[],IJumps0,[]),
    (	Jumps0==[]
    ->	Fa1=Fa, Cycle=nocycle
    ;	sort(Jumps0,Jumps),
	sort(IJumps0,IJumps),
	fsa_data:fsa_states_number(Fa0,Size),
	fsa_array:fsa_array_new(JT,Size),
	fsa_array:fsa_array_new(IT,Size),
	fsa_data:transitions_to_thing(Jumps,JT),
	fsa_data:transitions_to_thing(IJumps,IT),
	length(Jumps,Len),
	catch( (   fsa_regex_aux:extend_trans(Trans1,Trans2,IT,JT,Len),
		   Cycle=nocycle ),
	       cycle,
	       (   Cycle=cycle,
		   Trans2=[] )
	     ),
	sort(Trans2,Trans)
    ).

collect_domain_jumps([],T,T,J,J,I,I).
collect_domain_jumps([H|X],T0,T,J0,J,I0,I):-
    collect_domain_jump(H,T0,T1,J0,J1,I0,I1),
    collect_domain_jumps(X,T1,T,J1,J,I1,I).

collect_domain_jump(trans(P,Left/Right,Q),T0,T,J0,J,I0,I):-
    (	Left==[]
    ->	J0=[trans(P,Right,Q)|J],
	I0=[trans(Q,Right,P)|I],
	T0=T
    ;   T0=[trans(P,Left/Right,Q)|T],
	J0=J,
	I0=I
    ).

extend_trans([],[],_IT,_JT,_).
extend_trans([trans(P,L/R,Q)|Tr],Trans,IT,JT,Len):-
    extend_trans0(P,L,R,Q,Trans,Trans0,IT,JT,Len),
    extend_trans(Tr,Trans0,IT,JT,Len).

extend_trans0(P0,L,R0,Q0,Trans0,Trans,IT,JT,Len):-
    findall(trans(P,L/R,Q),
	    fsa_regex_aux:extend_trans1(P0,R0,Q0,P,R,Q,IT,JT,Len),
	    Trans0,Trans).

extend_trans1(P0,R0,Q0,P,R,Q,IT,JT,Len):-
    lists:append(R0,Right0,Right),
    lpath(P0,P,IT,Right,R,0,Len),
    path(Q0,Q,JT,Right0,[],0,Len).

%% path(State0,State,Table,Seq0,Seq,Len0,Len)
%% find a path from State0 to State using transitions in Table
%% collecting the symbols in Seq0-Seq. Also checks that no
%% cycles are possible by making sure number of transitions is
%% smaller than Len.
%% lpath is different from path only because outputs are appended in
%% different order (left-right rather than right-left).
path(P,P,_,R,R,_,_).
path(P0,P,Table,R0,R,Len0,Len):-
    Len1 is +(Len0,1),
    fsa_array:fsa_array_get(P0,List,Table),
    lists:member(Sym-List1,List),
    lists:append(Sym,R1,R0),
    lists:member(P1,List1),
    (	Len1 < +(Len,1)
    ->	true
    ;   throw(cycle)
    ),
    path(P1,P,Table,R1,R,Len1,Len).

lpath(P,P,_,R,R,_,_).
lpath(P0,P,Table,R0,R,Len0,Len):-
    Len1 is +(Len0,1),
    fsa_array:fsa_array_get(P0,List,Table),
    lists:member(Sym-List1,List),
    lists:append(Sym,R0,R1),
    lists:member(P1,List1),
    (	Len1 < +(Len,1)
    ->	true
    ;   throw(cycle)
    ),
    lpath(P1,P,Table,R1,R,Len1,Len).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% UNAMBIGUOUS %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Predicates: first normalize input preds...

%% Algorithm taken from Figure 1.28, page 32 from:
%% Roche and Schabes, Introduction. In: Roche and Schabes (eds),
%% Finite State Language Processing. MIT Press 1997.
%%
macro(unambiguous(Expr),unambiguous0(efree_domain(Expr))).

rx(unambiguous0(Expr),Fa) :-
    fsa_regex:rx(Expr,Fa0),
    unambiguous(Fa0,Fa).

unambiguous(Fa00,Fa):-
    fsa_regex:rx(determinize(fa(Fa00)),Fa0),
    fsa_data:fsa_symbols_decl(Fa0,Syms),
    fsa_data:fsa_start_states(Fa0,Starts0),
    fsa_hash:fsa_hash_new(Done),
    fsa_data:index_transitions(Fa0,TransT),
    treat_start_states(Starts0,Starts,Fa0,Ag,Done,Finals1,Finals0,0,Nr0,N,0),
    closure(N,Ag,0,Done,Fa0,TransT,Finals0,Trans,Nr0,Nr),
    sort(Finals1,Finals),
    fsa_data:fsa_components(Syms,Nr,Starts,Finals,Trans,[],Fa1),
    fsa_regex:rx(coaccessible(fa(Fa1)),Fa).

treat_start_states([],[],_,Ag-Ag,_,F,F,Nr,Nr,Ns,Ns).
treat_start_states([H|T],[Nr0|TT],Fa0,[g(H,[])|Ag0]-Ag,D,F0,F,Nr0,Nr,s(N0),N):-
    add_final(g(H,[]),Fa0,F0,F1,Nr0),
    Nr1 is +(Nr0,1),
    fsa_hash:fsa_hash_access(g(H,[]),Nr0,D),
    treat_start_states(T,TT,Fa0,Ag0-Ag,D,F1,F,Nr1,Nr,N0,N).

add_final(g(State,States),Fa0,F0,F,Nr):-
    fsa_data:fsa_final_states(Fa0,Finals),
    (	lists:memberchk(State,Finals),
	\+ ordsets:ord_intersect(States,Finals)
    ->	F0= [Nr|F]
    ;   F0=F
    ).

closure(0,_,_,_,_,_,[],[],N,N).
closure(s(AgC0),[H|Ag0]-Ag1,NrQ0,Done,Fa0,TransT,Finals,Trans,Nr0,Nr):-
    add_final(H,Fa0,Finals,Finals0,NrQ0),
    %% findall could go away
    findall(Ts,fsa_regex_aux:new_trans(H,TransT,Ts),News),
    add_news(News,Ag1,Ag,Done,Trans0,NrQ0,Nr0,Nr1,AgC0,AgC),
    NrQ1 is +(NrQ0,1),
    sort(Trans0,Trans1),
    lists:append(Trans1,Trans2,Trans),
    closure(AgC,Ag0-Ag,NrQ1,Done,Fa0,TransT,Finals0,Trans2,Nr1,Nr).

%% new_trans(+State,+TransTable,-Trans)
new_trans(g(X1,S),Table,trans(g(X1,S),A/W,g(Y1,S1))):-
    fsa_array:fsa_array_get(X1,Keylist,Table),
    same_inputs(Keylist,A,WandT),
    %% WantT is a sorted keylist W-Targets, so we need to pick
    %% one non-deterministically and all keyvals preceeding that
    %% one go into S1
    lists:append(Prefix,[W-Targets|_],WandT),
    append_vals(Prefix,S100),
    get_s1(S,Table,A,S10,S100),
    sort(S10,S1),
    lists:member(Y1,Targets),
    \+ ordsets:ord_member(Y1,S1).

append_vals([],[]).
append_vals([_-V|Tail],Vals):-
    lists:append(V,Vals1,Vals),
    append_vals(Tail,Vals1).

same_inputs([A/W-Targets|Tail],A,[W-Targets|Ntail]):-
    same_inputs_take(Tail,A,Ntail).
same_inputs([A/_-_|Tail],A0,Ntail):-
    same_inputs_not_take(Tail,A,A0,Ntail).

same_inputs_not_take([A/W-Targets|Tail],Not,Take,Ntail):-
    (	A==Not
    ->	same_inputs_not_take(Tail,Not,Take,Ntail)
    ;   same_inputs([A/W-Targets|Tail],Take,Ntail)
    ).

same_inputs_take([],_,[]).
same_inputs_take([A0/W-Targets|Tail],A,Ntail):-
    (	A0==A
    ->	Ntail=[W-Targets|Ntail0],
	same_inputs_take(Tail,A,Ntail0)
    ;   Ntail=[]
    ).

%% Y2 is the list of Y s.t. there is a transition from X to Y,
%% X in S, with input A.
%% get_s1(S,Table,A,Y2,Y2t)
get_s1([],_,_,Y,Y).
get_s1([H|T],Table,A,Y0,Y):-
    get_s10(H,Table,A,Y0,Y1),
    get_s1(T,Table,A,Y1,Y).

get_s10(X,Table,A,Y0,Y):-
    fsa_array:fsa_array_access(X,Keylist,[],Table),
    get_s100(Keylist,A,Y0,Y).

get_s100([],_,Y,Y).
get_s100([A/_-Targets|Tail],A0,Y0,Y):-
    (	A0 @< A
    ->	Y0=Y
    ;   A0 == A
    ->	lists:append(Targets,Y1,Y0),
	get_s100(Tail,A0,Y1,Y)
    ;   get_s100(Tail,A0,Y0,Y)
    ).

add_news([],Ag,Ag,_,[],_,Nr,Nr,C,C).
add_news([trans(_P,Sym,Q)|Trs],Ag0,Ag,Done,[trans(Cur,Sym,V)|Tr0],
	 Cur,Nr0,Nr,C0,C):-
    add_new(Q,Ag0,Ag1,Done,Nr0,Nr1,C0,C1,V),
    add_news(Trs,Ag1,Ag,Done,Tr0,Cur,Nr1,Nr,C1,C).

add_new(Q,Ag0,Ag,Done,Nr0,Nr,C0,C,V):-
    fsa_hash:fsa_hash_access(Q,V,Done),
    (	var(V)
    ->	V=Nr0,
	Nr is +(Nr0,1),
	C = s(C0),
	Ag0=[Q|Ag]
    ;   Ag0=Ag,
	Nr0=Nr,
	C0=C,
	Ag0=Ag
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% IS_FUNCTIONAL %%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Algorithm taken from Figure 1.32, page 36 from:
%% Roche and Schabes, Introduction. In: Roche and Schabes (eds),
%% Finite State Language Processing. MIT Press 1997.
%%
%% However, that algorithm contains a crucial bug. We try to fix that here.
%% [[   Counterexample: is_functional({[a*,b],[a:[]*,c]})
%%      is wrongly determined to be non-functional.
%% ]]
%% This fix is due to an idea by Dale Gerdemann. He wrote:
%% > I thought of something a bit different.
%% > 
%% > Roche & Schabes write (p. 35):
%% > 
%% > 2. if there is a path ((x1,x2,t),om,om',(x1,x2,t')) \in E3 with t \neq t'
%% > ...
%% > 
%% > This is presented as a second terminiation condition. But this 
%% > doesn't seem right. It seems that in this case you ought to generate
%% > a new state (x1,x2,OUT). And then to make sure that the algorithm
%% > terminates, a record should be kept that a state (x1,x2,_) has been 
%% > created and no more such states should be allowed.

%% nb there are typos on line 16 and 17
%% 16: use u.w' instead of w'.u
%% 17: use u.w  instead of w.u
%% why do they state that it's not neccessary to maintain the set
%% of transitions E3? These are used in special_path check. 
is_functional(Fa0,Result):-
    %% add special begin marker to get rid of transductions on epsilon input
    %% a transducer T is functional iff [#,T] is functional.
    fsa_regex:rx(['#',fa(Fa0)],Fa1),
    efree_domain(Fa1,Fa2,Result1),
    (	Result1==cylce
    ->	Result=no
    ;	unambiguous(Fa2,Fa3),
	equivalent(Fa2,Fa3,Result)
    ).

equivalent(Fa2,Fa0Unamb,Result):-
    %% fsa_visualization:fsa_visualization(dotty,Fa2),
    %% fsa_visualization:fsa_visualization(dotty,Fa0Unamb),    
    fsa_data:index_transitions(Fa2,Fa2Table),
    fsa_data:index_transitions(Fa0Unamb,Fa0Table),
    fsa_data:fsa_start_states(Fa2,Starts2),
    fsa_data:fsa_start_states(Fa0Unamb,Starts1),
    fsa_hash:fsa_hash_new(Done),
    treat_start_states_eq(Starts1,Starts2,Ag,Done),
    fsa_hash:fsa_hash_new(E3),
    closure_eq(Ag,Done,Fa0Unamb,Fa0Table,Fa2,Fa2Table,E3,Result).

treat_start_states_eq([],_,[],_).
treat_start_states_eq([S1|S1s],S2s,Ag,Done):-
    treat_start_states_eq0(S2s,S1,Ag,Ag0,Done),
    treat_start_states_eq(S1s,S2s,Ag0,Done).

treat_start_states_eq0([],_,Ag,Ag,_).
treat_start_states_eq0([S2|S2s],S1,[t(S1,S2,plus([]))|Ag0],Ag,Done):-
    fsa_hash:fsa_hash_access(t(S1,S2,plus([])),done,Done),
    treat_start_states_eq0(S2s,S1,Ag0,Ag,Done).

closure_eq([],_,_,_,_,_,_,yes).
closure_eq([t(X1,X2,U)|T],Done,Fa1,Tab1,Fa2,Tab2,E3,Result):-
    (	U \== plus([]),
	fsa_data:fsa_final_state(Fa1,X1),
	fsa_data:fsa_final_state(Fa2,X2)
    ->	Result=no
    ;	news_eq(X1,X2,U,Tab1,Tab2,News),
	add_news_eq(News,Ns,Done,t(X1,X2,U)),
	add_e3(Ns,t(X1,X2,U),E3),
	alter_looping_eq(Ns,Ag,T,E3,Done),
	closure_eq(Ag,Done,Fa1,Tab1,Fa2,Tab2,E3,Result)
    ).

alter_looping_eq([],Ag,Ag,_e3,_done).
alter_looping_eq([t(Y1,Y2,V)|T],Ag0,Ag,E3,Done) :-
    (	inv_path_hash(t(Y1,Y2,U),t(Y1,Y2,V),E3,[t(Y1,Y2,V)]),
	U \== V
    ->	W=out,
	fsa_hash:fsa_hash_access(t(Y1,Y2,W),Val,Done),
	(   var(Val)
	->  Val=done,
	    Ag0=[t(Y1,Y2,W)|T2]
	;   Ag0=T2
	)
    ;   Ag0=[t(Y1,Y2,V)|T2]
    ),
    alter_looping_eq(T,T2,Ag,E3,Done).

add_e3([],_,_).
add_e3([Key|Keys],Val,Assoc):-
    fsa_hash:fsa_hash_access(Key,Prev,Assoc),
    add_vlist(Val,Prev),
    add_e3(Keys,Val,Assoc).

add_vlist(Val,Vlist):-
    (	var(Vlist)
    ->	Vlist=[Val|_]
    ;   Vlist=[_|Vlist0],
	add_vlist(Val,Vlist0)
    ).

in_vlist(Val,Vlist):-
    nonvar(Vlist),
    Vlist=[H|T],
    (	Val=H
    ;   in_vlist(Val,T)
    ).

special_path_hash(Ns,E3):-
    lists:member(t(Y1,Y2,V),Ns),
    inv_path_hash(t(Y1,Y2,U),t(Y1,Y2,V),E3,[t(Y1,Y2,V)]),
    U \== V.

crazy_path(Ns,E3):-
    lists:member(t(X1,U1,X2,U2),Ns),
    inv_path_hash(t(X1,V1,X2,V2),t(X1,U1,X2,U2),E3,[t(X1,U1,X2,U2)]),
    (	U1 \== V1
    ;   U2 \== V2
    ).

inv_path_hash(P,P,_,_).
inv_path_hash(P,Q,E3,Done):-
    fsa_hash:fsa_hash_access(Q,List,E3),
    in_vlist(R,List),
    \+ lists:member(R,Done),
    inv_path_hash(P,R,E3,[R|Done]).

add_news_eq([],[],_,_).
add_news_eq([H|N],News,Done,Source):-
    fsa_hash:fsa_hash_access(H,Val,Done),
    (	var(Val)
    ->	Val=done,
	T0=[H|T1],
	News=[H|News0]
    ;   T0=T1, News=News0
    ),
    add_news_eq(N,News0,Done,Source).

news_eq(X1,X2,U,Tab1,Tab2,News) :-
    fsa_array:fsa_array_access(X1,Keylist1,[],Tab1),
    fsa_array:fsa_array_access(X2,Keylist2,[],Tab2),
    news_eq0(Keylist1,Keylist2,U,News,[]).

news_eq0([],_,_,N,N).
news_eq0([A/W1-Y1s|Rest],Y2s,U,News0,News):-
    news_eq00(Y2s,Y1s,A,W1,U,News0,News1),
    news_eq0(Rest,Y2s,U,News1,News).

news_eq00([],_,_,_,_,N,N).
news_eq00([A2/W2-Y2s|Rest],Y1s,A,W1,U,News0,News):-
    (	A @< A2
    ->	News0=News
    ;   A == A2
    ->  news_eq000(Y1s,Y2s,W1,W2,U,News0,News1),
	news_eq00(Rest,Y1s,A,W1,U,News1,News)
    ;   news_eq00(Rest,Y1s,A,W1,U,News0,News)
    ).

news_eq000(Y1s,Y2s,W1,W2,U,Ns0,Ns):-
    whats_v(U,W1,W2,V),
    news_eq0000(Y1s,Y2s,V,Ns0,Ns).

news_eq0000([],_,_,Ns,Ns).
news_eq0000([Y1|Y1s],Y2s,V,Ns0,Ns):-
    news_eq00000(Y2s,Y1,V,Ns0,Ns1),
    news_eq0000(Y1s,Y2s,V,Ns1,Ns).

news_eq00000([],_,_,Ns,Ns).
news_eq00000([Y2|Y2s],Y1,V,[t(Y1,Y2,V)|Ns0],Ns):-
    news_eq00000(Y2s,Y1,V,Ns0,Ns).

whats_v(U,W1,W2,V):-
    (	U==out
    ->	V=out
    ;	(   U=plus(Uval)
	->  V1=W1, lists:append(Uval,W2,V2)
	;   U=minus(Uval),
	    V2=W2, lists:append(Uval,W1,V1)
	),
	length(V1,V1l),
	length(V2,V2l),
	(   V1l > V2l
	->  (   lists:append(V2,Rest,V1)
	    ->	V=minus(Rest)
	    ;	V=out
	    )
	;   (   lists:append(V1,Rest,V2)
	    ->	V=plus(Rest)
	    ;	V=out
	    )
	)
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%% IS_SUBSEQUENTIAL %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Algorithm taken from Figure 1.45, page 53 from:
%% Roche and Schabes, Introduction. In: Roche and Schabes (eds),
%% Finite State Language Processing. MIT Press 1997.
%% (this is the first algorithm described there; so this is _not_
%% the twinning algorithm. Why this one? Looked easier to implement)
is_subsequential(Fa0,Result) :-
    %% add special begin marker to get rid of transductions on epsilon input
    fsa_regex:rx(['#',fa(Fa0)],Fa1),
    efree_domain(Fa1,Fa2,Result1),
    (	Result1==cycle
    ->	Result=no
    ;   unambiguous(Fa2,Fa3),
	equivalent(Fa2,Fa3,Result2),
	(   Result2==no    %% not functional
	->  Result=no      %% hence not subsequential
	;   fsa_data:fsa_start_states(Fa3,Starts),
	    fsa_data:index_transitions(Fa3,Table),
	    fsa_hash:fsa_hash_new(Done),
	    fsa_hash:fsa_hash_new(E3),
	    treat_starts_sq(Starts,Ag,Done),
	    closure_sq(Ag,Table,Done,E3,Result)
	)
    ).

treat_starts_sq([],[],_).
treat_starts_sq([H|T],[t(H,[],H,[])|Tail],Done):-
    fsa_hash:fsa_hash_access(t(H,[],H,[]),done,Done),
    treat_starts_sq(T,Tail,Done).

closure_sq([],_,_,_,yes).
closure_sq([t(X1,U1,X2,U2)|Ag],Fa,Done,E3,Bool):-
    new_sqs(X1,U1,X2,U2,Fa,News),
    add_news_sq(News,Ns,t(X1,U1,X2,U2),Ag0,Ag,Done),
    add_e3(Ns,t(X1,U1,X2,U2),E3),
    (	crazy_path(Ns,E3)
    ->	Bool=no
    ;   closure_sq(Ag0,Fa,Done,E3,Bool)
    ).

add_news_sq([],[],_,Ag,Ag,_).
add_news_sq([State|States],Ns,Source,Ag0,Ag,Done):-
    fsa_hash:fsa_hash_access(State,Val,Done),
    (	var(Val)
    ->	Val=done,
	Ns=[State|Ns1],
	Ag0=[State|Ag1]
    ;   Ns=Ns1, Ag0=Ag1
    ),
    add_news_sq(States,Ns1,Source,Ag1,Ag,Done).

new_sqs(X1,U1,X2,U2,Table,News):-
    fsa_array:fsa_array_access(X1,Keylist1,[],Table),
    fsa_array:fsa_array_access(X2,Keylist2,[],Table),
    new_sqs0(Keylist1,Keylist2,U1,U2,News,[]).

new_sqs0([],_,_,_,Ns,Ns).
new_sqs0([A/W1-Y1s|Rest],Keylist2,U1,U2,Ns0,Ns):-
    new_sqs00(Keylist2,A,W1,Y1s,U1,U2,Ns0,Ns1),
    new_sqs0(Rest,Keylist2,U1,U2,Ns1,Ns).

new_sqs00([],_,_,_,_,_,Ns,Ns).
new_sqs00([A0/W2-Y2s|Rest],A,W1,Y1s,U1,U2,Ns0,Ns):-
    (	A @< A0
    ->	Ns0=Ns
    ;   A0==A
    ->  new_sqs000(W1,W2,Y1s,Y2s,U1,U2,Ns0,Ns1),
	new_sqs00(Rest,A,W1,Y1s,U1,U2,Ns1,Ns)
    ;   new_sqs00(Rest,A,W1,Y1s,U1,U2,Ns0,Ns)
    ).

new_sqs000(W1,W2,Y1s,Y2s,U1,U2,Ns0,Ns):-
    lists:append(U1,W1,U1W1),
    lists:append(U2,W2,U2W2),
    fsa_strings:longest_common_prefix(U1W1,U2W2,V),
    lists:append(V,V1,U1W1),
    lists:append(V,V2,U2W2),
    new_sqs0000(Y1s,Y2s,V1,V2,Ns0,Ns).

new_sqs0000([],_,_,_,Ns,Ns).
new_sqs0000([Y1|Y1s],Y2s,V1,V2,Ns0,Ns):-
    new_sqs00000(Y2s,Y1,V1,V2,Ns0,Ns1),
    new_sqs0000(Y1s,Y2s,V1,V2,Ns1,Ns).

new_sqs00000([],_,_,_,Ns,Ns).
new_sqs00000([Y2|Y2s],Y1,V1,V2,[t(Y1,V1,Y2,V2)|Ns0],Ns):-
    new_sqs00000(Y2s,Y1,V1,V2,Ns0,Ns).
    

