/* Dale wrote:

Here's one other problem I've given some thought to: I think it would
be useful to be able to find out for a transducer which inputs have
multiple outputs. Suppose you create a transducer which you expect to
be functional, but then you test it with is_functional and find out
that it's not. You'd like to fix the problem, but you have no idea
what inputs to look at. What's needed is an algorithm for finding the
non-functional domain (the set of inputs having more than one output).

I've thought of an approximation algorithm which is rather similar to
our OT operator. So I thought I'd run it past you to see if it makes
sense.

First, represent the transducer as an FSA with input symbols followed
by 0 and output symbols followed by 1. For example:

    a:xy       b:z
0 -------> 1 -------> ((2))

is represented as:

  a    0    x    1    y    1    b    0    z    1
0 -> 1 -> 2 -> 3 -> 4 -> 5 -> 6 -> 7 -> 8 -> 9 -> ((10))


I assume a macro which puts a transducer into this form:

macro(flatten(T), ...).

Now make a filter which filters out cases where the input is the same,
but the output is somehow different from an alternative output. This
filter is constructed by using an analogue of add_violation. To
construct this, I assume that the output alphabet is known. Suppose
that the output alphabet is {x,y,z}, then one could ensure a different
output with the following:
*/

macro(input,[?, 0]).
macro(output,[?,1]).

macro(change_one_output,
   {   [x,1]:[{y,z}:1],
       [y,1]:[{x,z}:1],
       [z,1]:[{x,y}:1]  }).

macro(possibly_change_one_output, { change_one_output, [{x,y,z},1] }).

macro(alter_output,
   { input,possibly_change_one_output }*
                  &
   [{input,possibly_change_one_output }*,
    change_one_output,                      % contains at least one change
    {input,possibly_change_one_output }*]).

macro(shorten_output,
   [{input,possibly_change_one_output}*,  
    [?,1]:[],                             % delete first one
    { input, [?,1]:[] }*                  % possibly delete more
   ]).

macro(lengthen_output, 
   [{input,possibly_change_one_output}*,
    ([]:[{x,y,z},1])+]).

macro(change_output(T), 
    range(T 
          o
 {alter_output, shorten_output, lengthen_output}
          o
    permute_inputs    % defined below
 )).


macro(functional_domain(T),
   range( (flatten(T) - change_output(flatten(T)))
                   o
          { [?,0:[]], [?,1]:[] })).

macro(non_functional_domain(T),
   domain(T) - functional_domain(T)).


% For permute_inputs, we need to know the input alphabet. Suppose
% the alphabet is {a,b}
macro(permute_inputs,
  [{ [[{input,output}*, [a,0]:[], output+, []:[a,0]]*],
     [[{input,output}*, [b,0]:[], output+, []:[b,0]]*],
     [[{input,output}*, []:[a,0], output+, [a,0]:[]]*],
     [[{input,output}*, []:[b,0], output+, [b,0]:[]]*] },
   {input,output}*]).

/*
With permutation, we get an approximation like in the OT paper. To
test whether it is exact, you can check:

is_functional(functional_domain(T) o T)

Questions:

Does the algorithm make sense? Have I made any terrible mistakes? 

Do we really need an approximation? Is the non-functional domain of a
transducer irregular? 

How practical is this algorithm? If the input/output alphabets are
large, this would quickly become exceedingly complex.



Even if this algorithm is not generally practical, it could possibly
be refined for a special case like two level rules. For same-length
transducers there would be no need for permute_inputs. And since
there's a restricted list of feasible outputs for each input, one
could avoid mapping every output symbol to every other one in
change_one_output. And if the algorithm can be refined to be useful
for Kimmo-style two level rules, then maybe it could also be useful
for OT as well. Afterall, OT also has two levels and something like
feasible pairs is encoded in faithfulness constraints.

-- Dale


*/



macro(flatten(T),
      flatten0(letter_transducer(T))).

rx(flatten0(X),Fa) :-
   fsa_regex:rx(X,Fa0),
   fsa_data:fsa_copy_except(symbols,Fa0,Fa1,t(M,M),r(M)),
   fsa_data:fsa_copy_except(states,Fa1,Fa2,States0,States),
   fsa_data:fsa_copy_except(transitions,Fa2,Fa,Trans,FlattenedTrans),
   flatten_trans(Trans,States0,States,FlattenedTrans).


flatten_trans([],States,States,[]).
flatten_trans([trans(Source,In/Out,Target)|Ts],States0,States,Trans0) :-
   (   Out == []
   ->  mk_in_trans(Source,Target,In,States0,States1,Trans0,Trans)
   ;   In == []
   ->  mk_out_trans(Source,Target,Out,States0,States1,Trans0,Trans)
   ;   mk_both_trans(Source,Target,In,Out,States0,States1,Trans0,Trans)
   ),
   flatten_trans(Ts,States1,States,Trans).


mk_in_trans(Source,Target,In,States0,States,
     [InTrans1,InTrans2|Trans],Trans) :-
   States is '+'(States0,1),
   InTrans1 = trans(Source,In,States0),
   InTrans2 = trans(States0,0,Target).
mk_out_trans(Source,Target,Out,States0,States,
      [OutTrans1,OutTrans2|Trans],Trans) :-
   States is '+'(States0,1),
   OutTrans1 = trans(Source,Out,States0),
   OutTrans2 = trans(States0,1,Target).
mk_both_trans(Source,Target,In,Out,States0,States,
     [InTrans1,InTrans2,OutTrans1,OutTrans2|Trans],Trans) :-
   States is '+'(States0,3),
   Mid is '+'(States0,1),
   OutputMid is '+'(States0,2),
   InTrans1 = trans(Source,In,States0),
   InTrans2 = trans(States0,0,Mid),
   OutTrans1 = trans(Mid,Out,OutputMid),
   OutTrans2 = trans(OutputMid,1,Target).



