#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "typedef.h"
#include "xfst2fsa.tab.h"
FILE  *fsa_file;			/* file into which the fsa code will be written */
extern FILE *yyin;			/* a pointer to the program file that should be compiled */
extern int yyparse (void);  /* do syntax checking and build the tree */   

/*************************************/
/* int main(int argc, char* argv[]): */
/*									 */
/* The main function.				 */
/*************************************/

int main(int argc, char* argv[])
{
	char *file; 
	/* set initial values */
    SetValues( );
	/* check input arguments */
    if ( argc < 3 )
	{
		printf( "\nInput file or out put file is missing.\n" );
	    exit (-1);
    }
	/* create xfst grammar tree */
	getTree(argv[1]);
	/* open input file */
	if((file=(char*) malloc (strlen(argv[2])+1))==NULL)
	{
		fprintf(stderr,"Error: malloc failed.\n");
		exit(-1);
	}
	strcpy(file,argv[2]);
	if ((fsa_file=fopen(file,"w")) == NULL)
	{ 
	   fprintf(stderr,"Error: problem in opening file %s \n",file);
       exit(-1);
    }
	/* Produce the fsa code into the output file */
	fprintf(fsa_file,"%%%% This file contains the fsa code for the xfst code in %s.\n\n",argv[1]);
	fprintf(fsa_file,":- multifile macro/2.\n");
	fprintf(fsa_file,":- multifile rx/2.\n");
	fprintf(fsa_file,"%%%%  Load macros in macros.pl\n");
	fprintf(fsa_file,":- ensure_loaded(macros).\n\n");
	print_fsa(root);
	PrintMacro( );
	/* free all mallocs */
	free_macro( );
	free_tree(root);
	/* Print results about untranslated xfst code (if exist such) */
	if(unsolved!=0)
		fprintf(stdout,"\nParsing completed. There were %d occurences of undefined operators.\n",unsolved);
	fprintf(stdout,"\n");
	/* program ended successfully */
	return (0);
}

/*************************************************************/
/* void SetValues(void):									 */
/*															 */
/* This function set initial values of the global variables. */
/*************************************************************/


void SetValues(void)
{
	head=NULL;
	tail=NULL;
	index=1;
	unsolved=0;
	return;
}

/*********************************************************************/
/* void getTree(char *codeFile):									 */
/*																	 */
/* This function creates the xfst grammar tree using bison and flex. */
/*********************************************************************/

void getTree(char *codeFile) 
{
	char *file;
	if((file = (char*) malloc (strlen(codeFile)+1))==NULL)
	{
		fprintf(stderr,"Error: malloc failed.\n");
		exit(-1);
	}
	strcpy(file,codeFile);
	/* yyin is an external variable that is used in yyparse as pointer to the xfst source file */
	if ((yyin = fopen(codeFile,"r")) == NULL)
	{ 
	   fprintf(stderr,"Error: file %s does not exist \n",file);
       exit(-1);
    }
	/* yyparse is yacc (bison) function that parse over the xfst code,
	   check syntax and build the xfst tree. */
	yyparse();
	/* root was initialized in yyparse while it was building the tree.
	   root is the pointer to the returned tree. */
	if (root==NULL) 
	{ 
	   fprintf(stderr,"Error: xfst tree was not initialized\n");
       exit(-1);
    }
	return;
}

/**************************************************************************/
/* void print_fsa(NODE r):												  */
/*																     	  */
/* This function is responsible for going over the xfst grammar tree and  */
/* producing the fsa code into the output file.							  */
/* For each node in the tree the function either prints the code into the */
/* output file or calls another functions (in more complicated cases).	  */
/**************************************************************************/

void print_fsa(NODE r)
{
	int j,n,k;
	switch (r->op){
		case REMARK:			/* ! whatever */
			fprintf(fsa_file,"%%%% %s",(r->str)+1); /* no need to print \n since REMARK includes \n */
			break; 
		case DEFINE:			/* define DefName expression; */
			fprintf(fsa_file,"macro(%s,",(r->str)+7); 
			print_fsa(r->child1);
			fprintf(fsa_file,").\n");
			break; 
		case REGEX:				/* read regex expression; */
			fprintf(fsa_file,"macro(regex,"); 
			print_fsa(r->child1);
			fprintf(fsa_file,").\n");
			break; 
		case BEGIN_CODE:		/* tree root operation */
			print_fsa(r->child1);
			break; 
 		case CODE:				/* help operation for generating definitions, remarks and read regex */
			print_fsa(r->child1);
			if(r->children==2)
				print_fsa(r->child2);
			break; 
		case TERM_COMPLEMENT:	/* \A */
			fprintf(fsa_file,"~ (");
			print_fsa(r->child1);
			fprintf(fsa_file,") & ?");
			break;
		case COMPLEMENT:		/* ~A */
			fprintf(fsa_file,"~ (");
			print_fsa(r->child1);
			fprintf(fsa_file,")");
			break; 
		case KLEENE_PLUS:		/* A+ */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,")+");
			break; 
		case KLEENE_STAR:		/* A* */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,")*");
			break; 
		case CONTAINMENT:		/* $A */
			fprintf(fsa_file,"$(");
			print_fsa(r->child1);
			fprintf(fsa_file,")");
			break; 
		case ONE_CONTAINMENT:	/* $.A */
			fprintf(fsa_file,"contain_one((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case MAX_ONE_CONTAINMENT:	/* $?A */
			fprintf(fsa_file,"contain_max_one((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case CONCATENATION_EXPR:	/* A B */
			fprintf(fsa_file,"[(");
			print_fsa(r->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child2);
			fprintf(fsa_file,")]");
			break; 
		case CONCATENATION_REL:		/* R1 R2 */
			fprintf(fsa_file,"[(");
			print_fsa(r->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child2);
			fprintf(fsa_file,")]");
			break; 
		case N_CONCATENATION:	/* A^n */
			fprintf(fsa_file,"match_n(%d,(",atoi((r->str)+1));
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case N_K_CONCATENATION:	/* A^{n,k} */
			r->str[strlen(r->str)-1]='\0';
			k=atoi(strchr(r->str,',')+1);
			(strchr(r->str,','))[0]='\0';
			n=atoi((r->str)+2);
			if(n>k)
			{
				j=k;
				k=n;
				n=j;
			}
			fprintf(fsa_file,"{match_n(%d,(",n);
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			for( j=n+1 ; j<=k ; j++ )
			{
				fprintf(fsa_file,",match_n(%d,(",j);
				print_fsa(r->child1);
				fprintf(fsa_file,"))");
			}
			fprintf(fsa_file,"}");
			break; 
		case MORE_N_CONCATENATION: /* A^>n */
			fprintf(fsa_file,"[match_n(%d,(",atoi((r->str)+2));
			print_fsa(r->child1);
			fprintf(fsa_file,")),(");
			print_fsa(r->child1);
			fprintf(fsa_file,")+]");
			break; 
		case LESS_N_CONCATENATION:	/* A^<n */
			fprintf(fsa_file,"{[]");
			for( j=1 ; j < atoi((r->str)+2) ; j++)
			{
				fprintf(fsa_file,",match_n(%d,(",j);
				print_fsa(r->child1);
				fprintf(fsa_file,"))");
			}
			fprintf(fsa_file,"}");
			break; 
		case UNION_EXPR:	/* A | B */
			fprintf(fsa_file,"{(");
			print_fsa(r->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child2);
			fprintf(fsa_file,")}");
			break; 
		case INTERSECTION_EXPR:	/* A & B */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,") & (");
			print_fsa(r->child2);
			fprintf(fsa_file,")");
			break; 
		case UNION_REL:		/* R1 | R2 */
			fprintf(fsa_file,"{(");
			print_fsa(r->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child2);
			fprintf(fsa_file,")}");
			break; 
		case INTERSECTION_REL:	/* R1 & R2 */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break; 
		case MINUS:		/* A-B */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,") - (");
			print_fsa(r->child2);
			fprintf(fsa_file,")");
			break; 
		case IGNORE:	/* A/B */
			fprintf(fsa_file,"ignore((");
			print_fsa(r->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child2);
			fprintf(fsa_file,"))");
			break; 
		case INTERNAL_IGNORE: /* A./.B */
			fprintf(fsa_file,"internal_ignore((");
			print_fsa(r->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child2);
			fprintf(fsa_file,"))");
			break; 
		case OPTION:	/* (A) */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,")^");
			break; 
		case ORDER:		/* [A] */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,")");
			break; 
		case REVERSE:	/* A.r */
			fprintf(fsa_file,"reverse((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case DOMAIN:	/* A.u or A.1 */
			fprintf(fsa_file,"domain((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case RANGE:		/* A.r or A.2 */
			fprintf(fsa_file,"range((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case CROSSPRODUCT:	/* A .x. B */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,") x (");
			print_fsa(r->child2);
			fprintf(fsa_file,")");
			break; 
		case COMPOSITION:	/* R1 .o. R2 */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,") o (");
			print_fsa(r->child2);
			fprintf(fsa_file,")");
			break; 
		case PAIR:			/* a:b */
			fprintf(fsa_file,"(");
			print_fsa(r->child1);
			fprintf(fsa_file,"):(");
			print_fsa(r->child2);
			fprintf(fsa_file,")");
			break; 
		case INVERT:		/* R.i */
			fprintf(fsa_file,"invert((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case INVERT_EXPR:	/* A.i */
			fprintf(fsa_file,"invert((");
			print_fsa(r->child1);
			fprintf(fsa_file,"))");
			break; 
		case EPSILON:		/* [] or 0 or "" */
			fprintf(fsa_file,"[]");
			break; 
		case ONE_SYMBOL:	/* a or %saved_symbol */
			if(strlen(r->str)==2) /* letter of the form %letter (escape in xfst) */
				if(strcmp((r->str)+1,"?"))
					fprintf(fsa_file,"'%s'",(r->str)+1);
				else fprintf(fsa_file,"escape(?)");
			else fprintf(fsa_file,"'%s'",r->str);
			break;
		case ANY_SYMBOL:	/* ? */
			fprintf(fsa_file,"?");
			break;
		case BOUND:			/* .#. */
			fprintf(fsa_file,"not_defined");
			break;
		case OTHER_NO_QUOTES:	/* basic symbol such as abc%^ */
			print_no_quotes(r->str);
			break;
		case OTHER_QUOTES:		/* basic symbol such as "abc" */
			r->str[(strlen(r->str))-1]='\0';
			fprintf(fsa_file,"'%s'",(r->str)+1);
			break;
		case CHAR_CONCATENATION:	/* {abc} */
			print_char_concat(r->str);
			break;
		case REPLACE:	/* replace rule (can be multiple and/or with context) with -> arrow */
			replace(r);
			break;
		case REPLACE_OPTION:	/* replace rule (can be multiple and/or with context) with (->) arrow */
			replace(r);
			break;
		case REPLACE_DOUBLE:	/* replace rule (can be multiple and/or with context) with <-> arrow */
			replace(r);
			break;
		case REPLACE_DOUBLE_OPTION:	/* replace rule (can be multiple and/or with context) with (<->) arrow */
			replace(r);
			break;
		case REPLACE_REVERSE:	/* replace rule (can be multiple and/or with context) with <- arrow */
			replace(r);
			break;
		case REPLACE_REVERSE_OPTION:	/* replace rule (can be multiple and/or with context) with (<-) arrow */
			replace(r);
			break;
		case LONGEST_REPLACE:	/* replace rule (can be multiple and/or with context) with @-> arrow */
			replace_match(r);
			break;
		case LONGEST_REPLACE_RL:	/* replace rule (can be multiple and/or with context) with ->@ arrow */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case SHORTEST_REPLACE:	/* replace rule (can be multiple and/or with context) with @> arrow */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case SHORTEST_REPLACE_RL:	/* replace rule (can be multiple and/or with context) with >@ arrow */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case MARKUP:	/* A->L...R */
			fprintf(fsa_file,"uncond_rep(");
			print_fsa(r->child1);
			fprintf(fsa_file,",[");
			print_fsa(r->child2->child1);
			fprintf(fsa_file,",");
			print_fsa(r->child1);
			fprintf(fsa_file,",");
			print_fsa(r->child2->child2);
			fprintf(fsa_file,"])");
			break;
		case LONGEST_MARKUP: /* A@->L...R */
			fprintf(fsa_file,"longest_markup(");
			print_fsa(r->child1);
			fprintf(fsa_file,",");
			print_fsa(r->child2->child1);
			fprintf(fsa_file,",");
			print_fsa(r->child2->child2);
			fprintf(fsa_file,")");
			break;
		case SHORTEST_MARKUP_RL:	/* A>@L...R */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case LONGEST_MARKUP_RL:		/* A->@L...R */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case SHORTEST_MARKUP:	/* A@>L...R */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case RESTRICT:	/* A=>L1_R1,...,Ln_Rn */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case BEFORE:	/* A<B */
			fprintf(fsa_file,"( ~ $ [");
			print_fsa(r->child2);
			fprintf(fsa_file,",");
			print_fsa(r->child1);
			fprintf(fsa_file,"] )");
			break; 
		case AFTER:		/* A>B */
			fprintf(fsa_file,"( ~ $ [");
			print_fsa(r->child1);
			fprintf(fsa_file,",");
			print_fsa(r->child2);
			fprintf(fsa_file,"] )");
			break; 
		case SHUFFLE:	/* A<>B */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case ERROR:	/* parse error in yyparse() */
			fprintf(fsa_file,"%%%%parse_error_not_defined.\n");
			unsolved++;
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	return;
}

/**************************************************************************/
/* void print_no_quotes(char *str):										  */
/*																     	  */
/* This function prints basic symbols such as abc%^.					  */
/**************************************************************************/

void print_no_quotes(char *str)
{
	unsigned int i=0;
	fprintf(fsa_file,"'");
	if(str[0]=='%')
		i++;
	for( ; i <= strlen(str)-2 ; i++)
	{
		fprintf(fsa_file,"%c",str[i]);
		if(str[i+1]=='%')
			i++;
	}
	fprintf(fsa_file,"%c'",str[i]);
	return;
}

/**************************************************************************/
/* void print_char_concat(char *str):									  */
/*																     	  */
/* This function prints char concatenation, e.g. {abc} turns into		  */
/* a | b | c (in xfst) or [a,b,c] (in fsa).								  */
/**************************************************************************/

void print_char_concat(char *str)
{
	unsigned int i=0;
	fprintf(fsa_file,"[");
	if(str[0]=='%')
		i++;
	for( ; i <= strlen(str)-2 ; i++)
	{
		fprintf(fsa_file,"'%c',",str[i]);
		if(str[i+1]=='%')
			i++;
	}
	if(str[i]!='%')
		fprintf(fsa_file,"'%c']",str[i]);
	else
		fprintf(fsa_file,"[]]");
	return;
}

/**************************************************************************/
/* void replace(NODE r):												  */
/*																     	  */
/* This function deals with replace rule of all kinds:					  */
/* -- All arrows kinds (->,(->),<-,(<-),<->,(<->),@->,@>,->@,>@).		  */
/* -- All kind of contexts and rule (multiple or single).				  */
/**************************************************************************/

void replace(NODE r)
{
	if((r->child1->op)==NO_CONTEXT)			/* no context */
	{
		if((r->child1->child1->op)==RULE)	/* single rule A->B */
		{
			PrintReplace(r->op);
			print_fsa(r->child1->child1->child1);
			fprintf(fsa_file,"),(");
			print_fsa(r->child1->child1->child2);
			fprintf(fsa_file,"))");
		}
		else	/* multiple rules A1->B1,...,An->Bn where -> can be replaced consistently by (->),<-,(<-),<-> or (<->) */
		{
			if(r->op==REPLACE_DOUBLE_OPTION) /* (<->) is not dealt with */
			{
				fprintf(fsa_file,"not_defined");
				unsolved++;
			}
			else	/* print name and return to later (enter to macros list) */
			{
				fprintf(fsa_file,"regex%d",index);
				MakeMacro(r,index);
				index++;
			}
		}
	}
	else	/* there is context */
	{
		if((r->child1->child1->op)==MERGE_REPLACE)	/* paraller context replacement is not dealt with */
		{
			fprintf(fsa_file,"not_defined");
			unsolved++;
		}
		else	/* one replace but maybe multiple rules or contexts */
		{
			if(((r->child1->child1->child1->op)==MERGE_RULE) || ((r->child1->child1->child2->op)==MERGE_CONTEXT))	/*  parallel context replacement is not dealt with */
			{
				fprintf(fsa_file,"not_defined");
				unsolved++;
			}
			else	/* one context and one rule */
				PrintContextRule(r);
		}
	}
	return;
}

/**************************************************************************/
/* void MakeMacro(NODE r, int num):										  */
/*																     	  */
/* This function enters into the macro list an operation to be dealt with */
/* later.																  */
/**************************************************************************/

void MakeMacro(NODE r, int num)
{   
	MACRO tmp;
	if((tmp= (MACRO )malloc(sizeof(struct macro)))==NULL){
		fprintf(stderr,"Error: malloc failed.\n");
		exit(-1);
	}
	tmp->r=r;
	tmp->num=num;
	tmp->next=NULL;
	if(head==NULL){ /* empty list */
		head=tmp;
		tail=tmp;
	}
	else
	{
		tail->next=tmp;
		tail=tmp;
	}
	return;
}

/**************************************************************************/
/* void PrintReplace (int op):											  */
/*																     	  */
/* This function prints the macro name according to the kind of arrow	  */
/* (->,(->),<-,(<-),<->,(<->)).											  */
/**************************************************************************/

void PrintReplace (int op)
{
	switch (op){
		case REPLACE:
			fprintf(fsa_file,"uncond_rep((");
			break; 
		case REPLACE_OPTION:
			fprintf(fsa_file,"uncond_optional_rep((");
			break;
		case REPLACE_DOUBLE:
			fprintf(fsa_file,"uncond_double_rep((");
			break;
		case REPLACE_DOUBLE_OPTION:
			fprintf(fsa_file,"uncond_optional_double_rep((");
			break;
		case REPLACE_REVERSE:
			fprintf(fsa_file,"uncond_inv_rep((");
			break;
		case REPLACE_REVERSE_OPTION:
			fprintf(fsa_file,"uncond_optional_inv_rep((");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured\n");
			exit(-1);
	}
	return;
}

/**************************************************************************/
/* void PrintMacro (void):												  */
/*																     	  */
/* This function goes over the list of operations that were saved and     */
/* prints their macros.													  */
/**************************************************************************/

void PrintMacro (void)
{
	MACRO tmp=head;
	while (tmp!=NULL)
	{
		PrintParallelRules(tmp);
		tmp=tmp->next;
	}
	return;
}

/**************************************************************************/
/* void PrintContextRule(NODE r):										  */
/*																     	  */
/* This function prints rules with one rule and one contexts			  */													  
/**************************************************************************/

void PrintContextRule(NODE r)
{
	switch(r->op){
		case REPLACE:
			PrintContextRuleReplace(r->child1);
			break; 
		case REPLACE_OPTION:
			PrintContextRuleReplaceOption(r->child1);
			break;
		case REPLACE_DOUBLE:	/* is not dealt with */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case REPLACE_DOUBLE_OPTION: /* is not dealt with */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case REPLACE_REVERSE:
			PrintContextRuleReplaceReverse(r->child1);
			break;
		case REPLACE_REVERSE_OPTION:
			PrintContextRuleReplaceReverseOption(r->child1);
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	return;
}

/**************************************************************************/
/* void PrintParallelRules(MACRO m):									  */
/*																     	  */
/* This function prints rules with multiple rules and no context.		  */													  
/**************************************************************************/

void PrintParallelRules(MACRO m)
{
	int num;
	switch (m->r->op){
		case REPLACE:
			num=ReplaceParallel(m);
			break; 
		case REPLACE_OPTION:
			num=ReplaceOptionParallel(m);
			break;
		case REPLACE_DOUBLE:
			num=ReplaceDoubleParallel(m);
			break;
		case REPLACE_DOUBLE_OPTION:	/* is not dealt with */
			fprintf(fsa_file,"not_defined");
			unsolved++;
			break;
		case REPLACE_REVERSE:
			num=ReverseParallel(m);
			break;
		case REPLACE_REVERSE_OPTION:
			num=ReverseOptionParallel(m);
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	/* print macros for the uppers and lowers in the rules */
	if(m->r->op!=REPLACE_DOUBLE_OPTION)
		PrintUL(m,num);
	return;
}

/**************************************************************************/
/* int FindNumRules(NODE r):											  */
/*																     	  */
/* This function returns the number of rules in a rule with multiple	  */
/* rules and no context.		 										  */
/**************************************************************************/

int FindNumRules(NODE r)
{
	int num=1;
	while((r->op)==MERGE_RULE)
	{
		num++;
		r=r->child1;
	}
	return num;
}

/**************************************************************************/
/* int ReplaceParallel(MACRO m):										  */
/*																     	  */
/* This function prints rules of the form A1->B1,...,An->Bn.			  */		 										  
/**************************************************************************/

int ReplaceParallel(MACRO m)
{
	int i;
	/* find out the number of rules */
	int num=FindNumRules(m->r->child1->child1);
	/* print the general name */
	fprintf(fsa_file,"macro(regex%d,[[n_mac%d,r_mac%d]*,n_mac%d]).\n",m->num,m->num,m->num,m->num);
	/* print the help macros for the general one */
	fprintf(fsa_file,"macro(n_mac%d, ~ $ ({upper1_%d",m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",upper%d_%d",i,m->num);
	fprintf(fsa_file,"} - [])).\n");
	fprintf(fsa_file,"macro(r_mac%d,{upper1_%d x lower1_%d",m->num,m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",upper%d_%d x lower%d_%d",i,m->num,i,m->num);
	fprintf(fsa_file,"}).\n");
	return num;
}

/**************************************************************************/
/* int ReplaceOptionParallel(MACRO m):									  */
/*																     	  */
/* This function prints rules of the form A1(->)B1,...,An(->)Bn.		  */		 										  
/**************************************************************************/

int ReplaceOptionParallel(MACRO m)
{
	int i;
	/* find out the number of rules */
	int num=FindNumRules(m->r->child1->child1);
	/* print the general name */
	fprintf(fsa_file,"macro(regex%d,[[ ? * ,r_mac%d]*, ? * ]).\n",m->num,m->num);
	/* print the help macros for the general one */
	fprintf(fsa_file,"macro(r_mac%d,{upper1_%d x lower1_%d",m->num,m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",upper%d_%d x lower%d_%d",i,m->num,i,m->num);
	fprintf(fsa_file,"}).\n");
	return num;
}

/**************************************************************************/
/* int ReverseParallel(MACRO m):										  */
/*																     	  */
/* This function prints rules of the form A1<-B1,...,An<-Bn.			  */		 										  
/**************************************************************************/

int ReverseParallel(MACRO m)
{
	int i;
	/* find out the number of rules */
	int num=FindNumRules(m->r->child1->child1);
	/* print the general name */
	fprintf(fsa_file,"macro(regex%d,invert([[n_mac%d,r_mac%d]*,n_mac%d])).\n",m->num,m->num,m->num,m->num);
	/* print the help macros for the general one */
	fprintf(fsa_file,"macro(n_mac%d, ~ $ ({lower1_%d",m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",lower%d_%d",i,m->num);
	fprintf(fsa_file,"} - [])).\n");
	fprintf(fsa_file,"macro(r_mac%d,{lower1_%d x upper1_%d",m->num,m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",lower%d_%d x upper%d_%d",i,m->num,i,m->num);
	fprintf(fsa_file,"}).\n");
	return num;
}

/**************************************************************************/
/* int ReverseOptionParallel(MACRO m):									  */
/*																     	  */
/* This function prints rules of the form A1(<-)B1,...,An(<-)Bn.		  */
/**************************************************************************/

int ReverseOptionParallel(MACRO m)
{
	int i;
	/* find out the number of rules */
	int num=FindNumRules(m->r->child1->child1);
	/* print the general name */
	fprintf(fsa_file,"macro(regex%d,invert([[ ? * ,r_mac%d]*, ? * ])).\n",m->num,m->num);
	/* print the help macros for the general one */
	fprintf(fsa_file,"macro(r_mac%d,{lower1_%d x upper1_%d",m->num,m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",lower%d_%d x upper%d_%d",i,m->num,i,m->num);
	fprintf(fsa_file,"}).\n");
	return num;
}

/**************************************************************************/
/* int ReplaceDoubleParallel(MACRO m):									  */
/*																     	  */
/* This function prints rules of the form A1<->B1,...,An<->Bn.			  */
/**************************************************************************/

int ReplaceDoubleParallel(MACRO m)
{
	int i;
	/* find out the number of rules */
	int num=FindNumRules(m->r->child1->child1);
	/* regex */
	fprintf(fsa_file,"macro(regex%d,stage1_%d o stage2_%d o stage3_%d o stage1_%d).\n",m->num,m->num,m->num,m->num,m->num);
	/* stage 1 */
	fprintf(fsa_file,"macro(stage1_%d, ~ $ ({'^^^1'",m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",'^^^%d'",i);
	fprintf(fsa_file,"})).\n");
	/* stage 2 */
	fprintf(fsa_file,"macro(stage2_%d,[[n2_mac%d,r2_mac%d]* ,n2_mac%d] ).\n",m->num,m->num,m->num,m->num);
	/* n_mac for stage 2 */
	fprintf(fsa_file,"macro(n2_mac%d, ~ $ ({upper1_%d",m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",upper%d_%d",i,m->num);
	fprintf(fsa_file,"} - [])).\n");
	/* r_mac for stage 2 */
	fprintf(fsa_file,"macro(r2_mac%d,{upper1_%d x '^^^1'",m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",upper%d_%d x '^^^%d'",i,m->num,i);
	fprintf(fsa_file,"}).\n");
	/* stage 3 */
	fprintf(fsa_file,"macro(stage3_%d,invert([[n3_mac%d,r3_mac%d]* ,n3_mac%d]) ).\n",m->num,m->num,m->num,m->num);
	/* n_mac for stage 3 */
	fprintf(fsa_file,"macro(n3_mac%d, ~ $ ({lower1_%d",m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",lower%d_%d",i,m->num);
	fprintf(fsa_file,"} - [])).\n");
	/* r_mac for stage 3 */
	fprintf(fsa_file,"macro(r3_mac%d,{lower1_%d x '^^^1'",m->num,m->num);
	for ( i=2 ; i<=num ; i++)
		fprintf(fsa_file,",lower%d_%d x '^^^%d'",i,m->num,i);
	fprintf(fsa_file,"}).\n");
	return num;
}

/**************************************************************************/
/* void PrintUL(MACRO m,int num):										  */
/*																     	  */
/* This function prints macros for the UPPERs and LOWERs of rules.		  */
/**************************************************************************/

void PrintUL(MACRO m,int num)
{
	int i;
	NODE tmp=m->r->child1->child1;
	for( i=1 ; i<=num-2 ; i++)
	{
		/* Upper macro */
		fprintf(fsa_file,"macro(upper%d_%d,(",i,m->num);
		print_fsa(tmp->child2->child1);
		fprintf(fsa_file,") ).\n");
		/* Lower macro */
		fprintf(fsa_file,"macro(lower%d_%d,(",i,m->num);
		print_fsa(tmp->child2->child2);
		fprintf(fsa_file,") ).\n");
		/*move to next rule */
		tmp=tmp->child1;
	}
	/* move to last MERGE_RULE */
	/* Upper macro - one before last */
	fprintf(fsa_file,"macro(upper%d_%d,(",num-1,m->num);
	print_fsa(tmp->child2->child1);
	fprintf(fsa_file,") ).\n");
	/* Lower macro - one before last */
	fprintf(fsa_file,"macro(lower%d_%d,(",num-1,m->num);
	print_fsa(tmp->child2->child2);
	fprintf(fsa_file,") ).\n");
	/* Upper macro - last one */
	fprintf(fsa_file,"macro(upper%d_%d,(",num,m->num);
	print_fsa(tmp->child1->child1);
	fprintf(fsa_file,") ).\n");
	/* Lower macro - last one */
	fprintf(fsa_file,"macro(lower%d_%d,(",num,m->num);
	print_fsa(tmp->child1->child2);
	fprintf(fsa_file,") ).\n");
	return;
}

/**************************************************************************/
/* void PrintContextRuleReplace (NODE r):								  */
/*																     	  */
/* This function prints rules of the form UPPER->LOWER||LEFT_RIGHT.		  */
/* || can be replaced by either \\, // or \/.							  */
/**************************************************************************/

void PrintContextRuleReplace (NODE r)
{
	switch(r->op){
		case OR_OR:
			fprintf(fsa_file,"cond_rep_or_or");
			break; 
		case SL_SL:
			fprintf(fsa_file,"cond_rep_sl_sl");
			break;
		case BSL_BSL:
			fprintf(fsa_file,"cond_rep_bsl_bsl");
			break;
		case BSL_SL:
			fprintf(fsa_file,"cond_rep_bsl_sl");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	/* check for boundary case - LEFT is at beginning */
	PrintStart(r->child1->child2);
	/* check for boundary case - RIGHT is at end */
	PrintEnd(r->child1->child2);
	fprintf(fsa_file,"((");
	/* print UPPER */
	print_fsa(r->child1->child1->child1);
	fprintf(fsa_file,"),(");
	/* print LOWER */
	print_fsa(r->child1->child1->child2);
	fprintf(fsa_file,"),(");
	/* print LEFT context */
	PrintLeft(r->child1->child2);
	fprintf(fsa_file,"),(");
	/* print RIGHT context */
	PrintRight(r->child1->child2);
	fprintf(fsa_file,"))");
	return;
}

/**************************************************************************/
/* void PrintLeft(NODE r):												  */
/*																     	  */
/* This function prints the left context of a replacement rule with only  */
/* one context.															  */
/**************************************************************************/

void PrintLeft(NODE r)
{
	switch(r->op){
		case CONTEXT:					/* there is left context */
			print_fsa(r->child1);
			break; 
		case CONTEXT_START:				/* there is left context */
			print_fsa(r->child1);
			break; 
		case CONTEXT_END:				/* there is left context */
			print_fsa(r->child1);
			break; 
		case CONTEXT_START_END:			/* there is left context */
			print_fsa(r->child1);
			break; 
		case RIGHT_CONTEXT:				/* there is left context */
			print_fsa(r->child1);
			break;
		case RIGHT_CONTEXT_START:		/* there is left context */
			print_fsa(r->child1);
			break;
		case RIGHT_CONTEXT_END:			/* there is left context */
			print_fsa(r->child1);
			break;
		case RIGHT_CONTEXT_START_END:	/* there is left context */
			print_fsa(r->child1);
			break;
		case LEFT_CONTEXT:				/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case LEFT_CONTEXT_START:		/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case LEFT_CONTEXT_END:			/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case LEFT_CONTEXT_START_END:	/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS:			/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS_START:		/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS_END:		/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS_START_END:	/* left context is empty */
			fprintf(fsa_file,"[]");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	return;
}

/**************************************************************************/
/* void PrintRight(NODE r):												  */
/*																     	  */
/* This function prints the right context of a replacement rule with only */
/* one context.															  */
/**************************************************************************/

void PrintRight(NODE r)
{
	switch(r->op){
		case CONTEXT:						/* there is right context */
			print_fsa(r->child2);
			break; 
		case CONTEXT_START:					/* there is right context */
			print_fsa(r->child2);
			break; 
		case CONTEXT_END:					/* there is right context */
			print_fsa(r->child2);
			break; 
		case CONTEXT_START_END:				/* there is right context */
			print_fsa(r->child2);
			break; 
		case RIGHT_CONTEXT:					/* right context is empty */
			fprintf(fsa_file,"[]");			
			break;
		case RIGHT_CONTEXT_START:			/* right context is empty */	
			fprintf(fsa_file,"[]");
			break;
		case RIGHT_CONTEXT_END:				/* right context is empty */
			fprintf(fsa_file,"[]");
			break;
		case RIGHT_CONTEXT_START_END:		/* right context is empty */
			fprintf(fsa_file,"[]");
			break;
		case LEFT_CONTEXT:					/* there is right context */
			print_fsa(r->child1);
			break;
		case LEFT_CONTEXT_START:			/* there is right context */
			print_fsa(r->child1);
			break;
		case LEFT_CONTEXT_END:				/* there is right context */
			print_fsa(r->child1);
			break;
		case LEFT_CONTEXT_START_END:		/* there is right context */
			print_fsa(r->child1);
			break;
		case EMPTY_CONTEXTS:				/* right context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS_START:			/* right context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS_END:			/* right context is empty */
			fprintf(fsa_file,"[]");
			break;
		case EMPTY_CONTEXTS_START_END:		/* right context is empty */
			fprintf(fsa_file,"[]");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	return;
}

/**************************************************************************/
/* void PrintContextRuleReplaceOption (NODE r):							  */
/*																     	  */
/* This function prints rules of the form UPPER(->)LOWER||LEFT_RIGHT.	  */
/* || can be replaced by either \\, // or \/.							  */
/**************************************************************************/

void PrintContextRuleReplaceOption (NODE r)
{
	switch(r->op){
		case OR_OR:
			fprintf(fsa_file,"cond_rep_option_or_or");
			break; 
		case SL_SL:
			fprintf(fsa_file,"cond_rep_option_sl_sl");
			break;
		case BSL_BSL:
			fprintf(fsa_file,"cond_rep_option_bsl_bsl");
			break;
		case BSL_SL:
			fprintf(fsa_file,"cond_rep_option_bsl_sl");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	/* check for boundary case - LEFT is at beginning */
	PrintStart(r->child1->child2);
	/* check for boundary case - RIGHT is at end */
	PrintEnd(r->child1->child2);
	fprintf(fsa_file,"((");
	/* print UPPER */
	print_fsa(r->child1->child1->child1);
	fprintf(fsa_file,"),(");
	/* print LOWER */
	print_fsa(r->child1->child1->child2);
	fprintf(fsa_file,"),(");
	/* print LEFT context */
	PrintLeft(r->child1->child2);
	fprintf(fsa_file,"),(");
	/* print RIGHT context */
	PrintRight(r->child1->child2);
	fprintf(fsa_file,"))");
	return;
}

/**************************************************************************/
/* void PrintContextRuleReplaceReverse (NODE r):						  */
/*																     	  */
/* This function prints rules of the form UPPER<-LOWER||LEFT_RIGHT.		  */
/* || can be replaced by either \\, // or \/.							  */
/**************************************************************************/

void PrintContextRuleReplaceReverse (NODE r)
{
	switch(r->op){
		case OR_OR:
			fprintf(fsa_file,"cond_rep_reverse_or_or");
			break; 
		case SL_SL:
			fprintf(fsa_file,"cond_rep_reverse_sl_sl");
			break;
		case BSL_BSL:
			fprintf(fsa_file,"cond_rep_reverse_bsl_bsl");
			break;
		case BSL_SL:
			fprintf(fsa_file,"cond_rep_reverse_bsl_sl");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	/* check for boundary case - LEFT is at beginning */
	PrintStart(r->child1->child2);
	/* check for boundary case - RIGHT is at end */
	PrintEnd(r->child1->child2);
	fprintf(fsa_file,"((");
	/* print UPPER */
	print_fsa(r->child1->child1->child1);
	fprintf(fsa_file,"),(");
	/* print LOWER */
	print_fsa(r->child1->child1->child2);
	fprintf(fsa_file,"),(");
	/* print LEFT context */
	PrintLeft(r->child1->child2);
	fprintf(fsa_file,"),(");
	/* print RIGHT context */
	PrintRight(r->child1->child2);
	fprintf(fsa_file,"))");
	return;
}

/**************************************************************************/
/* void PrintContextRuleReplaceReverseOption (NODE r):					  */
/*																     	  */
/* This function prints rules of the form UPPER(<-)LOWER||LEFT_RIGHT.	  */
/* || can be replaced by either \\, // or \/.							  */
/**************************************************************************/

void PrintContextRuleReplaceReverseOption (NODE r)
{
	switch(r->op){
		case OR_OR:
			fprintf(fsa_file,"cond_rep_reverse_option_or_or");
			break; 
		case SL_SL:
			fprintf(fsa_file,"cond_rep_reverse_option_sl_sl");
			break;
		case BSL_BSL:
			fprintf(fsa_file,"cond_rep_reverse_option_bsl_bsl");
			break;
		case BSL_SL:
			fprintf(fsa_file,"cond_rep_reverse_option_bsl_sl");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	/* check for boundary case - LEFT is at beginning */
	PrintStart(r->child1->child2);
	/* check for boundary case - RIGHT is at end */
	PrintEnd(r->child1->child2);
	fprintf(fsa_file,"((");
	/* print UPPER */
	print_fsa(r->child1->child1->child1);
	fprintf(fsa_file,"),(");
	/* print LOWER */
	print_fsa(r->child1->child1->child2);
	fprintf(fsa_file,"),(");
	/* print LEFT context */
	PrintLeft(r->child1->child2);
	fprintf(fsa_file,"),(");
	/* print RIGHT context */
	PrintRight(r->child1->child2);
	fprintf(fsa_file,"))");
	return;
}

/**************************************************************************/
/* void replace_match (NODE r):											  */
/*																     	  */
/* This function deals with replacement rules with @-> arrow.			  */
/**************************************************************************/

void replace_match(NODE r)
{
	if(((r->child1->op)==NO_CONTEXT) & ((r->child1->child1->op)==RULE))	/* single rule A@->B */
	{
		fprintf(fsa_file,"longest_match_replacement((");
		print_fsa(r->child1->child1->child1);
		fprintf(fsa_file,"),(");
		print_fsa(r->child1->child1->child2);
		fprintf(fsa_file,"))");
	}
	else	/* there is context or multiple rules without contexts */
	{
		fprintf(fsa_file,"not_defined");
		unsolved++;
	}
	return;
}

/**************************************************************************/
/* void free_tree (NODE r):												  */
/*																     	  */
/* This function frees the allocation of the grammar tree.				  */
/**************************************************************************/

void free_tree(NODE r)
{
	if(r->child1!=NULL)
		free_tree(r->child1);
	if(r->child2!=NULL)
		free_tree(r->child2);
	free(r);
	return;
}

/**************************************************************************/
/* void free_macro (NODE r):											  */
/*																     	  */
/* This function frees the allocation of list of macros.				  */
/**************************************************************************/

void free_macro(void)
{
	MACRO tmp;
	while(head!=NULL)
	{
		tmp=head;
		head=head->next;
		free(tmp);
	}
	return;
}

/**************************************************************************/
/* void PrintStart(NODE r):												  */
/*																     	  */
/* This function prints extensions to replace macros if it is a boundary  */
/* case where the left context is at the beginning.						  */
/**************************************************************************/

void PrintStart(NODE r)
{
	switch(r->op){
		case CONTEXT:
			break; 
		case CONTEXT_START:					/* at start */
			fprintf(fsa_file,"_start");
			break; 
		case CONTEXT_END:					/* not at start */
			break; 
		case CONTEXT_START_END:				/* at start */
			fprintf(fsa_file,"_start");
			break; 
		case RIGHT_CONTEXT:					/* not at start */
			break;
		case RIGHT_CONTEXT_START:			/* at start */
			fprintf(fsa_file,"_start");
			break;
		case RIGHT_CONTEXT_END:				/* not at start */
			break;
		case RIGHT_CONTEXT_START_END:		/* at start */
			fprintf(fsa_file,"_start");
			break;
		case LEFT_CONTEXT:					/* not at start */
			break;
		case LEFT_CONTEXT_START:			/* at start */
			fprintf(fsa_file,"_start");
			break;
		case LEFT_CONTEXT_END:				/* not at start */
			break;
		case LEFT_CONTEXT_START_END:		/* at start */
			fprintf(fsa_file,"_start");
			break;
		case EMPTY_CONTEXTS:				/* not at start */
			break;
		case EMPTY_CONTEXTS_START:			/* at start */
			fprintf(fsa_file,"_start");
			break;
		case EMPTY_CONTEXTS_END:			/* not at start */
			break;
		case EMPTY_CONTEXTS_START_END:		/* at start */
			fprintf(fsa_file,"_start");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	return;
}

/**************************************************************************/
/* void PrintEnd(NODE r):												  */
/*																     	  */
/* This function prints extensions to replace macros if it is a boundary  */
/* case where the right context is at the end.							  */
/**************************************************************************/

void PrintEnd(NODE r)
{
	switch(r->op){
		case CONTEXT:					/* not at end */
			break; 
		case CONTEXT_START:				/* not at end */
			break; 
		case CONTEXT_END:				/* at end */
			fprintf(fsa_file,"_end");
			break; 
		case CONTEXT_START_END:			/* at end */
			fprintf(fsa_file,"_end");
			break; 
		case RIGHT_CONTEXT:				/* not at end */
			break;
		case RIGHT_CONTEXT_START:		/* not at end */
			break;
		case RIGHT_CONTEXT_END:			/* at end */
			fprintf(fsa_file,"_end");
			break;
		case RIGHT_CONTEXT_START_END:	/* at end */
			fprintf(fsa_file,"_end");
			break;
		case LEFT_CONTEXT:				/* not at end */
			break;
		case LEFT_CONTEXT_START:		/* not at end */
			break;
		case LEFT_CONTEXT_END:			/* at end */
			fprintf(fsa_file,"_end");
			break;
		case LEFT_CONTEXT_START_END:	/* at end */
			fprintf(fsa_file,"_end");
			break;
		case EMPTY_CONTEXTS:			/* not at end */
			break;
		case EMPTY_CONTEXTS_START:		/* not at end */
			break;
		case EMPTY_CONTEXTS_END:		/* at end */
			fprintf(fsa_file,"_end");
			break;
		case EMPTY_CONTEXTS_START_END:	/* at end */
			fprintf(fsa_file,"_end");
			break;
		default:
			fprintf(stderr,"Unknown Token. Parse error occured.\n");
			exit(-1);
	}
	return;
}

