%% Finite State Encoding of the Knapsack Problem
%% As an example problem, try
%%
%%   knapsack(50,[item(i1,10,60,1),item(i2,20,100,1),item(i3,30,120,1)])
%%
%% This is the example found in Cormen et al, "Introduction to Algorithms"
%% Each item is represented by: item(ItemName,Weight,Value,Quantity)
%%
%% The empty bag in this implementation is represented by a sequence
%% of 1's. Each 1 represents some capacity that can be filled. Filling
%% the bag creates a new sequence of 1's where the 1's here represent
%% value rather than weight. The output side also has item names
%% interspersed with the 1's, so that we know what sequence of items
%% has the value represented by the 1's.
%%
%% Note (for those interested in linguistic problems): this problem
%% doesn't appear to have much linguistic interest. The strategy for
%% solving the problem, however, may share some features with some
%% approaches to optimality theory.
%% 
%% Note (for those interested in efficient algorithms): this approach
%% is pretty slow. Could be that you don't really want to solve this
%% problem using finite state methods. Could also be that if you do
%% use finite state methods, you can still find a more efficient
%% approach. One idea to try might be to encode numbers (Weights and
%% Values) with binary numbers instead of unary.

macro(knapsack(BagSize,Items),
  minimize(range(
     concat_n(BagSize)
          o
       FillBag
          o
    most_valuable(BagSize,FillBag,set(ItemNames))
          o
    set([1 x []|ItemNames])*
  ))) :-
 
  bag_filler(Items,FillBag),
  item_names(Items,ItemNames).


%% most_valuable/3
%% This is the filter used to ensure that value is optimized

macro(most_valuable(BagSize,FillBag,Items),
  ~ range( concat_n(BagSize) 
                 o 
              FillBag 
                 o 
  [{1,Items x []},[] x Items^]* % the filter must ignore item names
                 o
        [? *, 1 x [], ? *]+ % decrement the value of the filled bag by
                            % at least 1
  )).                       


bag_filler([],[(? * x [])]).
bag_filler([item(ItemName,Weight,Value,Quantity)|R0],
           [concat_leq_n(Quantity,
                         (concat_n(Weight) 
                                x
                    [concat_n(Value),ItemName]))|R]) :-
   bag_filler(R0,R).


item_names([],[]).
item_names([item(ItemName,_,_,_)|R0],[ItemName|R]) :-
   item_names(R0,R).

macro(concat_n(0),[]).
macro(concat_n(N),[1,concat_n(N1)]) :-
   N1 is N-1.

macro(concat_leq_n(0,_),[]).
macro(concat_leq_n(N,A),[A^,concat_leq_n(N1,A)]) :-
   N1 is N-1.



