/*----------------------------------------------------------------------
caslr.pl

http://www.ling.ohio-state.edu/~nvail/msol.html

Simulation of
Closed Alphabet Same-Length Relations
in recoginziers

(c) Nathan Vaillette (nvail@ling.ohio-state.edu)

Last updated July 10, 01

Tested with the Finite State Utilities v. 6.234

               *             *               *

This file provides macros for simulating transducers in recognizers;
e.g. the input/output pair a:b is simulated with the symbol ab.  We
use the alphabet {a,b,c,d,e}, plus the "blank" symbol "o". Our intention
is for "o" to stand proxy for "epsilon", like the 0 in two-level
morphology.

-------------------------------------------------------------------------*/

:- multifile macro/2 .
:- multifile rx/2 .



%% mimics ?
macro(caslr_alphabet, {a,b,c,d,e,o}).

%% mimics id(?)
macro(caslr_id,{aa,bb,cc,dd,ee,oo}).

%% input_s mimics s x ?
macro(input_o,{
	      oa,
	      ob,
	      oc,
	      od,
	      oe,
	      oo
	     }).

macro(input_a,{
	       aa,
	       ab,
	       ac,
	       ad,
	       ae,
	       ao
	      
	      }).
macro(input_b,{
	       ba,
	       bb,
	       bc,
	       bd,
	       be,
	       bo
	       }).
macro(input_c,{
	       ca,
	       cb,
	       cc,
	       cd,
	       ce,
	       co
	       }).
macro(input_d,{
	       da,
	       db,
	       dc,
	       dd,
	       de,
	       do
	      
	       }).
macro(input_e,{
	       ea,
	       eb,
	       ec,
	       ed,
	       ee,
	       eo
		       
	       }).



%% output_s mimics ? x s
macro(output_o,{
	       oo,
	       ao,
	       bo,
	       co,
	       do,
	       eo
	     }).
	     
	     

macro(output_a,{
		oa,
	     aa,
	     ba,
	     ca,
	     da,
	     ea
	       }).
macro(output_b,{
		ob,
	     ab,
	     bb,
	     cb,
	     db,
	     eb
	       }).
macro(output_c,{
		oc,

	     ac,
	     bc,
	     cc,
	     dc,
	     ec
	       }).
macro(output_d,{
		od,

	     ad,
	     bd,
	     cd,
	     dd,
	     ed
	       }).
macro(output_e,{oe,	       

	     ae,
	     be,
	     ce,
	     de,
	     ee
	       }).


%% mimics (? x ?)*
macro(alpha,{input_a,input_b,input_c,input_d,input_e,input_o}).

%% mimics domain(L)
macro(caslr_domain(L),
      domain({o x input_o,
	      a x input_a,
	      b x input_b,
	      c x input_c,
	      d x input_d,
	      e x input_e}*
	    o
	    L)
     ).

%% mimics range(L)
macro(caslr_range(L),
      range(

	    L
	   o{ output_o x o,
	     output_a x a,
	      output_b x b,
	      output_c x c,
	      output_d x d,
	      output_e x e}*
	   )

     ).

%% mimics id(L)
macro(caslr_id(L),
      range(L
	   o
	   {a x aa,
	    b x bb,
	    c x cc,
	    d x dd,
	    e x ee,
	    o x oo
	  }*
	   )
     ).

%% = non_det_range(caslr_id(L))
%% mimics L xx ? * 
macro(string_times_caslr(L),
      range(L
	   o
	   {o x input_o,
	     a x input_a,
	      b x input_b,
	      c x input_c,
	      d x input_d,
	      e x input_e}*
	   )
     ).

%% mimics ? * xx L
macro(caslr_times_string(L),
      non_det_domain(caslr_id(L))).


%% mimics L o (? x ?)*
macro(non_det_range(L),
      range(L
	   o
	    { input_o x input_o,
	     input_a x input_a,
	      input_b x input_b,
	      input_c x input_c,
	      input_d x input_d,
	      input_e x input_e}*
	    )
     ).

%% mimics (? x ?)* o L
macro(non_det_domain(L),
      range(L
	   o
	    { output_o x output_o,
	     output_a x output_a,
	      output_b x output_b,
	      output_c x output_c,
	      output_d x output_d,
	      output_e x output_e}*
	    )
     ).


%% mimics range(String o T)
macro(caslr_apply(T,String),
      caslr_range(string_times_caslr(String) & T)).

%% converts a caslr simulation of a relation to a real
%% transducer for the relation:

macro(caslr_to_trans(L),
      { o x input_o,
	a x input_a,
	b x input_b,
	c x input_c,
	d x input_d,
	e x input_e}*
      
      o L 
 o
      
       { output_o x o,
 	output_a x a,
 	output_b x b,
 	output_c x c,
 	output_d x d,
 	output_e x e}*
).

%% converts a transducer over {a,b,c,d,e} into its caslr representation
macro(trans_to_caslr(R),sltrans_to_caslr(slify(R,o))).

%% turns a transducer for R into a same_legnth transducer for the R with
%% S's representing epsilons

rx(slify(R,S),FSA) :-
	fsa_regex:rx(letter_transducer(R),FSA0),
	
	fsa_data:fsa_start_states(FSA0,Starts),
	fsa_data:fsa_final_states(FSA0,Finals),
	fsa_data:fsa_jumps(FSA0,Jumps),
	fsa_data:fsa_transitions(FSA0,Edges0),
	slify(Edges0,S,Edges),
	fsa_data:fsa_construct_rename_states(Starts,Finals,Edges,Jumps,FSA).

slify([],_,[]).
slify([trans(P,X/[],Q)|Edges0],S,[trans(P,X/S,Q)|Edges]) :-
	atomic(X),
	!,
	slify(Edges0,S,Edges).
slify([trans(P,[]/Y,Q)|Edges0],S,[trans(P,S/Y,Q)|Edges]) :-
	atomic(Y),
	!,
	slify(Edges0,S,Edges).

slify([trans(P,X/Y,Q)|Edges0],S,[trans(P,X/Y,Q)|Edges]) :-
	atomic(X),
	atomic(Y),
	!,
	slify(Edges0,S,Edges).

slify([trans(P,in(L)/[],Q)|Edges0],S,[trans(P,in(L)/S,Q)|Edges]) :-
	!,
	slify(Edges0,S,Edges).
slify([trans(P,not_in(L)/[],Q)|Edges0],S,[trans(P,not_in(L)/S,Q)|Edges]) :-
	!,
	slify(Edges0,S,Edges).
slify([trans(P,[]/in(L),Q)|Edges0],S,[trans(P,S/in(L),Q)|Edges]) :-
	!,
	slify(Edges0,S,Edges).
slify([trans(P,[]/not_in(L),Q)|Edges0],S,[trans(P,S/not_in(L),Q)|Edges]) :-
	!,
	slify(Edges0,S,Edges).

slify([trans(P,'$@'(X)/'$@'(X),Q)|Edges0],S,[trans(P,'$@'(X)/'$@'(X),Q)|Edges]) :-
	!,
     slify(Edges0,S,Edges).

slify([trans(P,T,Q)|Edges0],S,[trans(P,T,Q)|Edges]) :-
     slify(Edges0,S,Edges).

rx(sltrans_to_caslr(R),FSA) :-
	fsa_regex:rx(R,FSA0),
	
	fsa_data:fsa_start_states(FSA0,Starts),
	fsa_data:fsa_final_states(FSA0,Finals),
	fsa_data:fsa_jumps(FSA0,Jumps),
	fsa_data:fsa_transitions(FSA0,Edges0),
	findall(trans(P,T,Q),
		(   member(trans(P,T0,Q),Edges0),
		    caslr_trans(T0,T)
		),
		Edges),
	fsa_data:fsa_construct_rename_states(Starts,Finals,Edges,Jumps,FSA).

caslr_trans('$@'(I0)/'$@'(I0),T) :-
	!,
	fix_alphabet(I0,I),
	caslr_expand_id(I,T).
caslr_trans(I0/O0,T) :-
	fix_alphabet(I0,I),
	fix_alphabet(O0,O),
	caslr_expand(I,O,T).



fix_alphabet(not_in(L0),L) :-
	!,
	findall(X,
		(   member(X,[a,b,c,d,e]),
		    \+ member(X,L0)
		),
		L).
fix_alphabet(in(L),L) :- !.
fix_alphabet(X,[X]).

caslr_expand([X],[Y],Sym) :-
	!,
	name(X,[A]),
	name(Y,[B]),
	name(Sym,[A,B]).
caslr_expand(L1,L2,in(T)) :-
	findall(Sym,
		(   
		    member(X,L1),
		    member(Y,L2),
		    name(X,[A]),
		    name(Y,[B]),
		    name(Sym,[A,B])
		),
		T).
caslr_expand_id(L,in(T)) :-
	findall(Sym,
		(   
		    member(X,L),
		    name(X,[A]),
		    name(Sym,[A,A])
		),
		T).


%% FIN

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
