/*----------------------------------------------------------------------
expansion.pl

http://www.ling.ohio-state.edu/~nvail/msol.html

Macro and abstract expansion code for msol

(c) Nathan Vaillette (nvail@ling.ohio-state.edu)

Last updated July 10, 01

Tested with the Finite State Utilities v. 6.234

-------------------------------------------------------------------------*/
:- multifile msol_macro/2 .

%% GvN: 
%% for stand-alone binaries these should not be re-loaded - they're in
%% the system already anyway...
%% :- use_module(library(lists)).
%% predicates from library(lists) are prefixed with lists: below

expand_msol_macros(regex(R),regex(R)) :- !.

expand_msol_macros(F0,F) :-

	%% F0 matches a macro template
	msol_macro(F0,_),
	!,

	
	free_vars(F0,VF0,_),

	%% wrap template variables in temp(...)
	temp_vars(VF0,F0,F1),

	%% instantiate the macro with the temp vars
	msol_macro(F1,F2),

	%% find all the other non-temp vars in F2
	all_vars(F2,VF2A),
	sort(VF2A,VF2),
	fresh_vars(VF2,Fresh),

	%% rename them
	rename_vars(Fresh,F2,F3),

	%% strip off temp(...)
	detemp_vars(VF0,F3,F4),
	
	%% recursive clause.
	expand_msol_macros(F4,F).

expand_msol_macros(F,F) :-
	    atomic(F),
	    !.

expand_msol_macros(F0,F) :-
	F0 =.. [Functor|Args],
	expand_msol_macros_rec(Args,Args1),
	F =.. [Functor|Args1].

expand_msol_macros_rec([],[]).
expand_msol_macros_rec([Form0|Forms0],[Form|Forms]) :-
	expand_msol_macros(Form0,Form),
	expand_msol_macros_rec(Forms0,Forms).

%% free_vars lists free variables:

free_vars(F,Vars,N) :-
	free_vars(F,Vars0),
	sort(Vars0,Vars),
	length(Vars,N).

%% ignore regular expressions
free_vars(regex(_),[]) :- !.

%% variables:
free_vars(v(V),[V]) :- !.

%% quantifiers:
free_vars(exists(v(V),F),Vars) :-
	!,
	free_vars(F,Vars0),
	remove_elt(Vars0,V,Vars).
free_vars(all(v(V),F),Vars) :-
	!,
	free_vars(F,Vars0),
	remove_elt(Vars0,V,Vars).

%% connectives and relation symbols
free_vars(F,Vars) :-
	F =.. [_Functor|Args],
	free_vars_rec(Args,[],Vars).

free_vars_rec([],Vars,Vars).
free_vars_rec([F|Fs],Vars0,Vars) :-
	free_vars(F,Vars1),
	lists:append(Vars0,Vars1,Vars2),
	free_vars_rec(Fs,Vars2,Vars).

%% removes all occurances of elt. from list
remove_elt([],_,[]).
remove_elt([X|L0],X,L) :-
        !,
	remove_elt(L0,X,L).
remove_elt([H|L0],X,[H|L]) :-
	remove_elt(L0,X,L).



%% temp_vars replaces Vars with temp(int)
 
temp_vars([],F,F).
temp_vars([V|Vars],F0,F) :-
	replace_in_formula(v(V),v(temp(V)),F0,F1),
	temp_vars(Vars,F1,F).
	
%% detemp_vars removes temp(...)

detemp_vars([],F,F).
detemp_vars([V|Vars],F0,F) :-
	replace_in_formula(v(temp(V)),v(V),F0,F1),
	detemp_vars(Vars,F1,F).

replace_in_formula(X,Y, X,Y) :- !.

replace_in_formula(X,Y, Form0, Form1) :-
   (   atomic(Form0)
   ->
       Form1 = Form0
   ;
       (Form0 =.. List0,
        replace_in_formula_rec(X,Y, List0, List1),
        Form1 =.. List1)
   ).

replace_in_formula_rec(_,_,[],[]).
replace_in_formula_rec(X,Y,[Form0|Forms0], [Form1|Forms1]):-
   replace_in_formula(X,Y,Form0,Form1),
   replace_in_formula_rec(X,Y,Forms0,Forms1).



%% all_vars collects all vars, ignoring temp(var)'s

%% ingnore temporarily renamed variables 
all_vars(v(temp(_)),[]) :- !.

all_vars(v(V),[V]) :- !.
	

%% quantifiers:
all_vars(exists(v(V),F),[V|Vars]) :-
	!,
	all_vars(F,Vars).
all_vars(all(v(V),F),[V|Vars]) :-
	!,
	all_vars(F,Vars).


%% connectives and relation symbols
all_vars(F,Vars) :-
	F =.. [_Functor|Args],
	all_vars_rec(Args,[],Vars).

all_vars_rec([],Vars,Vars).
all_vars_rec([F|Fs],Vars0,Vars) :-
	all_vars(F,Vars1),
	lists:append(Vars0,Vars1,Vars2),
	all_vars_rec(Fs,Vars2,Vars).

%% fresh_vars associates list of variables with fresh replacements

fresh_vars([],[]).
fresh_vars([Var|Vars],[Var-FVar|Fresh]) :-
	fresh_var(FVar),
	fresh_vars(Vars,Fresh).

:- dynamic fresh_var_counter/1.
fresh_var_counter(1).
fresh_var(fresh_var(M)) :-
      fresh_var_counter(N),
      retract( fresh_var_counter(N) ),
      M is +(N,1),
      assert( fresh_var_counter(M) ).


%% renames variables:
rename_vars([],F,F).
rename_vars([Var-FVar|Fresh],F0,F) :-
	replace_in_formula(v(Var),v(FVar),F0,F1),
	rename_vars(Fresh,F1,F).


msol_macro(beta_reduce(lambda(l(Param),Form0),Repl0),Form) :-
	expand_msol_macros(Form0,Form1),
	expand_msol_macros(Repl0,Repl1),
	temp_free_params([],Form1,Form2),
	replace_in_formula(l(temp(Param)),Repl,Form2,Form).


%% temp_free_params replaces free lambda params l(L) with l(temp(L))

temp_free_params(_Bound,regex(L),regex(L)) :- !.
temp_free_params(_Bound,v(L),v(L)) :- !.

temp_free_params(Bound,l(L),l(L)) :-
	lists:member(L,Bound),
	!.
temp_free_params(_Bound,l(L),l(temp(L))) :-
	!.

temp_free_params(Bound,lambda(l(L),Form0),lambda(l(L),Form)) :-
	!,
	temp_free_params([L|Bound],Form0,Form).

%% connectives and relation symbols
temp_free_params(Bound,Form0,Form) :-
	Form0 =.. [Functor|Args0],
	temp_free_params_rec(Bound,Args0,Args),
	Form =.. [Functor|Args].

temp_free_params_rec(_Bound,[],[]).
temp_free_params_rec(Bound,[Arg0|Args0],[Arg|Args]) :-
	temp_free_params(Bound,Arg0,Arg),
	temp_free_params_rec(Bound,Args0,Args).



%% temp_bound_vars replaces bound var v(L) with v(temp(L))

temp_bound_vars(_Bound,regex(L),regex(L)) :- !.
temp_bound_vars(_Bound,l(L),l(L)) :- !.

temp_bound_vars(Bound,v(L),v(temp(L))) :-
	lists:member(L,Bound),
	!.
temp_bound_vars(_Bound,v(L),v(L)) :-
	!.

temp_bound_vars(Bound,exists(v(L),Form0),exists(v(L),Form)) :-
	!,
	temp_bound_vars([L|Bound],Form0,Form).

temp_bound_vars(Bound,all(v(L),Form0),all(v(L),Form)) :-
	!,
	temp_bound_vars([L|Bound],Form0,Form).

%% connectives and relation symbols
temp_bound_vars(Bound,Form0,Form) :-
	Form0 =.. [Functor|Args0],
	temp_bound_vars_rec(Bound,Args0,Args),
	Form =.. [Functor|Args].

temp_bound_vars_rec(_Bound,[],[]).
temp_bound_vars_rec(Bound,[Arg0|Args0],[Arg|Args]) :-
	temp_bound_vars(Bound,Arg0,Arg),
	temp_bound_vars_rec(Bound,Args0,Args).


%% tests for variable renaming:

msol_macro(foo(v(X)),
           and(
               matches(v(X),a),
               exists(v(y),and(
                               not(empty(v(y))),
			       bar(v(X),v(y)))))).


msol_macro(testbeta,
	   exists(v(x),
		  beta_reduce(lambda(l(l),bar(l(l))),
			      v(x)))
	  ).

msol_macro(bar(v(V)),exists(v(x),foo(v(V),v(x)))).


