%%% Gertjan van Noord, Dale Gerdemann, (c) 1998
%%%
%%% Mehryar Mohri and Richard Sproat, An Efficient Compiler for Weighted
%%% Rewrite Rules. 34th Annual Meeting of the ACL, Santa Cruz 1996, pages
%%% 230-238. This only treats the non-weighted case.
%%%
%%% this version marks symbols such that there cannot be confusion
%%% between `input' symbols and (temporary) `marker' symbols
%%%
%%% moreover, it `solves' the sloppy_ignore problem:
%%% First, Dale explained why ignore cannot be applied to transducers:
% > 
% > Second question: isn't it a bit dangerous to use the recognizer
% > operation 'ignore' on a transducer? I think it works in this case
% > since the pairs you're ignoring all have epsilon outputs. But suppose
% > the pair is a:b and the transduction is identity(c). The problem is
% > that there's more than one machine that implements this transduction.
% > 
% > The simple version is simply
% > 
% > Initial: 0
% > Final: 1
% > 
% > 0 --c:c--> 1
% > 
% > A more complex version is.
% > 
% > Initial: 0
% > Final: 2
% > 
% > 0 --c:epsilon--> 1 ---epsilon:c--> 2
% > 
% > 
% > In the first case you get transductions like:
% > 
% > a c a
% > b c b
% > 
% > In the second case you get:
% > 
% > a c a
% > b b c
% > 
% > 
% > The example is a bit artificial, but I think it is generally not good
% > for the user to be concerned with the actual machine that implements a
% > transduction. 
% >
% >
%%% And then Dale explained how the use of ignore to a transduer
%%% in Mohri & Sproat can be removed:
% >
% > If T is a filter transducer, you can ignore instances of E by
% > 
% > macro(ignore_in_filter(T,E), identity(ignore(domain(T),E)))
% > 
% > The type 2 transducer, however, is not purely a filter since it
% > also deletes the marker it is looking for. You could fix this to
% > make it purely a filter by making the loop over #:#. Then use
% > ignore_in_filter and compose it with a transducer that throws away #.
% > 
% > It looks to me like Mohri & Sproat were glossing over a few nasty
% > details.

:- multifile macro/2.
:- multifile rx/2.

:- discontiguous mark/4.


%%% replace(Left,Phi,Psi,Right) === Phi -> Psi /Left _ Right
%%% i.e. the left-to-right obligatory replacement rule
macro(replace(Phi0,Psi,Left0,Right), pragma([Phi-Phi0,Left-Left0],
     non_markers
         o
      r(Right)
         o
      f(Phi)
         o
   replace(Phi,Psi)
         o
     l1(Left)
         o
      l2(Left)
         o
   inverse(non_markers)
     )).

%% Use this test with the string abababa
%% and (default) symbol separator 0.
macro(test,replace([a,b],x,[a,b],a)).

%% Use this test with the string <1 <2 <1 <2 <1 <2 <1
%% and using symbol_separator 32.
macro(test2,replace(['<1','<2'],>,['<1','<2'],'<1')).

macro(bool,{0,1}).

macro(rbrack,[>,1]).

macro(lbrack1,['<1',1]).

macro(lbrack2,['<2',1]).

macro(ext_sigma,[?,bool]).

macro(brack,{rbrack,lbrack1,lbrack2}).

macro(sigma,ext_sigma - brack).

macro(non_markers, [identity(?),[] x 0]*).

macro(non_markers(Expr), 
      range(identity(Expr) o [identity(?),[] x 0]*)).

macro(xignore(E1,E20),pragma([E2-E20], ignore(E1,E2) - [E2,? *])).

macro(r(R),
      reverse(marker(1,
		     reverse([non_markers(R),sigma*]),
                     [seq(1,>)]
		    )
	     )
     ).

macro(f(F),
      reverse(marker(1,
                     reverse([non_markers(xignore(F,rbrack)),
                              rbrack,
                              {sigma,rbrack}*]),
		     [seq(1,'<1'),seq(1,'<2')]
		    )
	     )
     ).

macro(l1(L), ignore(marker(2,
                           [sigma*,non_markers(L)],
                           seq('<1',1)
                          ),
                    lbrack2)
                      o
          [[{sigma,lbrack2}*,lbrack1 x []]*,{sigma,lbrack2}*]
     ).

macro(l2(L),marker(3,
		   [sigma*,non_markers(L)],
		   seq('<2',1)
		  )
     ). 

macro(replace(Phi,Psi), 
      {{sigma,lbrack2, rbrack x []}, 
       [lbrack1,
	ignore(non_markers(Phi), brack) x non_markers(Psi),
	rbrack x []
       ]
      }*).

%%%%%%%%%%%%%%%%
%%% MARKER/3 %%%
%%%%%%%%%%%%%%%%

%% we only implement the markers that we really use, so not in full
%% generality (since I am lazy).
%% marker(1,RegEx,List)
%%                List is a list of symbol(pairs) to be inserted
%% marker(2,RegEx,Sym)
%%                Sym is a symbol(pair) to be deleted
%% marker(2,RegEx,Sym)
%%                Sym is a symbol(pair) to be deleted

%% DG: I've made marker2 a special case which outputs a recognizer
%% rather than a transducer. It would be more consistent to do this
%% with marker3 as well.

rx(marker(2,Expr,Change),Fa) :- !,
    fsa_regex:rx(determinize(Expr),Fa0),
    mark(2,Change,Fa0,Fa).

rx(marker(Type,Expr,Change),Fa) :-
    fsa_regex:rx(identity(determinize(Expr)),Fa0),
    mark(Type,Change,Fa0,Fa).

%%%%%%%%%%%%%%
%%% type 1 %%%
%%%%%%%%%%%%%%

%% NB: for type 1 we only need insertions ...
mark(1,Insertions0,Fa1,Fa) :-
    fsa_data:fsa_symbols_decl(Fa1,t(ModuleT,Module)),
    regex_notation_to_predicate_list(Insertions0,Insertions,Module),
    fsa_data:fsa_start_states(Fa1,Starts),
    fsa_data:fsa_transitions(Fa1,Trans0),
    fsa_data:fsa_final_states(Fa1,Finals),
    fsa_data:fsa_states_set(Fa1,AllStates),
    ordsets:ord_subtract(AllStates,Finals,NewFinals0),
    add_insertions(Finals,Insertions,NewFinals,NewFinals0,Trans,Trans1),
    replace_transitions_starting_final(Trans0,Trans1,Fa1),
    fsa_data:fsa_construct_rename_states(t(ModuleT,Module),Starts,NewFinals,Trans,[],Fa).

%% this is slow for large automata
%% in such cases it's worthwhile creating a thing for final_states
%% moreover it creates symbolic names first, which are renamed later
%% it would be better to create larger integers for state names here
%% immediately
replace_transitions_starting_final([],[],_).
replace_transitions_starting_final([trans(A0,B,C)|T0],[trans(A,B,C)|T],Fa):-
    (	fsa_data:fsa_final_state(Fa,A0)
    ->	A=q(A0)
    ;	A=A0
    ),
    replace_transitions_starting_final(T0,T,Fa).

add_insertions([],_,F,F) --> [].
add_insertions([F0|Fs],Ins,[q(F0)|NewF0],NewF) -->
    add_ins(Ins,F0),
    add_insertions(Fs,Ins,NewF0,NewF).

add_ins([],_F) --> [].
%% Special case added: a marker can be a sequence of two symbols
add_ins([seq(Sym1,Sym2)|Syms],F) -->
    {!},
    [trans(F,[]/Sym1,foo),trans(foo,[]/Sym2,q(F))],
    add_ins(Syms,F).
add_ins([Sym|Syms],F) -->
	{nl,nl,nl},

    [trans(F,[]/Sym,q(F))],
    add_ins(Syms,F).

%%%%%%%%%%%%%%
%%% type 2 %%%
%%%%%%%%%%%%%%

%% macro(l1(L), ignore(marker(2,[sigma*,non_markers(L)],seq('<1',1)),'<2')).

%% NB: for type 2 and 3 we only need a single deletion
%% DG: Filter changed here to filter.
mark(2,seq(A0,B0),Fa1,Fa) :-
    fsa_data:fsa_symbols_decl(Fa1,r(Module)),
    Module:regex_notation_to_predicate(A0,A),
    Module:regex_notation_to_predicate(B0,B),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa2,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa2,Fa3,Finals,AllStates),
    fsa_data:fsa_copy_except(states,Fa3,Fa,Num0,Num),
    fsa_data:fsa_states_set(Fa1,AllStates),
    add_2_filters(Finals,seq(A,B),Num0,Num,Trans1,Trans0),
    sort(Trans1,Trans).

add_2_filters([],_,Num,Num) --> [].
add_2_filters([Final|Fs],seq(A,B),Num0,Num) -->
    {user:is(Num1,+(Num0,1))},
    [trans(Final,A,Num0),trans(Num0,B,Final)],
    add_2_filters(Fs,seq(A,B),Num1,Num).

%%%%%%%%%%%%%%
%%% type 3 %%%
%%%%%%%%%%%%%%

%% NB: for type 2 and 3 we only need a single filter
mark(3,seq(A0,B0),Fa1,Fa) :-
    fsa_data:fsa_symbols_decl(Fa1,t(Module,_)),
    Module:regex_notation_to_predicate(A0,A),
    Module:regex_notation_to_predicate(B0,B),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa2,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa2,Fa3,Finals,AllStates),
    fsa_data:fsa_copy_except(states,Fa3,Fa,Num0,Num),
    fsa_data:fsa_states_set(Fa1,AllStates),
    ordsets:ord_subtract(AllStates,Finals,NonFinals),
    add_2_deletions(NonFinals,seq(A,B),Num0,Num,Trans1,Trans0),
    sort(Trans1,Trans).

add_2_deletions([],_,Num,Num) --> [].
add_2_deletions([Final|Fs],seq(A,B),Num0,Num) -->
    {user:is(Num1,+(Num0,1))},
      [trans(Final,A/[],Num0),trans(Num0,B/[],Final)],
    add_2_deletions(Fs,seq(A,B),Num1,Num).



regex_notation_to_predicate_list([],[],_).
regex_notation_to_predicate_list([H0|T0],[H|T],Mod):-
    (	H0=seq(A0,B0)
    ->	H=seq(A,B),
	Mod:regex_notation_to_predicate(A0,A),
	Mod:regex_notation_to_predicate(B0,B)
    ;	Mod:regex_notation_to_predicate(H0,H)
    ),
    regex_notation_to_predicate_list(T0,T,Mod).

