:- module( bit_types, []).

:- use_module(bitsets).

:- multifile user:term_expansion/2.

user:term_expansion(hierarchy(Type,Ds),Preds):-
    expand_hierarchy(Type,Ds,Preds).

expand_hierarchy(Type,Ds,Preds):-
    collect_subtypes(Ds,Type,0,_,Preds0,[]),
    sort(Preds0,Preds).

collect_subtypes([],_,N,N) -->
    [].
collect_subtypes([H|T],Type,N0,N) -->
    [subtype(Type,Sub)],
    {type_of(H,Sub,Ds,R)},
    collect_subtypes0(Ds,Sub,N0,N1,R),
    collect_subtypes(T,Type,N1,N).

leaf(r,_,N,N) --> [].
leaf(nr,Type,N0,N) -->
    [leaf(Type,N0)],
    { N is +(N0,1) }.

collect_subtypes0([],Type,N0,N,R) -->
    leaf(R,Type,N0,N).
collect_subtypes0([H|T],Type,N0,N,_) -->
    [subtype(Type,Sub)],
    {type_of(H,Sub,Ds,R)},
    collect_subtypes0(Ds,Sub,N0,N1,R),
    collect_subtypes(T,Type,N1,N).

type_of(t(T,Ds),T,Ds,nr).
type_of(+T,T,[],r).

%% end of utilities for term expansion. Now define an example hierarchy:
%% this is automatically term-expanded into clauses
%% true/1
%% type/1
%% subtype/2
%%
%% hierarchy is given as a list t(Type,ListOfTypes)
%% multiple inheritance is indicated using a `+' prefix for every
%% second occurrence of that type
hierarchy(
 top,[t(red,   [t(bright_red,[]),
		t(ordinary_red,[]),
	        t(dark_red,[])
	       ]),
      t(white, [t(bright_white,[]),
	        t(dark_white,[])
	       ]),
      t(blue,  [t(bright_blue,[]),
	        t(dark_blue,[])
	       ]),
      t(bright,[+bright_red,
		+bright_white,
		+bright_blue
	       ]),
      t(dark,  [+dark_red,
		+dark_white,
		+dark_blue
	       ])
     ]).

%% end of hierarchy. Now follow definitions which hold for any type
%% hierarchy.

true(L) :-
    findall(X,leaf(_,X),Xs),
    bitsets:list_to_bitset(Xs,L).

conjunction(A,B,C) :-
    bitsets:bit_intersection(A,B,C),
    non_empty(C).

non_empty(C):-
    C =\= 0.

disjunction(A,B,C) :-
    bitsets:bit_union(A,B,C).

negation(P0,P) :-
    true(P1),
    bitsets:bit_subtract(P1,P0,P).


regex_notation_to_predicate(Type,L) :-
    findall(X,subsumes_leaf(Type,X),Xs),
    bitsets:list_to_bitset(Xs,L).

subsumes_leaf(X,N):-
    leaf(X,N).
subsumes_leaf(X,N):-
    subtype(X,Y),
    subsumes_leaf(Y,N).

evaluate_predicate(Bitset,Leaf) :-
    bitsets:bitset_to_list(Bitset,List),
    lists:member(N,List),
    leaf(Leaf,N).

talks_about(Bitset,Leaf) :-
    bitsets:bitset_to_list(Bitset,List),
    lists:member(N,List),
    leaf(Leaf,N).

ns_to_leafs([],[]).
ns_to_leafs([H0|T0],[H|T]) :-
    leaf(H,H0),
    ns_to_leafs(T0,T).

display_predicate(X,Conj) :-
    bitsets:bitset_to_list(X,List0),
    ns_to_leafs(List0,List),
    fsa_visualization:abbreviate_label0(List,[H|T]),
    display_predicate_in(T,H,Conj).

display_predicate_in([],H,H).
display_predicate_in([H|T],A,{Conj}):-
    list_to_conj([H|T],A,Conj).

list_to_conj([],A,A).
list_to_conj([H|T],A,(A,B)):-
    list_to_conj(T,H,B).

prepare_complement_of_set(_,_).

complement_of_set(Preds,_,Comps):-
    true(True),
    (	conjoin_negated(Preds,True,Comp)
    ->	Comps=[Comp]
    ;   Comps=[]
    ).

conjoin_negated([],Pred,Pred).
conjoin_negated([P0|Preds],P1,Pred) :-
    negation(P0,NotP0),
    conjunction(P1,NotP0,P2),
    conjoin_negated(Preds,P2,Pred).

determinize_preds(U0,U):-
    fsa_determinizer:determinize_preds(U0,U,bit_types).
    
t_determinize_preds(U0,U):-
    fsa_t_determinizer:t_determinize_preds(U0,U,bit_types).
    
identity(Pred,'$@'(Pred)).

class_to_pred(List0,[Pred]):-
    sort(List0,List),
    del_func(in(List),Pred).

negated_class_to_pred(List0,[not_in(List)]):-
    sort(List0,List).

cleanup([],[]).
cleanup([H|T],[Pred]) :-
    cleanup(T,H,Pred).

cleanup([],P,P).
cleanup([H|T],Pred0,Pred) :-
    disjunction(H,Pred0,Pred1),
    cleanup(T,Pred1,Pred).
