:- module( types, []).

:- multifile user:term_expansion/2.

user:term_expansion(hierarchy(Type,Ds),Preds):-
    expand_hierarchy(Type,Ds,Preds).

expand_hierarchy(Type,Ds,[true(Type)|Preds]):-
    collect_subtypes(Ds,Type,Preds0,[]),
    sort(Preds0,Preds).

collect_subtypes([],Type) -->
    [type(Type)].
collect_subtypes([H|T],Type) -->
    [subtype(Type,Sub)],
    {type_of(H,Sub,Ds)},
    collect_subtypes0(Ds,Sub),
    collect_subtypes(T,Type).

collect_subtypes0([],Type) -->
    [type(Type), leaf(Type)].
collect_subtypes0([H|T],Type) -->
    [subtype(Type,Sub)],
    {type_of(H,Sub,Ds)},
    collect_subtypes0(Ds,Sub),
    collect_subtypes(T,Type).

type_of(t(T,Ds),T,Ds).
type_of(+T,T,[]).

%% end of utilities for term expansion. Now define an example hierarchy:
%% this is automatically term-expanded into clauses
%% true/1
%% type/1
%% subtype/2
%%
%% hierarchy is given as a list t(Type,ListOfTypes)
%% multiple inheritance is indicated using a `+' prefix for every
%% second occurrence of that type
hierarchy(
 top,[t(red,   [t(bright_red,[]),
		t(ordinary_red,[]),
	        t(dark_red,[])
	       ]),
      t(white, [t(bright_white,[]),
	        t(dark_white,[])
	       ]),
      t(blue,  [t(bright_blue,[]),
	        t(dark_blue,[])
	       ]),
      t(bright,[+bright_red,
		+bright_white,
		+bright_blue
	       ]),
      t(dark,  [+dark_red,
		+dark_white,
		+dark_blue
	       ])
     ]).

%% end of hierarchy. Now follow definitions which hold for any type
%% hierarchy.

%% subsumption is straightforward.
subsumes(A,A).
subsumes(A,C) :-
    subtype(A,B),
    subsumes(B,C).

proper_subsumes(A,C):-
    subtype(A,B),
    subsumes(B,C).

%% conjunction(A,B,C) == unification of types A and B
%% naive algorithm:
%% find all types subsumed by both A and B, and select maximal one.
%% compensate for using the naive algorithm by using memo-ization

unknown_predicate_handler(conj(_,_,_,_),fail).

conjunction(A,B,C) :-
    terms:term_hash(conj(A,B),-1,16'2000000,Index),
    (	conj(Index,A,B,List)
    ->	true
    ;   findall(C,compute_conjunction(A,B,C),List),
	assertz(conj(Index,A,B,List))
    ),
    lists:member(C,List).

compute_conjunction(A,B,C) :-
    findall(C,( subsumes(A,C),
		subsumes(B,C)), List0),
    sort(List0,List),
    lists:member(C,List),
    \+ ( lists:member(D,List),
	 proper_subsumes(D,C)
       ).

%% conjoin_with_negation(A,B,C):-
%%    negation(B,NotB),
%%    conjunction(A,NotB,C).
%% This tries to be smarter and ensures that C is maximal;
%% C will not be unique though (in the general case).
%% could be precompiled/memo-ized.
conjoin_with_negation(In,A,C):-
    findall(T,( type(B),
		\+ conjunction(A,B,_),
		conjunction(In,B,T)
	      ), Types0 ),
    sort(Types0,Types),
    lists:member(C,Types),
    \+ (  lists:member(D,Types),
	  proper_subsumes(D,C)
       ).
    
regex_notation_to_predicate(Atom,Atom).

evaluate_predicate(Type,Sub) :-
    leaf(Sub),
    subsumes(Type,Sub).

talks_about(Type,Sub) :-
    leaf(Sub),
    subsumes(Type,Sub).

display_predicate(X,X).

prepare_complement_of_set(_,_).

%% Comps is the list of types inconsistent with any of the types
%% in Preds; The shorter the length of Comps, the better.
complement_of_set(Preds,_,Comps):-
    true(True),
    findall(Comp,conjoin_negated(Preds,True,Comp),Comps).

conjoin_negated([],Pred,Pred).
conjoin_negated([P0|Preds],P1,Pred) :-
    conjoin_with_negation(P1,P0,P2),
    conjoin_negated(Preds,P2,Pred).

%% TODO......... is broken
%% Dale's brute force algorithm, extended to deal with $@; but first
%% we pick out the types that have no overlap with any of the others.
%% In many cases this yields reasonable results; but it's clearly
%% not perfect.
determinize_preds(Preds0,Preds) :-
    non_overlapping(Preds0,Preds1),
    sort(Preds1,Preds).

non_overlapping(Preds0,Preds):-
    (	non_overlap_one(Preds0,Preds1,P)
    ->	Preds=[P|Preds2],
	non_overlapping(Preds1,Preds2)
    ;   expand(Preds0,Preds)
    ).

expand([],[]).
expand([T-States|Types],List):-
    findall(Type-States,expand_1(T,Type),List,List0),
    expand(Types,List0).

expand_1(T,Type):-
    leaf(Type),
    subsumes(T,Type).

t_determinize_preds(_Preds0,_Preds) :-
    throw(fsa_error(
        "pred_module types: t_determinize_preds not supported~n",[])).

non_overlap_one(Preds0,Preds,Pred-List) :-
    lists:select(Pred-List,Preds0,Preds),
    no_conjunction(Preds,Pred).

no_conjunction([],_).
no_conjunction([Pred-_|Tail],Pred0):-
    \+ conjunction(Pred0,Pred,_),
    no_conjunction(Tail,Pred0).

identity(Pred,'$@'(Pred)).

class_to_pred(L,L).

negated_class_to_pred(L0,_,L):-
    complement_of_set(L0,L).

%% cleanup
%% List0 is a list of types, interpreted as disjunction
%% List is equivalent, but shorter list of types.
cleanup(List0,List):-
    %% first, find all types C such that C is not true of
    %% any species not in List0.
    findall(C, smaller_type(C,List0), Cs0),  sort(Cs0,Cs),
    %% second, now select from C only the minimal ones.
    %% there could still be overlap in the result, of course.
    findall(C, select_maximal(C,Cs), List).

%% C is a type which is subsumed by List (interpreted as a disjunction)
%% i.e. there are no species which are true for C but not for List
smaller_type(C,List) :-
    type(C),
    \+ other_species(C,List).

%% other_species(C,List)
%% C subsumes a type which is not subsumed by any of the types in List
other_species(C,List) :-
    leaf(D),
    subsumes(C,D),
    \+ subsumes_list(List,D).

%% subsumes_list(List,El)
%% there is a type in List which subsumes El
subsumes_list(List,El) :-
    lists:member(El0,List),
    subsumes(El0,El).

%% select_maximal(C,List)
%% select a type from List such that there is no more genaral type in List
select_maximal(C,List) :-
    lists:select(C,List,List1),
    \+ more_general_member(C,List1).

%% more_general_member(C,List)
%% List contains a type which is strictly more general than C
more_general_member(C,List) :-
    lists:member(D,List),
    proper_subsumes(D,C).
