%% by Dale Gerdemann dg@sfs.nphil.uni-tuebingen.de
%% -----------------------------------------------
%%
%% BTW: I've got a little piece of code here for your amusement (Maybe
%% you can guess that it's between semesters here). Anyway I've tried to
%% solve the 4 queens problem with the finite state utilities. As far as
%% I can tell, I've done it right. But I can't quite test the whole thing
%% at once. It just gets too complicated. Anyway, here it is:
%% I was a bit interested in trying this problem since the algorithms for
%% compiling rewrite rules also have this generate-and-test flavor. For
%% example, Kartunnen's "directed replacement" uses a transducer to
%% generate possible markups of a string and then filters out those
%% markups that don't correspond to a longest match. To implement such an
%% algorithm, I think it would help to have a general understanding of
%% how generate-and-test can be implemented as efficiently as possible.

%% Minor changes by GvN. FSA is fast enough now for such problems.

:- multifile macro/2.

macro(four_queens_problem,
        board
          &
      four_queens
          &
      safe_rows
          &
       l_diag1
          &
       r_diag5
          &
       l_diag5
          &
       r_diag1
          &
       l_diag2
          &
       l_diag4
          &
       r_diag2
          &
       r_diag4
          &
       l_diag3
          &
       r_diag3
          &
     safe_columns ). 


%% bool:
%% -----
%% f for 'no queen', t for queen

macro(bool,{f,t}).

%% Make a 4x4 board: Even this small size is pretty hard.

macro(board,[bool,bool,bool,bool,
             bool,bool,bool,bool,
             bool,bool,bool,bool,
             bool,bool,bool,bool]).

macro(four_queens,[f*,t,f*,t,f*,t,f*,t,f*]).

macro(safe_rows,[safe_row,safe_row,safe_row,safe_row]).

macro(safe_row,
     [bool,bool,bool,bool]
              &
            q_chk ).

%%% q_chk:
%%% -----
%%% A little macro to check that exactly one queen exists in some sequence.

macro(q_chk,[f*,bool,f*]).

%% safe_columns:
%% -------------
%% Note: This would perhaps be easier if one could write a transducer
%% to transpose the chess board and then check safe_rows.
%% But doing transposition with a transducer is easier said than done.
%% This could also be written in a way analogous to the checking of 
%% diagonals.

macro(safe_columns,
   ~[bool*,t,[bool,bool,bool,bool]*,bool,bool,bool,t,bool*]).

%%% safe_left_diags/safe_right_diags:
%%% ---------------------------------
%%% The following code looks like a pretty redundant way of checking
%%% left and right diagonals. But every other way I tried goes off
%%% the deep end. Of course, this does too, but at least you can check
%%% one diagonal at a time.

macro(l_diag1,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([1:2,2:1],Diag).

macro(l_diag2,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([1:3,2:2,3:1],Diag).

macro(l_diag3,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([1:4,2:3,3:2,4:1],Diag).

macro(l_diag4,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([2:4,3:3,4:2],Diag).

macro(l_diag5,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([3:4,4:3],Diag).


macro(r_diag1,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([3:1,4:2],Diag).

macro(r_diag2,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([2:1,3:2,4:3],Diag).

macro(r_diag3,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([1:1,2:2,3:3,4:4],Diag).

macro(r_diag4,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([1:2,2:3,3:4],Diag).

macro(r_diag5,
   domain(           Diag
      
                       o
                 identity(q_chk)
      )):-  diag([1:3,2:4],Diag).

%%% diag(ListOfSquares,Regex)
%%% ----------------
%%% Prolog procedure to construct a transducer which picks out
%%% the squares listed in ListOfSquares. Each square is encoded
%%% as 'Row:Column'. The squares listed in ListOfSquares must come
%%% in ascending order: Row 1 before Row 2 etc, and within a given
%%% row, column 1 before column 2 etc.

diag(L0,Regex) :-
   multiply_out_rows(L0,L),
   pick_out(L,1,Regex).


multiply_out_rows([],[]).
multiply_out_rows([Row:Column|R0],[N|R]) :-
    N is +(*(4,(Row-1)),Column),
    multiply_out_rows(R0,R).


pick_out([],_,[kleene_star(bool x [])]).
pick_out([N|R0],N,[identity(bool)|R]):-
    !,
    N1 is +(N,1),
    pick_out(R0,N1,R).
pick_out(L,N,[bool x []|R]) :-
    N1 is +(N,1),
    pick_out(L,N1,R).

