%% (c) 	Carola Fricke
%% fricke@sfs.nphil.uni-tuebingen.de 
%% http://www.sfs.nphil.uni-tuebingen.de/~fricke/

%% Problems with the use of the program could arise from the following facts:
%%
%% 1. Most of the algorithms I use here work with one start state only, while Gertjan van
%%    Noord's package admits for more that one start state.
%%
%% 2. For I didn't know what the last argument in van Noord's fa-datastructure is for,
%%    I just ignored it while implementing the algorithms.
%%
%% 3. I did't take transducers with jumps into account. If one wanted to work with transducers
%%    of this kind, one would have to change almost all the algorithms, I guess.



:- ensure_loaded(library(lists)).




%% The program's main procedure is 'transduce', which takes a given transducer and an input 
%% string as an argument, checks if it is functional or subsequential (via 'decision')
%% and produces an output string in the following way:
%% If the given transducer is not functional it computes an intersection of the input string
%% with the transducer, building a new transducer to process the input string on. However,
%% this intersection is not implemented here!! Anyway, the necessary procedures already
%% exists in Gertjan van Noord's package.
%% If the tranducer is functional and subsequential, the program just transduces the input in
%% the traditional way.
%% In the last case, if it is functional but not subsequential, a bimachine is built to do 
%% the computation of the output in a deterministic way.



%% transduce(fa([b,a],_,[0],[0],[trans(0,a/[c],0),trans(0,a/[c,b],0)],[],_),[a,a],Output).

%% transduce(fa([a,b],_,[0],[2],[trans(0,a/[b],1),trans(0,a/[b,c],3),trans(0,a/[b,c,d],2),trans(1,b/[c,d],2),trans(3,b/[d],2),trans(2,b/[],2)],[],_),[a,b],Output).

%% transduce( fa([a,keep,under,control,out,of,reach],7,[0],[3],[trans(0,keep/[keep_1],1),trans(0,keep/[keep_2],4),trans(1,a/[a],1),trans(4,a/[a],4),trans(1,under/[under],2),trans(2,control/[control],3),trans(4,out/[out],5),trans(5,of/[of],6),trans(6,reach/[reach],3)],_,_),[keep,out,of,reach],Output).

%% transduce(fa([c,o,plus,min,a,o],6,[0],[4],[trans(0,c/[c],1),trans(1,o/[o],2),trans(2,plus/[],3),trans(2,plus/[min],5),trans(3,o/[o],4),trans(5,o/[o],4)],[],[]) ,[c,o,plus,o],Output).

%% transduce(fa([b,a],4,[0],[3],[trans(0,b/[z],1),trans(1,a/[x,x],1),trans(1,b/[z],3),trans(1,a/[x],2),trans(2,a/[],1)],[],_),[b,a,a,b],Output).





transduce(UnknownTransducerType,InString,OutString):-
    decision(UnknownTransducerType,KnownTransducerType),!,
    ( KnownTransducerType = fa(_,_,_,_,_,_,_,-)
     -> intersect_with_domain(InString,KnownTransducerType,OutString)
     ; (KnownTransducerType = fa(_,_,_,_,_,_,_,+,+)
	-> run_transducer(InString,KnownTransducerType,OutString)
	; run_bimachine(InString,KnownTransducerType,OutString))).








%% 'decision' checks if a transducer is functional or subsequential and and finally marks
%% the input-transducer respectively to the result of the checks - a '-' means that
%% the transducer is not functional, two times '+' means it is functional and subsequential,
%% and in the last case a bimachine is built which is a triple of two deterministic
%% automata and a transition function delta.

decision(NonFun,NonFun):-
   NonFun == fa(_,_,_,_,_,_,_,-), 
   !.
decision(Bimachine,Bimachine):-
   Bimachine == (_,_,_),
   !.
decision(Funct_and_Subsequential,Funct_and_Subsequential):-
   Funct_and_Subsequential == fa(_,_,_,_,_,_,_,+,+),
   !.
decision(FST,FSTout):-
    unambiguous(FST,FST1),
    is_function(FST,FST1,Bool),
    (Bool == false
     -> mark_as_non_fun(FST,FSTout),
     write('I`m sorry, Your Transducer is not functional.'),nl
     ; write('Congratulations, your Transducer is functional,'),
     is_subsequential(FST1,Bool1),     
     (Bool1 == false
      -> write(' but it is not subsequential.'),nl,
      write('Therefor we`ll build a nice bimachine:'),nl,nl,
      bimachine(FST,A1,A2,Delta),
      write('The first Automaton A1 (deterministic) is:'),nl,
      write(A1),nl,
      write('The second Automaton A2 (deterministic) is:'),nl,
      write(A2),nl,
      write('And the Transition Function delta:'),nl,
      write(Delta),nl,
      FSTout = (A1,A2,Delta)
      ;write(' and it also is subsequential'),
      mark_as_sub_and-fun(FST,FSTout))). 


mark_as_non_fun(fa(A,B,C,D,E,F,G),fa(A,B,C,D,E,F,G,-)).
mark_as_sub_and-fun(fa(A,B,C,D,E,F,G),fa(A,B,C,D,E,F,G,+,+)).





%%------------------not implemented properly yet-------------


intersect_with_domain(InString,NonFunTransducer,OutString) :-
  % range, compose and identity are all defined in Gertjan's program, but they
  % are not actually carried out here
   OutString = range(compose(identity(InString),NonFunTransducer)).




%%---------------------check for subsequentiality of a transducer------------
%% 'is_subsequential' takes an unambiguous transducer and tells wether it is 
%% subsequential 
%% or not, following the algorithm in Roche & Schabes' (Titel), Figure 1.45.

is_subsequential(fa(Alpha,_,_,_,Trans,_,_),Bool):-
    OldC = [[0,[],0,[]]],
    is_sub0(Alpha,Trans,OldC,[],_,0,1,Bool1),
    (Bool1 == false
     -> Bool = false
     ; Bool = true).


is_sub0(_,_,_,_,_,_,0,_).
is_sub0(Alpha,Trans,OldC,OldTrans,NewTrans,Q,_,Bool1):-
    is_sub1(Alpha,Trans,OldC,NewC,OldTrans,NewTrans1,Q,Bool),
    (Bool == false
     -> Bool1 = false
    ;length(NewC,X),
    Q1 is Q + 1,
    Count is X - Q1,
    is_sub0(Alpha,Trans,NewC,NewTrans1,NewTrans,Q1,Count,Bool1)).


is_sub1([],_,X,X,Y,Y,_,_).
is_sub1([A|Alpha],Trans,OldC,NewC,OldTrans,NewTrans,Q,Bool):-
    array_at(Q,OldC,[X1,_,X2,_]),
    findall(trans(X1,A/W,Y1),member(trans(X1,A/W,Y1),Trans),Done1),
    findall(trans(X2,A/WPrime,Y2),member(trans(X2,A/WPrime,Y2),Trans),Done2),
    is_sub2(OldC,NewC1,OldTrans,NewTrans1,Q,Done1,Done2,Bool1),
    (Bool1 == false
     -> Bool = false    
    ;is_sub1(Alpha,Trans,NewC1,NewC,NewTrans1,NewTrans,Q,Bool)).

    


is_sub2(X,X,Y,Y,_,[],_,_).
is_sub2(OldC,NewC,OldTrans,NewTrans,Q,[A|Done1],Done2,Bool):-
    is_sub3(OldC,NewC1,OldTrans,NewTrans1,Q,A,Done2,Bool1),
    (Bool1 == false
     -> Bool = false
    ;is_sub2(NewC1,NewC,NewTrans1,NewTrans,Q,Done1,Done2,Bool)).

    
  
is_sub3(X,X,Y,Y,_,_,[],_).
is_sub3(OldC,NewC,OldTrans,NewTrans,Q,A,[B|Done2],Bool):-
    is_sub4(OldC,NewC1,OldTrans,NewTrans1,Q,A,B,Bool1),
    (Bool1 == false
     -> Bool = false
    ;is_sub3(NewC1,NewC,NewTrans1,NewTrans,Q,A,Done2,Bool)).

    

is_sub4(OldC,NewC,OldTrans,NewTrans,Q,trans(X1,A/W,Y1),trans(X2,A/WPrime,Y2),Bool):-
    array_at(Q,OldC,[X1,U1,X2,U2]),
    append(U1,W,Done),
    append(U2,WPrime,Done1),
    common_pref(Done,Done1,V),                 %%V
    delete_common_pref(V,Done,V1),             %%V1
    delete_common_pref(V,Done1,V2),            %%V2
    length(OldC,X),
    (member([Y1,V1,Y2,V2],OldC)
     -> NewC = OldC,
     array_at(Index,OldC,[Y1,V1,Y2,V2]),
     append([trans(Q,A/V,Index)],OldTrans,NewTrans)
     ;NewC = [[Y1,V1,Y2,V2]|OldC],
     append([trans(Q,A/V,X)],OldTrans,NewTrans)),
    findall([P,[Y1,U3,Y2,U4]],array_at(P,OldC,[Y1,U3,Y2,U4]),Result),
    test_it(Result,NewTrans,W,Q,[Y1,V1,Y2,V2],Bool).


test_it([],_,_,_,_,_).
test_it([[P,[Y1,U3,Y2,U4]]|Rest],NewTrans,W,Q,[Y1,V1,Y2,V2],Bool):-
    (Q > P   
     -> (path1(NewTrans,Q,_,W,P)
	 -> (U3 == V1
	     -> Bool = Bool1
	     ; Bool = false),
	 (U4 == V2
	  -> Bool = Bool1
	  ; Bool = false)
	 ;Bool = Bool1)
     ;Bool = Bool1),
    test_it(Rest,NewTrans,W,Q,[Y1,V1,Y2,V2],Bool1).

    

common_pref([],_,[]).
common_pref(_,[],[]).
common_pref([X|Rest1],[X|Rest2],[X|Result]):-
    common_pref(Rest1,Rest2,Result).
common_pref(_,_,[]).



%------------------check for functionality of a transducer-------------------

%% 'is_function' takes a transducer and its unambiguous representation as arguments to
%% decide wether the original transducer is functional, comparing the transducers with 
%% one another. You'll find the pseudocode in Roche & Schabes, Figure 1.32.
%% (there should be only one start state!)



%% is_function(fa([a,b],_,[0],[2],[trans(0,a/[b],1),trans(0,a/[b,c],3),trans(0,a/[b,c,d],2),trans(1,b/[c,d],2),trans(3,b/[d],2),trans(2,b/[],2)],[],_),Bool).
%% is_function(fa([b,a],_,[0],[0],[trans(0,a/[c],0),trans(0,a/[c,b],0)],[],_),FST,Bool).
%% is_function(fa([b,a],_,[0],[3],[trans(0,b/[z],1),trans(1,a/[x,x],1),trans(1,b/[z],3),trans(1,a/[x],2),trans(2,a/[],1)],[],_),FST,Bool).
%% is_function(fa([cC,oO,a,plus,o],_,[0],[4],[trans(0,cC/[c],1),trans(1,oO/[o],2),trans(2,plus/[],3),trans(3,o/[o],4),trans(3,a/[a],4),trans(2,plus/[min],5),trans(5,o/[o],4)],[],_),FST,Bool).


is_function(fa(Alpha,_,[Starts],Finals,Trans,_,_),fa(Alpha,_,[Starts],Finals1,Trans1,_,_),Bool):-
   Starts3 = 0,
   Q = 0,
   OldC = [(Starts,Starts3,[[],[]])],
   is_function1(Alpha,Finals,Trans,OldC,NewC,Q,_,Finals1,Trans1,[],_,1,Bool),
    write('Testing functionality, your C is: '),nl,
    write(NewC),nl,nl.



is_function1(_,_,_,X,X,_,[],_,_,Y,Y,0,_).
is_function1(Alpha,Finals2,Trans2,OldC,NewC1,Q,[Q|NewStates],Finals1,Trans1,OldTrans,NewTrans3,_,Bool):-
    check_if_false(Finals1,Finals2,OldC,Q,Bool1),
    (Bool1 == true
     -> part0(Alpha,Trans2,OldC,NewC2,Q,Trans1,OldTrans,NewTrans,Bool2)
     ; Bool2 = false),
    (Bool2 == false
      -> Bool = false,
     NewC1 = NewC2
      ; length(NewC2,X),
      Q1 is Q + 1,
      Count is X - Q1,
      is_function1(Alpha,Finals2,Trans2,NewC2,NewC1,Q1,NewStates,Finals1,Trans1,NewTrans,NewTrans3,Count,Bool3),
      (Bool3 == false
       -> Bool = false,
       (nonvar(NewC1)
	-> true
	;NewC1 = NewC2)
       ;Bool = true)).
   
 
check_if_false(Finals1,Finals2,OldC,Q,Bool1):-  
    array_at(Q,OldC,(X1,X2,[_,U])),
    (member(X1,Finals1)
     -> (member(X2,Finals2)
	-> (U == []
	    -> Bool1 = true
	    ;Bool1 = false)
	;Bool1 = true)
     ;Bool1 = true).


part0([],_,X,X,_,_,Y,Y,_).
part0([A|Alpha],Trans2,OldC,NewC,Q,Trans1,OldTrans,NewTrans,Bool):-          %do it for every input-sign
    part1(A,Trans2,OldC,NewC1,Q,Trans1,OldTrans,NewTrans1,Bool),
    (Bool == false
     -> NewC = NewC1
    ;part0(Alpha,Trans2,NewC1,NewC2,Q,Trans1,NewTrans1,NewTrans,Bool),
    put_together(NewC2,NewC1,NewC)).                             



part1(A,Trans2,OldC,NewC,Q,Trans1,OldTrans,NewTrans,Bool):-
    array_at(Q,OldC,(X1,X2,[_,U])),
    findall(trans(X1,A/W,Y1),member(trans(X1,A/W,Y1),Trans1),Done1),                    
    findall(trans(X2,A/WPrime,Y2),member(trans(X2,A/WPrime,Y2),Trans2),Done2),
    (U == out
     ->  make_new(Done1,Done2,OldC,NewC1,NewTrans1,[[],out],Q),
     NewC = NewC1,
     NewTrans = NewTrans1
     ;  part2(Done1,OldC,NewC2,Q,Done2,OldTrans,NewTrans2,Bool),
     put_together(OldC,NewC2,NewC),
     NewTrans = NewTrans2).


part2([],X,X,_,_,Y,Y,_).
part2([M|Trans1],OldC,NewC1,Q,Trans2,OldTrans,NewTrans,Bool):-
    part2_1(M,OldC,NewC,Q,Trans2,OldTrans,NewTrans1,Bool2),  
    (Bool2 == false
     -> Bool = false,
     NewC1 = NewC
     ;part2(Trans1,NewC,NewC1,Q,Trans2,NewTrans1,NewTrans,Bool1),
     Bool = Bool1).


part2_1(_,X,X,_,[],Y,Y,_).  
part2_1(trans(X1,A/W,Y1),OldC,NewC2,Q,[trans(X2,A/WPrime,Y2)|Trans2],OldTrans,NewTrans,Bool):-
     array_at(Q,OldC,(X1,X2,[Sign,U])),
    (Sign == +                             
     -> V1 = W,
     append(U,WPrime,V2)
     ; append(U,W,V1),
     V2 = WPrime),
    (longer(V1,V2)
     -> 
     delete_common_pref(V2,V1,Result1),
     (Result1 == no                         %% there was no common prefix
      -> V = [[],out]
      ; V = [-,Result1])
     ;
     delete_common_pref(V1,V2,Result3),
     (Result3 == []
	 -> V = [[],[]]
      ;(Result3 == no                         %% there was no common prefix
	-> V = [[],out]
	;V = [+,Result3]))),
     compare1(OldTrans,Q,OldC,Y1,Y2,V,Bool),
     make_new1(trans(X1,A/W,Y1),[trans(X2,A/WPrime,Y2)],OldC,NewC1,NewTrans1,V,Q), 
    (Bool == false
     -> append(NewTrans1,OldTrans,NewTrans),
     NewC2 = NewC1
     ; append(NewTrans1,OldTrans,OldTrans1),
     part2_1(trans(X1,A/W,Y1),NewC1,NewC2,Q,Trans2,OldTrans1,NewTrans,Bool)).
    


compare1(Trans,Q,C,Y1,Y2,V,Bool):-
    path(Trans,Q,Path,_),
    compare1_2(Path,C,Y1,Y2,V,Bool).

compare1_2([],_,_,_,_,_).
compare1_2([trans(X,_/_,_)|Path],C,Y1,Y2,[Sign1,V],Bool):-
    array_at(X,C,(X1,X2,[Sign,U])),
    (X1 == Y1
     -> (X2 == Y2
	 -> ([Sign1,V] == [Sign,U]
	     -> true
	     ;Bool = false)    %% non-functionality: break!
	 ;Bool = true)
     ;Bool = true),
    (Bool == false
     -> true
    ;compare1_2(Path,C,Y1,Y2,[Sign1,V],Bool)).
    

%% path([trans(0,a/[b],1),trans(1,a/[b,c],3),trans(3,a/[b,c,d],2)],2,P,Wort).

path([],_,[],[]).
path(Trans,Q,Path,Word):-
    (member(trans(X,A/W,Q),Trans)
     -> (X == Q
        -> pop_out(trans(X,A/W,Q),Trans,Trans1),
           path(Trans1,X,Path1,Word1),!
        ;path(Trans,X,Path1,Word1)),
	(member(trans(X,A/W,Q),Path1)
	 -> Path = Path1,
	 Word = Word1
	 ; two_poss(Word1,W,Word,trans(X,A/W,Q),Path1,Path))
     ; Path = [],
     Word = []).




path1(Trans,Q,Path,Word,State):-
    reverse(Word,Word1),
    path1_2(Trans,Q,Path,Word1,State).

path1_2(_,Q,[],[],Q).
path1_2(Trans,Q,Path,Word,State):-
    (member(trans(X,A/W,Q),Trans)
     -> rev1(W,W1),
     (common_pref(W1,Word,W1)
      -> delete_common_pref(W1,Word,Word1),
      path1_2(Trans,X,Path1,Word1,State1),
      append([trans(X,A/W,Q)],Path1,Path),
      State = State1
      ;true)
     ;true).

rev1(W,W1):-
    reverse(W,W1).
rev1(W,W).

two_poss(Word1,W,Word,trans(X,A/W,Q),Path1,Path):-
    append(Word1,W,Word),
    Path = [trans(X,A/W,Q)|Path1].
two_poss(_,_,Word,trans(X,A/W,Q),_,Path):-
    Path = [trans(X,A/W,Q)],
    Word = W.



delete_common_pref([],[],[]).
delete_common_pref([],Rest,Result):-
   Result = Rest.
delete_common_pref([X|Rest1],[X|Rest2],Result):-
   delete_common_pref(Rest1,Rest2,Result).
delete_common_pref(_,_,no).       



array_at(Index,List,Elem):-
    reverse(List,List1),
    nth0(Index,List1,Elem).


longer(V1,V2):-
   length(V1,X1),
   length(V2,Y1),
   X1 > Y1.



make_new([],_,X,X,[],_,_).
make_new([M|Trans1],Trans2,OldC,NewC,NewTrans,OUTORNOT,Q):-
   make_new1(M,Trans2,OldC,NewC1,NewTrans1,OUTORNOT,Q),
   make_new(Trans1,Trans2,NewC1,NewC2,NewTrans2,OUTORNOT,Q),  
   append(NewTrans1,NewTrans2,NewTrans),
   append(NewC1,NewC2,NewC3),
   only_one(NewC3,NewC).

make_new1(_,[],OldC,OldC,[],_,_).
make_new1(trans(X1,A/W,Y1),[trans(_,A/_,Y2)|Trans],OldC,NewC,NewTrans,OUTORNOT,Q):-
    make_new1(trans(X1,A/W,Y1),Trans,OldC,NewC1,NewTrans1,OUTORNOT,Q),
    (member((Y1,Y2,OUTORNOT),NewC1)
    -> NewC = NewC1,
    array_at(Num,NewC1,(Y1,Y2,OUTORNOT)),
    NewTrans = [trans(Q,A/W,Num)|NewTrans1]
    ;NewC = [(Y1,Y2,OUTORNOT)|NewC1],
    length(NewC,X),
    Num1 is X - 1,
    NewTrans = [trans(Q,A/W,Num1)|NewTrans1]).   


%% 'only_one' avoids dubble entries for entries in a list

only_one([],[]). 
only_one([A|Set_of_trans1],Set_of_trans):-
   member(A,Set_of_trans1),
   only_one(Set_of_trans1,Set_of_trans).
only_one([A|Set_of_trans1],[A|Set_of_trans]):-
   only_one(Set_of_trans1,Set_of_trans).



%----------------------------building a bimachine-----------------------

%% It becomes necessary to buil a bimachine if you want to run an input on a functional 
%% notsequential transducer. If you didn't use a bimachine, there would result an unfinite 
%% number of outputs! In Gertjan van Noord's package he just produces some of the possible
%% outputs and then interrupts the process, which is one, but not a very nice way to do things.

%bimachine( fa([keep,under,control,out,of,reach],7,[0],[3],[trans(0,keep/[keep_1],1),trans(0,keep/[keep_2],4),trans(1,under/[under],2),trans(2,control/[control],3),trans(4,out/[out],5),trans(5,of/[of],6),trans(6,reach/[reach],3)],Jumps,_), Delta).



bimachine(fa(Alpha,NumStates,Starts,Finals,Trans,Jumps,_),A1,A2,Delta):-
   bimachine1(Alpha,NumStates,Starts,Finals,Trans,Jumps,A1,A2,Delta).


bimachine1(Alpha,NumStates,Starts,Finals,Trans,Jumps,A1,A2,Delta):-
    turn_into_fsa(Trans,Result),
    turn_into_fsa(Jumps,Jumps1),
    my_determinize(Alpha,NumStates,Result,Starts,Finals,Jumps1,Subsets1,A1),
    A1 = fa(Alpha,Num1,[0],_,NewTrans1,[],_),
    reverse_fst(Trans,Result0),
    reverse_fst(Jumps,Jumps2), 
    turn_into_fsa(Result0,Result1),
    turn_into_fsa(Jumps2,Jumps3),
    my_determinize(Alpha,NumStates,Result1,Finals,Starts,Jumps3,Subsets2,A2),
    A2 = fa(Alpha,Num2,[0],_,NewTrans2,[],_),
    make_delta(Trans,Subsets1,Subsets2,Delta),
    count_states(NewTrans1,Num1),
    count_states(NewTrans2,Num2).


%% 'turn_into_fsa' makes an automaton out of a transducer in the sense that
%% it cuts away the output of every transition.

turn_into_fsa([],[]).                                                   
turn_into_fsa([trans(Beg,Inp/_,Fin)|T],[trans(Beg,Inp,Fin)|A1]):-       
   turn_into_fsa(T,A1),!.


reverse_fst([],[]).
reverse_fst([trans(Beg,Inp/Outp,Fin)|A1],[trans(Fin,Inp/Outp,Beg)|A2]):-
   reverse_fst(A1,A2).


%% 'make_delta' constructs the transition-function Delta out of the subsets
%% of the two deterministic automata and the transitions of the transducer.

make_delta(_,[],_,[]).   
make_delta(T,[X|Subsets1],Subsets2,Delta):-
   m_d0(T,X,Subsets2,D1),
   make_delta(T,Subsets1,Subsets2,D2),
   append(D1,D2,Delta).


m_d0(_,[_,[]],_,[]).
m_d0(T,[Num1,[A|Rest1]],B,Delta):-
    m_d1(T,[Num1,A],B,D1),
    m_d0(T,[Num1,Rest1],B,D2),
    append(D1,D2,Delta).


m_d1(_,_,[],[]).
m_d1(T,[Num1,A],[Set|Rest2],Delta):-
    m_d2(T,[Num1,A],Set,D1),
    m_d1(T,[Num1,A],Rest2,D2),
    append(D1,D2,Delta).
 

m_d2(_,_,[_,[]],[]).
m_d2(T,[Num1,A],[Num2,[B|Rest]],Delta):-
    m_d3(T,[Num1,A],[Num2,B],D1),
    m_d2(T,[Num1,A],[Num2,Rest],D2),
    append(D1,D2,Delta).    

m_d3(T,[Num1,A],[Num2,B],Delta):-
    (member(trans(A,C/D,B),T)
    -> Delta = [trans(Num1,C/D,Num2)]
    ; Delta = []).
	  

run_bimachine(InString,KnownTranducerType,OutString):-
    KnownTranducerType = (A1,A2,Delta),
    A1 = fa(_,_,Start1,Finals1,Trans1,_,_),
    A2 = fa(_,_,Start2,Finals2,Trans2,_,_),
    process_input(Trans1,Start1,Finals1,Trans2,Start2,Finals2,Delta,InString,OutString).



process_input(Trans1,[Start1],Finals1,Trans2,[Start2],Finals2,Delta,Input,Output):-
    with_A1(Trans1,Start1,Finals1,Input,Stack),
    reverse(Input,Input1),
    process_on_A2(Stack,Trans2,Start2,Finals2,Delta,Input1,Output1),
    reverse(Output1,Output2),
    merge(Output2,Output).



%% 'with_A1' processes the Input on the automaton 1 
%% and stores sign and states in a list

with_A1(_,State,Finals,[],[]):-  
    member(State,Finals). 
with_A1(Trans,Start,Finals,[Sign|Input],[[Start,Sign]|Stack]):-
    member(trans(Start,Sign,NewStart),Trans),
    with_A1(Trans,NewStart,Finals,Input,Stack).



%% 'process_on_A2' processes the Input on the automaton 2
%% with respect to A1 and Delta.

process_on_A2(_,_,State,Finals,_,[],[]):- 
    member(State,Finals).  
process_on_A2(Stack,Trans2,Start,Finals,Delta,[Sign|Input],[Output|Out]):-
    member(trans(Start,Sign,NewStart),Trans2),
    member([State,Sign],Stack),
    member(trans(State,Sign/Output,Start),Delta),
    process_on_A2(Stack,Trans2,NewStart,Finals,Delta,Input,Out).



    

%%-------------------------------determinisation------------------------------------
%% 'my_determinize' determinizes an automaton using the Myhill/Buechi algorithm.


%% my_determinize([a],3,[trans(0,a,2),trans(0,a,1)],[0],[0,2],[trans(1,[],0)],NewFinals,NewTrans,Subsets).
%% my_determinize([a],3,[trans(0,a,2),trans(0,a,1)],[0],[1,2],[trans(2,a,1)],NewFinals,NewTrans,Subsets).


my_determinize(Alpha,NumStates,Trans,[Start],Finals,Jumps,Result,fa(Alpha,Num,[0],NewFinals0,NewTrans0,[],_)):-
    list_of_states(NumStates,States),
    spontanous(States,Jumps,Successors),
    array_at(Start,Successors,(_,Done)),
    my_determinize2(Alpha,Successors,Trans,[Done],Finals,Jumps,NewFinals,NewTrans,_,[Done],Subsets1),
    length(Subsets1,X),
    reverse(Subsets1,Subsets2),
    append(Subsets2,[Done],Result1),
    rename_Subs(Result1,Result,X),
    rename_Trans(Result,NewTrans,NewTrans0),
    rename_Fin(NewFinals,Result,NewFinals1),
    only_one(NewFinals1,NewFinals0),
    count_states(NewTrans0,Num).


my_determinize2(_,_,_,[],_,_,[],[],[],_,[]).
my_determinize2(Alpha,Succ,Trans,[Starts|Rest],Finals,Jumps,NewFinals,NewTrans,_,States,Subsets):-
    my_determinize0(Jumps,Succ,Alpha,Trans,Starts,Finals,NewTrans1,Subsets1),
    put_together(Subsets1,States,States2),
    (member(Starts,Subsets1)
    -> pop_out(Starts,Subsets1,Done),
    append(Done,Rest,NewStarts)
    ;append(Subsets1,Rest,NewStarts)),
    my_determinize2(Alpha,Succ,Trans,NewStarts,Finals,Jumps,NewFinals2,NewTrans2,_,States2,Subsets2),
    intersect(Starts,Finals,Bool),
    ( Bool == 'true'
    -> NewFinals = [Starts|NewFinals2]
    ;  NewFinals = NewFinals2),
    append(NewTrans1,NewTrans2,NewTrans),
    put_together(Subsets1,Subsets2,Subsets).


rename_Trans(_,[],[]).
rename_Trans(Subsets,[trans(A,B,C)|Trans],[trans(Num1,B,Num2)|NewTrans]):-
    member([Num1,A],Subsets),
    member([Num2,C],Subsets),
    rename_Trans(Subsets,Trans,NewTrans).
    

rename_Subs([],[],-1).
rename_Subs([Set|Subsets],[[X,Set]|Result],X):-
    X1 is X - 1,
    rename_Subs(Subsets,Result,X1).

rename_Fin([],_,[]).
rename_Fin([Set|NewFinals],Subsets,[Num|Result]):-
    member([Num,Set],Subsets),
    rename_Fin(NewFinals,Subsets,Result).


count_states(Trans,Num):-
    count_states1(Trans,States),
    only_one(States,Done),
    length(Done,Num).

count_states1([],[]).
count_states1([trans(A,_,C)|Trans],[A,C|States]):-
    count_states1(Trans,States).


pop_out(_,[],[]).                                     
pop_out(Elem,[Elem|List],List).
pop_out(Elem,[X|List],[X|Done]):-
    pop_out(Elem,List,Done).

%% makes the union of two sets.
put_together([],X,X).
put_together([Set|Set1],Set2,Set3):-
    member(Set,Set2),
    put_together(Set1,Set2,Set3).
put_together([Set|Set1],Set2,[Set|Set3]):-
    put_together(Set1,Set2,Set3).


spontanous([],_,[]).
spontanous([S|States],Jumps,[(S,Result)|Succ]):-
    jump_from_to(S,Jumps,Result1),
    append([S],Result1,Result),
    spontanous(States,Jumps,Succ).
    


list_of_states(0,[]).
list_of_states(NumStates,List):-
    list_of_states(NumStates1,List1),
    NumStates is NumStates1 + 1,
    List = [NumStates1|List1].



collect_states(Start,Jumps,NewStarts):-
    jump_from_to(Start,Jumps,Result),
    merge(Result,Result2),
    append(Result2,Start,NewStarts).


my_determinize0(_,_,[],_,_,_,[],[]).
my_determinize0(Jumps,Succ,[B|Alpha],Trans,NewStarts,Finals,NewTrans,Subsets):-
    find_transitions(Jumps,Succ,Trans,NewStarts,B,Finals,Subsets1),
    my_determinize0(Jumps,Succ,Alpha,Trans,NewStarts,Finals,NewTrans2,Subsets2),
    (Subsets1 == []
    ->  NewTrans = NewTrans2,
    Subsets = Subsets2
    ;append([trans(NewStarts,B,Subsets1)],NewTrans2,NewTrans),
    append([Subsets1],Subsets2,Subsets3),
    only_one(Subsets3,Subsets)).


%% looks for possible transitions from one set of states to another
 
find_transitions(_,_,_,[],_,_,[]).                                     
find_transitions(Jumps,Succ,Trans,[A|Starts],B,Finals,Subset):-  
    findall(C,member(trans(A,B,C),Trans),X),  
    all_successors(Succ,X,Set_of_states),                                          
    append(Set_of_states,Subsets1,Subset),
    find_transitions(Jumps,Succ,Trans,Starts,B,Finals,Subsets1).


all_successors(_,[],[]).
all_successors(Succ,[S|X],Done):-
    array_at(S,Succ,(_,Done1)),
    all_successors(Succ,X,Done2),
    append(Done1,Done2,Done3),
    only_one(Done3,Done4),
    sort(Done4,Done).
   
 
%% finds all the jumps from all the members of a set of 
%% states an collects them.

jump_from_to([],_,[]).                        
jump_from_to([A|States],Jumps,[W|NewStates]):- 
   findall(C,member(trans(A,_,C),Jumps),W),
   jump_from_to(States,Jumps,NewStates).
jump_from_to(State,Jumps,W):-
   findall(C,member(trans(State,_,C),Jumps),W).


%% puts found things into one list

merge([],[]).                                 
merge([A|Rest],Result):-
   append(A,Result1,Result),
   merge(Rest,Result1).


%% intersect checks, if there is a nonempty intersection 
%% of "States" with "Finals", i.e. between two sets.

intersect([],_,_).
intersect(States,Finals,BoolOut):-
   intersect1(States,Finals,_,BoolOut).

intersect1([],_,false,false).
intersect1([A|States],Finals,Bool,BoolOut):-
   ( Bool == true
   -> BoolOut = true,
      intersect([],Finals,BoolOut)
   ; ( member(A,Finals)
      -> BoolOut = true,
         intersect([],Finals,BoolOut)
         ; Bool = false,
           intersect1(States,Finals,Bool,BoolOut))).


%%---------------------building an unambiguous transducer-----------------------

%% 'unambiguous' builds an unambiguous representation of a transducer
%% following the Roche & Schabes algorithm, Figure 1.28.
%% (!!here, 'Starts' is only a sigleton list, not a list!!)

%% unambiguous(fa([a,c,cb],_,[0],[0],[trans(0,a/[b],1),trans(0,a/[b,c],3),trans(0,a/[b,c,d],2),trans(1,b/[c,d],2),trans(3,b/[d],2)],[],_),Fsa2).
%% unambiguous(fa([a,c,cb],_,[0],[0],[trans(0,a/[c,b],0),trans(0,a/[c],0)],[],_),Fsa2).
%% unambiguous(fa([a,c,b],_,[0],[1],[trans(0,a/[c],1),trans(0,a/[b],1)],[],_),Fsa2).



unambiguous(FST,FSTout):-
   FST = fa(Alpha,_,[Starts],Finals,Trans,Jumps,_),
   C = [(Starts,[],0)],
   Q = 0,
   unambiguous1(Starts,Finals,Trans,Jumps,C,_,Q,NewStates,NewFinals,NewTrans,1),
   only_one(NewTrans,NewTrans1),
   length(NewStates,NumStates),
   FSTout = fa(Alpha,NumStates,[0],NewFinals,NewTrans1,Jumps,_).


unambiguous1(_,_,_,_,_,[],_,[],[],[],0).
unambiguous1(_,_,_,_,_,[],_,[],[],[],Y):-
   Y < 0.
unambiguous1(Starts,Finals,Trans,Jumps,OldC,NewC2,Q,[Q|NewStates],NewFinals,NewTrans0,_):-
   array_at(Q,OldC,(X1,Set,_)),
   findall(trans(X1,B/F,D),member(trans(X1,B/F,D),Trans),Done),    
   %% only the transitions that also begin with X1 are relevant
   loop1(Done,OldC,NewC,Q,_,NewTrans,Trans,Set), 
   length(NewC,X),
   Q1 is Q + 1,
   Y is X - Q1,
   unambiguous1(Starts,Finals,Trans,Jumps,NewC,NewC2,Q1,NewStates,NewFinals1,NewTrans1,Y),
   intersect(Finals,Set,Bool),  
   ( member(X1,Finals)                               %% collects the new final states
    -> (Bool == false
	-> append([Q],NewFinals1,NewFinals)
	; NewFinals = NewFinals1)
    ; NewFinals = NewFinals1),
   append(NewTrans,NewTrans1,NewTrans0).




loop1([],X,X,_,[],[],_,_).
loop1([trans(X1,B/F,D)|Trans],OldC,NewC2,Q,[SetPrime1,SetPrime2],NewTrans,Transi,Set):-
    loop2(trans(X1,B/F,D),OldC,NewC1,Q,SetPrime1,NewTrans1,Transi,Set),
    loop1(Trans,NewC1,NewC2,Q,SetPrime2,NewTrans2,Transi,Set),
    append(NewTrans1,NewTrans2,NewTrans).




loop2(trans(X1,B/F,D),OldC,NewC,Q,SetPrime4,NewTrans,Transi,Set):-
   findall(trans(X1,B/Z,Y),member(trans(X1,B/Z,Y),Transi),Done),   %%finds all comparable transitions
   compare(trans(X1,B/F,D),Done,Set,SetPrime1,Transi),
   compare2(trans(X1,B/F,D),Set,SetPrime2,Transi),
   append(SetPrime1,SetPrime2,SetPrime4),
   intersect([D],SetPrime4,Bool),
   ( Bool == false
    ->  (member((D,SetPrime4,I),OldC)    
	 -> NewC = OldC,
	 NewTrans = [trans(Q,B/F,I)]         %% new transition for the unambig. transducer.
	 ; length(OldC,X),
	 NewC = [(D,SetPrime4,X)|OldC],      %% here, the new entries for C are made        
	 NewTrans = [trans(Q,B/F,X)])        
    ; NewC = OldC,
      NewTrans = []).

%% looks for transitions with different start-
%% and arrival-state, but same in/output.

compare(_,[],_,[],_). 
compare(trans(_,_/C,_),[trans(_,_/Z,W)|Trans],Set,SetPrime,Transi):-  
   ( C @> Z                                  
   ->  append([W],SetPrime1,SetPrime)
    ; SetPrime = SetPrime1),
    compare(trans(_,_/C,_),Trans,Set,SetPrime1,Transi).


compare2(_,[],[],_).
compare2(trans(X1,B/F,D),[State|Set],SetPrime,Transi):-
    (member(trans(State,B/_,Y),Transi)
     -> SetPrime = [Y|SetPrime1]
     ;SetPrime = SetPrime1),
    compare2(trans(X1,B/F,D),Set,SetPrime1,Transi).



%%------------normal processing of a word on a transducer------------------
    
run_transducer([],fa(_,_,[State],Finals,_,_,_,_,_),[]):-
    (member(State,Finals)->
     true
     ;write('sorry, there is no path through the transducer with this input!')).
run_transducer([Sign|InString],fa(_,_,[Start],Finals,Trans,Jumps,_,+,+),OutString):-
    append(Trans,Jumps,Done),
    (member(trans(Start,Sign/Out,State),Done)
     -> run_transducer(InString,fa(_,_,[State],Finals,Trans,Jumps,_,+,+),OutString1),
     append(Out,OutString1,OutString)
     ; write('sorry, there is no path through the transducer with this input!')).



%%------------normal processing of a word on an automaton------------------
    
run_autom([],fa(_,_,[State],Finals,_,_,_,_,_),[]):-
    (member(State,Finals)->
     true
     ;write('sorry, there is no path through the transducer with this input!')).
run_autom([Sign|InString],fa(_,_,[Start],Finals,Trans,[],[]),OutString):-
    (member(trans(Start,Sign/Out,State),Trans)
     -> run_autom(InString,fa(_,_,[State],Finals,Trans,[],[]),OutString1),
     append(Out,OutString1,OutString)
     ; write('sorry, there is no path through the transducer with this input!')).






