%%           -*-Mode: prolog;-*-

%% Dale Gerdemann

%% Implementation of: Yli-Jyr and Koskenniemi, "Compiling Contextual
%% Restrictions on Strings into Finite-State Automata". A number of
%% well-known operators are redefined in terms of
%% "generalized_restriction".

%% The implementation is mostly straightforward from the article. The
%% only modification is to use the approach to marker symbols as in
%% Gerdemann & van Noord, ``Transducers from Rewrite Rules with
%% Backreferences''. EACL 99.

%% user:re:- nl,nl,fsa_regex:fsa_reconsult_aux_file('context_restrict.pl').
:- discontiguous macro/2.
:- op(402,yfx,(=>)).
:- op(402,yfx,(<=)).

% generalized_restriction
% Precondition: number of diamonds is the same in all strings.
% This is not defined as a macro because we want L and R to be
% Prolog lists not interpreted as the concatenation operator.
% It doesn't need to be a macro anyway, since end users would
% not likely use it.
generalized_restriction(L,R, ~elim_markers(set(L) - set(R))).

% context restrition
% Example: X => (V1,Y1;V2,Y2;...;VN,YN)
% Every X must occur in one of the contexts Vn __ Yn
macro(A0 => B0, GR^) :-
    two_diamond([A0],A),
    convert_contexts_alt(B0,L,R),
    two_diamond(L,R,B),
    generalized_restriction(A,B,GR).

% coercion
% Example: X <= (V1,Y1;V2,Y2;...;VN,YN)
% For every context Vn __ Yn, the gap must be filled by X
macro(A0 <= B0, GR^) :-
    two_diamond([A0],A),
    convert_contexts(B0,L,R),
    two_diamond(L,R,B),
    generalized_restriction(A,B,GR).

macro(if_p_then_s(L10,L20), GR^) :-
    diamond_left([L10],L1),
    diamond_right([L20],L2),
    generalized_restriction(L1,L2,GR).

macro(if_s_then_p(L10,L20), GR^) :-
    diamond_left([L10],L1),
    diamond_right([L20],L2),
    generalized_restriction(L2,L1,GR).

macro(nowhere(A),GR^) :-
    generalized_restriction([[sig*,non_markers(A),sig*]],[],GR).    


macro(sig,[?,0]).
macro(xsig,[?,{0,1}]).
macro(diamond, ['*',1]).
macro(non_markers,[?,[]:0]*).
macro(non_markers(E),
      range(E o non_markers)).
macro(elim_markers(E), minimize(range(E o {[?,0:[]],[?,1]:[]}*))).


diamond_left([],[]).
diamond_left([F|R],
	   [[non_markers(F),diamond,sig*]|Rest]) :-
    diamond_left(R,Rest).


diamond_right([],[]).
diamond_right([F|R],
	   [[sig*,diamond,non_markers(F)]|Rest]) :-
    diamond_right(R,Rest).


two_diamond([],[]).
two_diamond([F|R],
	   [[sig*,diamond,non_markers(F),diamond,sig*]|Rest]) :-
    two_diamond(R,Rest).


two_diamond([],[],[]).
two_diamond([F1|R1],[F2|R2],
       [[non_markers(F1),diamond,sig*,diamond,non_markers(F2)]|Rest]) :-
    two_diamond(R1,R2,Rest).
		 
   
convert_contexts(C,[FL|RL],[FR|RR]) :-
    functor(C,';',_), !,
    arg(1,C,(FL,FR)),
    arg(2,C,Rest),
    convert_contexts(Rest,RL,RR).
convert_contexts((A,B),[A],[B]) :- !.
convert_contexts('()',[],[]).   % little hack to allow empty context    


% Alternative version for xfst version of context restrict
convert_contexts_alt(C,[[? *,FL]|RL],[[FR,? *]|RR]) :-
    functor(C,';',_), !,
    arg(1,C,(FL,FR)),
    arg(2,C,Rest),
    convert_contexts_alt(Rest,RL,RR).
convert_contexts_alt((A,B),[[? *,A]],[[B,? *]]) :- !.
convert_contexts_alt('()',[],[]).   % little hack to allow empty context    


