:- multifile help_info/4.
:- discontiguous help_info/4.

#swi# :- style_check([-atom]).

#yap# :- use_module(library(lists),  [ append/3, memberchk/2 ]).
#sic# :- use_module(library(file_systems), [ current_directory/2 ]).
#yap# :- use_module(library(charsio),[ format_to_chars/3 ]).
#!yap# :- use_module(charsio,[ format_to_chars/3 ]).

:- dynamic fsa_command_alias:alias/2.

:- use_module(fsa_util).

:- discontiguous b_cmd/3.
:- discontiguous b_cmd_help/3.

help_info(class,command,"The Command Interpreter",
"The FSA command interpreter provides line-based interaction with the\
 \n FSA functionality. The command interpreter provides a history\
 \n mechanism, escape to the operating system, escape to Prolog and on-line\
 \n help information. All startup options for fsa are also available as\
 \n commands in the command interpreter. ").

help_info(command,syntax,"Syntax",
"A command is typed in by the user as one line of text; it's\
 \n tokenized as a sequence of `words', where spaces and tabs are\
 \n treated as separators. Each word is treated as an atomic (Prolog\
 \n atom or Prolog integer), unless it is written within { and }. In\
 \n the latter case the `word' is parsed as a Prolog term (in the latter\
 \n case spaces and tabs are not interpreted as separators).\
 \n \
 \n     |: flag jan jan(a,b,c)\
 \n \
 \n is equivalent to the Prolog goal \
 \n \
 \n     ?- fsa_globals:fsa_global_set(jan,'jan(a,b,c)')\
 \n \
 \n wherease\
 \n \
 \n     |: flag jan {jan(a,b,c)}\
 \n \
 \n is equivalent to the Prolog goal \
 \n \
 \n     ?- fsa_globals:fsa_global_set(jan,jan(a,b,c))\
 \n \
 \n Variables occuring in such terms have scope over the full \
 \n command-line!").

help_info(command,alias_and_history,"Alias and History",
"The command-interpreter has an alias mechanism which subsumes\
 \n a history mechanism as well. All occurences of $word are replaced \
 \n by the definition of the alias word. The alias command itself can \
 \n be used to define aliases:\
 \n \
 \n     19 |: alias hallo ! cat hallo\
 \n     20 |: $hallo\
 \n \
 \n so command number 20 will have the same effect as typing\
 \n \
 \n     33 |: ! cat hallo\
 \n \
 \n and if this command had indeed been typed as command number 33 then\
 \n typing\
 \n \
 \n     35 |: $33\
 \n \
 \n gives also the same result. The special meaning of $ can be turned off\
 \n by prefixing it with another $, e.g.:\
 \n \
 \n     |: cd $$HOME\
 \n \
 \n Moreover, if no alias has been defined, then it will apply the last\
 \n command that started with the name of the alias:\
 \n \
 \n     66 |: parse john kisses mary\
 \n     67 |: $parse\
 \n \
 \n will have the same meaning (in this order) if the macro parse \
 \n is not defined.").

help_info(command,prolog_goals,"Prolog goals",
"It is also possible to issue Prolog commands; however some\
 \n restrictions apply. \
 \n \
 \n     39 |: p {member(X,[X|T])}\
 \n \
 \n Note that this may succeed, but 'yes' or 'no' and variable bindings \
 \n will NOT be printed.").

help_info(command,start_stop,"Starting and Stopping the command interpreter",
"If no action option is provided, then the system runs in interactive\
 \n mode. If the variable *interpreter* is set to *on*, then fsa runs in\
 \n interactive mode after the action indicated by the action option has\
 \n been performed.  If the flag *interactive* has been set to *cmdint*,\
 \n then fsa runs the FSA command interpreter. The command interpreter can\
 \n be started any time from the Prolog prompt using the command r/0:\
 \n \
 \n     | ?- r.\
 \n     *** Welcome to the FSA Command Interpreter (type ? for help) ***\
 \n         5 |:\
 \n \
 \n You can stop the command interpreter using the *p* command:\
 \n \
 \n     5 |: p\
 \n     *** execution interrupted ***\
 \n \
 \n     yes\
 \n     | ?- \
 \n \
 \n You can quit FSA entirely using the *quit* command.  Note that you can\
 \n also use the command interpreter together with the graphical user\
 \n interface.  ").

:- public r/0, fsa_cmdint/0, restart/0.

r :-
    fsa_cmdint.

fsa_cmdint :-
    format(user_error,
	"*** Welcome to the FSA Command Interpreter (type ? for help) ***~n",
	   []),
   catch(restart,Pat,handle_exception(Pat)).

handle_exception(Exception) :-
    print_message(error,Exception),
    fsa_cmdint.

restart :-
    repeat,                     % repeat loop
    create_prompt(N,Prompt),	% incorporate number in prompt
    prompt(_,Prompt),           % and redefine prompt
    fsa_read_line(Input_line,End_input),
    (	End_input == end_of_file
    ->	true
    ;	replace_aliases(Input_line,[H|T]),
	usr_cmd(N,[H|T],End_input)
    ),
    !. % repeat cut
        
usr_cmd(N,[H|T],end_of_line) :-          % fail if Input line is empty
    (	parse_cmd(Cmd,[H|T],[])
    ->	increase_command_count,
	alias([N,H|T]),		% define alias for command counter
	exec_cmd(Cmd)		% execute it, and fail (unless Cmd=prolog)
    ;	format(user_output,"*** error: command cannot be parsed ***~n",[]),
	fail
    ).

help_info(command,prolog,"p[rolog]","Stops the command interpreter.").
exec_cmd(Cmd) :-
    (	Cmd == prolog
    ->	format(user_error,"*** execution interrupted ***~n",[])
    ;
#!swi#	     if( Cmd, 
#!swi#           true,
#swi#        (   Cmd 
#swi#        *-> true
#swi#        ;
                 format(user_error,
		        "~n*** error: command failed: ***~n~w~n",[Cmd])
        ),
   	fail
    ).

:- (   bb_get(command_count,_)
   ->  true
   ;   bb_put(command_count,1)
   ).

command_count(N) :-
    bb_get(command_count,N).

increase_command_count :-
    bb_get(command_count,N0),
    N is N0+1,
    bb_put(command_count,N).

create_prompt(N,Prompt):-
#sic#    dev_run(prolog:'$breaklevel'(BreakLevel,0),BreakLevel=0),
#!sic#    BreakLevel=0,
    command_count(N),
    (   BreakLevel =:= 0
    ->  charsio:format_to_chars('~w |: ',[N],Chars)
    ;   charsio:format_to_chars('{~w} ~w |: ',[BreakLevel,N],Chars)
    ),
    atom_codes(Prompt,Chars).

%% replace_aliases(+ListOfWords,-ListOfWords)
%% all words in the first list that start with
%% a dollar sign are replaced by appropriate
%% lists

%% 1: finished:
replace_aliases([],[]).

%% 2: escaped alias: note in that case that $ is concatenated to
%%    the following atom... (this is solely useful for the Tcl
%%    commands I believe).
%%
replace_aliases([Var|T0],T):-
    var(Var),
    !,
    T=[Var|T1],
    replace_aliases(T0,T1).
    
replace_aliases([$,$,W0|T0],[W|T]) :-
    !,
    atom_concat($,W0,W),
%    fsa_util:format_to_atom("$~a",[W0],W),
%%    fsa_util:concat($,W0,W),
    replace_aliases(T0,T).

%% 3: a proper alias
replace_aliases([$,Word|T],Res ):-
    command_alias(Word,Y),
    !,
    replace_aliases(T,T2),
    append(Y,T2,Res).

%% 4: find command that started with that word
replace_aliases([$,Word|T],Res ):-
    command_alias(_,[Word|Y]),
    !,
    replace_aliases(T,T2),
    append([Word|Y],T2,Res).

%% 5: not a dollar sign, continue
replace_aliases([H|T],[H|T2]):-
    replace_aliases(T,T2).

%%%%%%%%%%%%%%%%%%
%%% parse_cmd %%%%
%%%%%%%%%%%%%%%%%%

parse_cmd(X) -->
    command(X).

parse_cmd(X,[Start|L],[]) :-
    b_cmd(Start,X,L).

command(Cmd,[-,X|T],L) :-
    fsa_util:format_to_atom("-~a",[X],X1),
    fsa_options:fsa_command(Cmd,[X1|T],L).
command(Cmd,L0,L) :-
    fsa_options:fsa_command(Cmd,L0,L).
command(true,L0,L) :-
    fsa_options:parse_options(L0,L).

%% to set a flag
command((fsa_globals:fsa_global_set(Tag,New),
	 format(user_error,"~w=~w~n",[Tag,New]))) -->
        [flag],
        [Tag],
	[New].

% to see a flag
command(( fsa_globals:fsa_globals:fsa_global_get(Tag,Val),
	 format(user_error,"~w=~w~n",[Tag,Val]))) -->
        [flag],
        [Tag].
command(( fsa_globals:fsa_globals:fsa_global_get(Tag,Val),
	 format(user_error,"~w=~w~n",[Tag,Val]))) -->
        [Tag],
	{ fsa_globals:fsa_globals:fsa_global_decl(Tag,_,_,_,_) }.

help_info(command,Key,Short,Long) :-
	b_cmd_help(Key,Short,Long).

% for comments:
b_cmd('%',true,_).
b_cmd_help('%',"% Words","\
 \n ignores Words (comment). Note that there needs to be a space after\
 \n %."). 

#sic# b_cmd(fc,fcompile(List),List).
#sic# b_cmd_help(fc,"fc Files","fcompiles(Files)."). 

b_cmd(um,use_module(List),List).
b_cmd_help(um,"um Files","use_module(Files)."). 

b_cmd(el,ensure_loaded(List),List).
b_cmd_help(el,"el Files","ensure_loaded(Files)."). 

#!swi# b_cmd(c,compile(List),List).
#swi# b_cmd(c,consult(List),List).
b_cmd_help(c,"c Files","compile(Files)."). 

#!swi# b_cmd(rc,reconsult(List),List).
#swi# b_cmd(rc,consult(List),List).
b_cmd_help(rc,"rc Files","reconsult(Files)."). 

#sic# b_cmd(ld,load(List),List).
#sic# b_cmd_help(ld,"ld Files","load(Files)."). 

b_cmd(libum,use_module(List),List0):-
    add_l(List0,List).
b_cmd_help(libum,"libum Files",
		   "for each File, use_module(library(File)).").

#sic# b_cmd(librc,reconsult(List),List0):- add_l(List0,List).
#yap# b_cmd(librc,reconsult(List),List0):- add_l(List0,List).
#swi# b_cmd(librc,consult(List),List0):- add_l(List0,List).
b_cmd_help(librc,"librc Files",
		   "for each File, reconsult(library(File)).").

#sic# b_cmd(libc,compile(List),List0):- add_l(List0,List).
#yap# b_cmd(libc,compile(List),List0):- add_l(List0,List).
#swi# b_cmd(libc,consult(List),List0):- add_l(List0,List).
b_cmd_help(libc,"libc Files",
		   "for each File, compile(library(File)).").

b_cmd(libel,ensure_loaded(List),List0):-
    add_l(List0,List).
b_cmd_help(libel,"libel Files",
		   "for each File, ensure_loaded(library(File)).").

#sic# b_cmd(libld,load(List),List0):-    add_l(List0,List).
#sic# b_cmd_help(libld,"libld Files", "for each File, load(library(File)).").

add_l([],[]).
add_l([H|T0],[library(H)|T]) :-
    add_l(T0,T).

b_cmd(version,fsa_version,[]).
b_cmd_help(version,"version","displays version information."). 

b_cmd(quit,halt,[]).
b_cmd_help(quit,"quit","quits FSA."). 

b_cmd(b,break,[]).
b_cmd_help(b,"b","break; enters Prolog prompt at next break level."). 

b_cmd(d,debug,[]).
b_cmd_help(d,"d","debug/0."). 

b_cmd(nd,nodebug,[]).
b_cmd_help(nd,"nd","nodebug/0."). 

b_cmd(p,prolog,[]).
b_cmd(p,X,[X]).
b_cmd_help(p,"p [Goal]","\
 \n without Goal: quits command interpreter -- falls back to Prolog prompt;\
 \n with Goal: calls Goal. Normally you will need {} around the Goal. For example:\
 \n \
 \n     4 |: p { member(X,[a,b,c]), write(X), nl }\
 \n \
 \n       ").

b_cmd(!, ( fsa_util:concat_all([H|T],Cmd,' '),
	   fsa_shell(Cmd)
	 ),[H|T]).
b_cmd_help(!,"! Command","\
 \n Command is executed by the shell. Note that the space between ! and\
 \n Command is required."). 

b_cmd(alias,alias(List),List).
b_cmd_help(alias,"alias [Name [Val]]",
"No args: lists all aliases; one arg: displays alias Name; two args:\
 \n defines an alias Name with meaning Val."). 


b_cmd(help,help,[]).
b_cmd(help,fsa_help_module(O),[O]).
b_cmd(help,fsa_help_class(O,P),[O,P]).
b_cmd(help,fsa_help_key(O,P,Q),[O,P,Q]).
b_cmd_help(help,"help [Module [Class [Key]]]","\
 \n displays help on Module-Class-Key; use ? to get help on commands\
 \n only."). 

b_cmd(? ,fsa_help_class(command,user),[]).
b_cmd(? ,fsa_help_key(O,command,user),[O]).
b_cmd_help(? ,"?  [Cmd]",
"displays help on Cmd; without Cmd prints list of commands."). 


b_cmd(spy, spy(Module:Pred),L0) :-
    is_module(Module,L0,L1),
    pred(Pred,L1,[]).
b_cmd_help(spy,"spy [Module] Pred","\
 \n set spypoint on Module:Pred; Pred can either be Fun or Fun/Ar."). 

is_module(Module) -->
    [Module],
    { current_module(Module) }.
is_module(user,L,L).

%% for spy
pred(X/Y) -->
        [X],{atom(X)},
        ['/'],
        [Y],{integer(Y)}.
pred(X) -->
        [X],{atom(X)}.

#sic# b_cmd(cd,file_systems:current_directory(_,Dir),[Dir]).
#sic# b_cmd(cd,file_systems:current_directory(_,'$HOME'),[]).
#swi# b_cmd(cd,working_directory(_,Dir),[Dir]).
#swi# b_cmd(cd,working_directory(_,'$HOME'),[]).
#yap# b_cmd(cd,cd(Dir),[Dir]).
#yap# b_cmd(cd,cd('$HOME'),[]).

b_cmd_help(cd,"cd [Dir]","\
 \n change working directory to Dir; without argument cd to home\
 \n directory.").

b_cmd(pwd,fsa_shell(pwd),[]).
b_cmd_help(pwd,"pwd","print working directory.").

b_cmd(ls,fsa_shell(ls),[]).
b_cmd_help(ls,"ls","listing of directory contents").


%% alias
%% no arguments: list of all aliases
alias([]) :-
    (	command_alias(X,Z),
        write(X=Z),
        nl,
	fail
    ;	true
    ).

alias([H|T]) :-
    alias0(T,H).

%% one argument: show value of that particular alias
alias0([],H) :-
    command_alias(H,Z),
    write(H=Z),
    nl.

%% several arguments: set value of head of list to tail
alias0([Th|Tt],H):-
    add_command_alias(H,[Th|Tt]). 

b_cmd_help(fsa,"<any FSA startup option>","\
 \n Any valid option you can give to the fsa command is a valid command\
 \n for the command interpreter. For instance: \
 \n \
 \n     |: -d a.nd a.d\
 \n \
 \n     |: -m a.d a.m\
 \n \
 \n     |: -aux File\
 \n \
 \n     |: -tk\
 \n \
 \n     |: -r [[a,b]+,c]+").

command_alias(Alias,Command) :-
    fsa_command_alias:alias(Alias,Command),!.

add_command_alias(Alias,Command) :-
    retractall(fsa_command_alias:alias(Alias,_)),
    asserta(fsa_command_alias:alias(Alias,Command)).

% for some reason this one is missed by spxref...
:- public fsa_shell/1.
fsa_shell(Cmd) :-
#sic#    fsa_shell0(Cmd).
#yap#    unix(shell(Cmd)).
#swi#    unix(shell(Cmd)).


#sic# :- use_module(library(process)).
#sic# fsa_shell(Cmd):-
#sic#     process_create(sh,['-c',Cmd],[stdin(null),
#sic# 				  stdout(pipe(Out)),
#sic# 				  stderr(pipe(Err)),
#sic# 				  process(PID)]),
#sic#     copy_stream(Err),
#sic#     copy_stream(Out),
#sic#     close(Err),
#sic#     close(Out),
#sic#     process_wait(PID,Status),
#sic#     report_status(Status).
#sic# 
#sic# report_status(exit(0)) :- !.
#sic# report_status(Status) :-
#sic#     format(user_error,"warning: process has exit status ~w~n",[Status]).
#sic# 
#sic# copy_stream(Stream):-
#sic#     get_code(Stream,Char),
#sic#     copy_stream(Char,Stream).
#sic# 
#sic# copy_stream(Char0,Stream):-
#sic#     (   Char0 < 0
#sic#     ->  true
#sic#     ;   put_code(Char0),
#sic#         get_code(Stream,Char),
#sic#         copy_stream(Char,Stream)
#sic#     ).
#sic# 

%% special read_line for this file. This is all very messy.
%% code should be rewritten by using standard tokeniser...
%% fsa_read_line(-Word_list)
%% fsa_read_line(-Word_list,?End_input)
%%
%% fsa_read_line will read a line (ended with end of line)
%% and yields a list of words - a word is something between
%% spaces or between other 'syntax_char's

%read_line_from_stream(Stream,Line) :-
%    read_line_from_stream(Stream,Line,_End).
%
%read_line_from_stream(Stream,Line,End) :-
%    current_input(Old),
%    set_input(Stream),
%    call_cleanup(  read_line(Line,End), set_input(Old)  ).
%
%read_line_from_chars(Chars,Line):-
%    read_line_from_chars(Chars,Line,_End).
%
% :- use_module(library(charsio)).
%read_line_from_chars(Chars,Line,End) :-
%    charsio:open_chars_stream(Chars,read,Stream),
%    read_line_from_stream(Stream,Line,End).

%% to do: escape {} with \ 
fsa_read_line(List,End):-
#sic#    fsa_read_line(List,End,_).
#yap#    fsa_read_line(List,End,_).
#swi#    fsa_read_line(List,End,Chars0), 
#swi#    append(Chars,[_],Chars0),  % chop!
#swi#    atom_codes(Atom,Chars),
#swi#    rl_add_history(Atom).

char_type_backslash(C_mid,C_out,Type,Cs0,Cs) :-
    (	C_mid =:= 92
    ->	get_code(C_out),
	Cs0=[C_out|Cs],
	Type=normal
    ;   C_mid = C_out,
	Cs0=Cs,
	my_char_type(C_out,Type)
    ).

fsa_read_line(Word_list,End_input,[C_in|Cs]) :-
    get_code(C_in),
    skip_space(C_in,C_out,Cs,Cs1),
    read_words(C_out,Word_list,End_input,[],_Subs,Cs1,[]).

read_words(C_in0,Ws,End_input,Subs0,Subs,Cs00,Cs) :-
    char_type_backslash(C_in0,C_in,Type,Cs00,Cs0),
    (	Type == end_of_file
    ->  Ws=[], End_input=end_of_file, Subs0=Subs, Cs0=Cs
    ;   Type == end_of_line
    ->  Ws=[], End_input=end_of_line, Subs0=Subs, Cs0=Cs
    ;   Type == start_escape
    ->  Ws=[Term|Words],
	read_word_esc(Str,C_mid,Cs0,Cs1),
#sic#          read_from_chars(Str,Term,[variable_names(Pairs)]),
#yap#          read_from_chars(Str,Term), Pairs=[],
#swi#          atom_codes(Atom,Str), atom_to_term(Atom,Term,Pairs),
	add_subs(Pairs,Subs0,Subs1),
	skip_space(C_mid,C_out,Cs1,Cs2),
	read_words(C_out,Words,End_input,Subs1,Subs,Cs2,Cs)
    ;   Ws=[Word|Words],
	Subs0=Subs1,
        read_word(C_in,Str,C_mid,Cs0,Cs1),
        name(Word,Str),  %% real name/2, and not atom_codes!
        skip_space(C_mid,C_out,Cs1,Cs2),
        read_words(C_out,Words,End_input,Subs1,Subs,Cs2,Cs)
    ).

%% which syntax_char do we really need???
my_char_type(Char,Type) :-
    (	Char =:= 32
    ->	Type = space
    ;   Char =:= 9
    ->  Type = space
    ;   Char =:= 26
    ->  Type = end_of_file
    ;   Char =:= -1
    ->  Type = end_of_file
    ;   Char =:= 10
    ->  Type = end_of_line
    ;   Char =:= 36
    ->  Type = syntax_char
    ;   Char =:= 123
    ->  Type=start_escape
    ;   Char =:= 125
    ->  Type=end_escape
    ;   Type = normal
    ).

%% 32=<space>, 9=<tab>, 26=<^Z>, -1=<EOF>, 10=<CR>
%% 33=! 36=$ 40=( 41=) 42=* 43=+ 44=,
%% 58=: 59=; 60=< 61== 62=> 63=? 
%% 91=[ 92=\  93=] 123={ 124=| 125=}
read_word(Next,Word,Last,Cs0,Cs) :-
    my_char_type(Next,Type),
    read_word(Type,Next,Word,Last,Cs0,Cs).

read_word(space,Next,[],Next,Cs,Cs).
read_word(end_of_file,Next,[],Next,Cs,Cs).
read_word(end_of_line,Next,[],Next,Cs,Cs).
read_word(syntax_char,Next,[Next],Last,[Last|Cs],Cs):-
    get_code(Last).
read_word(normal,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == syntax_char
    ->	Chars = [],
	Last = C_mid,
	Cs1=Cs
    ;	read_word(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).

read_word(start_escape,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == syntax_char
    ->	Chars = [],
	Last = C_mid,
	Cs1=Cs
    ;	read_word(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).

read_word(end_escape,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == syntax_char
    ->	Chars = [],
	Last = C_mid,
	Cs1=Cs
    ;	read_word(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).

%% read_word reads anything up to the next space as an atom
%% read_esc_word reads anything up to the next } as a term, it
%% removes } and adds space . 32,46

read_word_esc(Word,Last,[Next0|Cs0],Cs) :-
    get_code(Next0),
    char_type_backslash(Next0,Next,Type,Cs0,Cs1),
    read_word_esc(Type,Next,Word,Last,Cs1,Cs).

read_word_esc(end_of_file,_,_,_,Cs,Cs):-
    throw(fsa_error("Unexpected end_of_file; expected }~n",[])).
read_word_esc(end_of_line,_,_,_,Cs,Cs):-
    throw(fsa_error("Unexpected end_of_line; expected }~n",[])).
read_word_esc(end_escape,_,_,_,Cs,Cs):-
    throw(fsa_error("Unexpected }~n",[])).
read_word_esc(normal,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == end_escape
    ->	Chars=[32,46],
	get_code(Last), Cs1=[Last|Cs]
    ;   read_word_esc(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).

read_word_esc(start_escape,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == end_escape
    ->	Chars=[32,46],
	get_code(Last), Cs1=[Last|Cs]
    ;   read_word_esc(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).

read_word_esc(space,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == end_escape
    ->	Chars=[32,46],
	get_code(Last), Cs1=[Last|Cs]
    ;   read_word_esc(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).

read_word_esc(syntax_char,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == end_escape
    ->	Chars=[32,46],
	get_code(Last), Cs1=[Last|Cs]
    ;   read_word_esc(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).
/* ??
read_word_esc(syntax_char,Next,[Next|Chars],Last,[C_mid0|Cs0],Cs):-
    get_code(C_mid0),
    char_type_backslash(C_mid0,C_mid,MidType,Cs0,Cs1),
    (	MidType == end_escape
    ->	Chars=[32,46],
	Last = C_mid, Cs1=Cs
    ;   read_word_esc(MidType,C_mid,Chars,Last,Cs1,Cs)
    ).
*/
skip_space(C_in,C_out,Cs0,Cs) :-
    my_char_type(C_in,Type),
    (	Type == space
    ->	get_code(C_mid),Cs0=[C_mid|Cs1],
        skip_space(C_mid,C_out,Cs1,Cs)
    ;	C_in = C_out, Cs0=Cs
    ).


#sic# :- use_module(library(codesio)).
#sic# read_from_chars(Chars, Term, Options) :-
#sic#     codesio:open_codes_stream(Chars,Stream),
#sic#     call_cleanup(read_term(Stream,Term0,Options),
#sic# 		 close(Stream)),
#sic#     Term=Term0.

add_subs([],L,L).
add_subs([H|T],L0,L):-
    subst_member(L0,H,L1),
    add_subs(T,L1,L).

subst_member([],Pair,[Pair]).
subst_member([H|T],El,List):-
    (	H=El
    ->  T=List
    ;   List=[H|List1],
	subst_member(T,El,List1)
    ).

#sic# dev_run(Dev,Run) :-
#sic#     prolog_flag(system_type,DR),
#sic#     dev_run(DR,Dev,Run).
#sic# 
#sic# dev_run(development,Dev,_):-
#sic#     call(Dev).
#sic# dev_run(runtime,_,Run):-
#sic#     call(Run).
#sic# 

