%%% -*- Module: fsa_compiler; -*-
:- module(fsa_compiler,
	  [ fsa_compile_to_prolog/1,
	    fsa_compile_to_prolog/2,
            fsa_compile_to_prolog_t/2,
            fsa_compile_to_prolog_w/2,
            fsa_compile_to_prolog_wt/2
	  ]).

:- use_module(fsa_util).

#swi# :- style_check([-atom]).

#yap# :- use_module(library(charsio),         [ format_to_chars/3 ]).
#!yap# :- use_module(charsio,         [ format_to_chars/3 ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

user:help_info(class,fsa_compiler,"Prolog Code Generation","\
 \n FSA supports the production of Prolog code on the basis of\
 \n a finite automaton.  Various tricks are employed to make the resulting\
 \n code efficient (rather than readable), but functionality has an ever\
 \n higher priority. The functionality is the same as that provided by the\
 \n fsa_interpreter module, only faster. For pure speed, you should\
 \n consider using the fsa_compiler_to_c module.\
 \n \
 \n Depending on the type of automaton, the compilation provides the\
 \n following Prolog predicate:\
 \n \
 \n * recognizer:\
 \n \
 \n         accepts(?String)\
 \n \
 \n * weighted recognizer:\
 \n \
 \n         w_accepts(+StringIn,?Weight)\
 \n \
 \n * transducer:\
 \n \
 \n         t_accepts(+StringIn,?StringOut)\
 \n \
 \n * weighted transducer:\
 \n \
 \n         wt_accepts(+StringIn,?StringOut,?Weight)\
 \n \
 \n accepts/1 can be used to check a given string for acceptance by the\
 \n automaton, or it can be used to generate all strings accepted by the\
 \n automaton. Since input can be non-deterministic we check for epsilon-cycles\
 \n by keeping track of a list of states visited after last consumption \
 \n of input). There are cases where it would make more sense to \
 \n pre-compute efree automata first. We provide it anyway.\
 \n \
 \n t_accepts/2 can be used to transduce a given string or to\
 \n produce pairs of strings, if the length of the input list is\
 \n known. Since input can be non-deterministic we check for\
 \n epsilon-cycles by keeping track of a list of states visited after last\
 \n consumption of input). The predicate uses special meta-notation |N+|\
 \n for `output' loops, to indicate that last N characters can be repeated\
 \n any number of times. The compilation supports unknown symbols,\
 \n including occurrences of delayed identity constraints (using a queue;\
 \n trick due to Tamas Gaal was explained to me by Lauri Karttunen, Xerox\
 \n Grenoble. Refer to Gerdemann and van Noord's paper in Grammars.\
 \n Try: tminimize({[a:b,?,c],[a,?,d]}).\
 \n \
 \n w_accepts/2 can be used to transduce a given string to the\
 \n corresponding weight. In case of output loops only the minimum weight\
 \n is produced. Since input can be non-deterministic we check for\
 \n epsilon-cycles by keeping track of a list of states visited after last\
 \n consumption of input.\
 \n \
 \n wt_accepts/3 can be used to produce the transduction and weight for\
 \n a given input string, or to generate triples of input, output and\
 \n weight as long as the length of the input string is known. ").

user:help_info(xpred,fsa_compile_to_prolog,
"fsa_compile_to_prolog(+Fa) fsa_compile_to_prolog(+FileIn,+FileOut)","\
 \n In the first variant, a Prolog program is written to standard output\
 \n for the automaton Fa. In the second variant, the Prolog\
 \n program is written to FileOut on the basis of the automaton read in\
 \n from FileIn.  ").

fsa_compile_to_prolog(Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_data:fsa_type(Fa,Type),
    fsa_compile_to_prolog(Type,Fa,Target).

fsa_compile_to_prolog(recognizer,Fa,Target) :-
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog(Fa),Target).
fsa_compile_to_prolog(w_recognizer,Fa,Target) :-
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog_w(Fa),Target).
fsa_compile_to_prolog(transducer(_),Fa,Target) :-
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog_t(Fa),Target).
fsa_compile_to_prolog(w_transducer(_),Fa,Target) :-
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog_wt(Fa),Target).

fsa_compile_to_prolog_t(Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog_t(Fa),Target).


fsa_compile_to_prolog_w(Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog_w(Fa),Target).

fsa_compile_to_prolog_wt(Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog_wt(Fa),Target).


%% utils
list_to_conj([],L,L).
list_to_conj([H|T],Last,(H,R)):-
    list_to_conj(T,Last,R).

fun(Atom,Int,FAtom) :-
    charsio:format_to_chars("~w~w",[Atom,Int],Chars),
    atom_codes(FAtom,Chars).

comment_message(X) :-
    write(
   '%% This file is generated automatically by FSA Utilities.\
 \n %% (c) 1993 - 1999 Gertjan van Noord RUG\
 \n %% More information from http://www.let.rug.nl/~vannoord/Fsa/ \
 \n :- unknown(_,fail).   % because acceptsN/3 where N is a state without\
 \n                       % outgoing transitions will not be defined.\
 \n %% :- use_module(library(lists), [ memberchk/2 ]).\
 \n                       % for loop checking in non-deterministic automata'),nl,
   comment_message_rest(X).

comment_message_rest(r) :-
    format(
"%% accepts(?String) can be used both to recognize a given string\
 \n %% or to produce a string according to Fa. This is why we (possibly)\
 \n %% use dif/2 below. We clearly prefer functionality over efficiency here.\
 \n %% If you didn't know: dif/2 is the SICStus built-in for inequality\
 \n %% which delays until arguments are sufficiently instantiated.~n~n",[]),
    format(
"%% Since input can be non-deterministic we check for epsilon-cycles\
 \n %% by keeping track of a list of states visited after last consumption \
 \n %% of input). There are cases where it would make more sense to \
 \n %% pre-compute efree automata first. We provide it anyway.~n",[]).

comment_message_rest(t) :-
    format(
"%% t_accepts(+In,?Out) can be used to transduce a given string\
 \n %% or to produce pairs of strings, if the length of the input list is\
 \n %% known. This is why we (possibly) use dif/2 below. We prefer \
 \n %% functionality over efficiency here.\
 \n %% If you didn't know: dif/2 is the SICStus built-in for inequality\
 \n %% which delays until arguments are sufficiently instantiated.~n~n",[]),
    format(
"%% Since input can be non-deterministic we check for epsilon-cycles\
 \n %% by keeping track of a list of states visited after last consumption \
 \n %% of input). This technique sometimes lead to spurious `_ is C+1' calls\
 \n %% below.~n~n",[]),
    format(
"%% Uses special meta-notation |N+| for `output' loops,\
 \n %% to indicate that last N characters can be repeated any number of\
 \n %% times.~n~n",[]),
    format(
"%% Supports unknown symbols ($? and $@), including occurrences of\
 \n %% `delayed $@', using a queue! Trick was explained to me by\
 \n %% Lauri Karttunen, Xerox Grenoble. I have never seen it described\
 \n %% in the literature. E.g. try regex tminimize({[a:b,?,c],[a,?,d]})~n",[]).

comment_message_rest(wt) :-
    format(
"%% wt_accepts(+In,?Out,?Weight) can be used to transduce a given string\
 \n %% or to produce triples of strings and weight, if length of the input list is\
 \n %% known. This is why we (possibly) use dif/2 below. We prefer \
 \n %% functionality over efficiency here.\
 \n %% If you didn't know: dif/2 is the SICStus built-in for inequality\
 \n %% which delays until arguments are sufficiently instantiated.~n~n",[]),
    format(
"%% Since input can be non-deterministic we check for epsilon-cycles\
 \n %% by keeping track of a list of states visited after last consumption \
 \n %% of input). This technique sometimes lead to spurious `_ is C+1' calls\
 \n %% below.~n~n",[]),
    format(
"%% Uses special meta-notation |N+| for `output' loops,\
 \n %% to indicate that last N characters can be repeated any number of\
 \n %% times.~n~n",[]),
    format(
"%% Supports unknown symbols ($? and $@), including occurrences of\
 \n %% `delayed $@', using a queue! Trick was explained to me by\
 \n %% Lauri Karttunen, Xerox Grenoble. I have never seen it described\
 \n %% in the literature. E.g. try regex tminimize({[a:b,?,c],[a,?,d]})~n",[]).

comment_message_rest(w) :-
    format(
"%% w_accepts(+String,?Weight) can be used to transduce a given String\
 \n %% to the corresponding Weight. In case of output loops only the minimum\
 \n %% weight is produced.\
 \n \
 \n %% Since input can be non-deterministic we check for epsilon-cycles\
 \n %% by keeping track of a list of states visited after last consumption \
 \n %% of input).~n",[]).




%% here we go...
fsa_compile_to_prolog(Fa) :-
    comment_message(r),
    (	fsa_data:fsa_start_state(Fa,X),
	fun(accepts,X,Fun),
	Body =.. [Fun,S,[X]],
	portray_clause((accepts(S) :- Body)),
	fail
    ;	true
    ),
    (	fsa_data:fsa_final_state(Fa,X),
	fun(accepts,X,Fun),
	Head =.. [Fun,[],_],
	portray_clause(Head),
	fail
    ;	true
    ),
    (	fsa_data:fsa_state(Fa,S),
	fun(accepts,S,Fun),
	Head =.. [Fun,[Sym|T],L],
	Body =.. [Fun,Sym,T,L],
	portray_clause((Head :- Body)),
	fail
    ;	true
    ),
    (	fsa_interpreter:
            apply_transition(Fa,State0,State,S0,S,States0,States,Goals,[]),
	fun(accepts,State0,F0),
	fun(accepts,State,F),
	(   nonvar(S0)
	->  S0=[S0h|S0t],
	    Head =.. [F0,S0h,S0t,States0]
	;   Head =.. [F,S0,States0]
	),
	fun(accepts,State,F),
	Lgoal =.. [F,S,States],
	list_to_conj(Goals,Lgoal,Body),
	portray_clause((Head :- Body)),
	fail
    ;   true
    ).

fsa_compile_to_prolog_t(Fa0) :-
    fsa_regex:rx(sequence_transducer(transducer(fa(Fa0),error),silent),Fa),
    comment_message(t),
    (	fsa_data:fsa_start_state(Fa,X),
	fun(t_accepts,X,Fun),
	fsa_interpreter:empty_q(Q),
	Body =.. [Fun,S,T,[],[X],0,Q],
        portray_clause((t_accepts(S,T) :- Body)),
	fail
    ;	true
    ),
    (	fsa_data:fsa_state(Fa,S),
	fun(t_accepts,S,Fun),
	Head =.. [Fun,[Sym|T],O,B,C,D,E],
	Body =.. [Fun,Sym,T,O,B,C,D,E],
	portray_clause((Head :- Body)),
	fail
    ;	true
    ),
    (	fsa_data:fsa_final_state(Fa,X),
	fun(t_accepts,X,Fun),
	Head =.. [Fun,[],[],_,_,_,_],
	portray_clause(Head),
	fail
    ;	true
    ),
    (	fsa_interpreter:apply_t_transition(
	    St0,St,S0,S,T0,T,Fa,Sts0,Sts,B0,B,L0,L,Q0,Q,G0,[]),
	fun(t_accepts,St0,F0),
	(   nonvar(S0)
	->  S0=[S0h|S0t],
	    Head =.. [F0,S0h,S0t,T0,B0,Sts0,L0,Q0]
	;   Head =.. [F0,S0,T0,B0,Sts0,L0,Q0]
	),
	fun(t_accepts,St,F),
	Lgoal =.. [F,S,T,B,Sts,L,Q],
	list_to_conj(G0,Lgoal,Body),
	portray_clause((Head :- Body)),
	fail
    ;   true
    ).

fsa_compile_to_prolog_wt(Fa0) :-
    fsa_regex:rx(weighted_sequence_transducer(fa(Fa0)),Fa),
    comment_message(wt),
    (	fsa_data:fsa_start_state(Fa,X),
	fun(wt_accepts,X,Fun),
	fsa_interpreter:empty_q(Q),
	Body =.. [Fun,S,T,[],[X],0,Q,0,Weight],
        portray_clause((wt_accepts(S,T,Weight) :- Body)),
	fail
    ;	true
    ),
    (	fsa_data:fsa_state(Fa,S),
	fun(wt_accepts,S,Fun),
	Head =.. [Fun,[Sym|T],O,B,C,D,E,W0,W],
	Body =.. [Fun,Sym,T,O,B,C,D,E,W0,W],
	portray_clause((Head :- Body)),
	fail
    ;	true
    ),
    (	fsa_data:fsa_final_state(Fa,X),
	fun(wt_accepts,X,Fun),
	Head =.. [Fun,[],[],_,_,_,_,W,W],
	portray_clause(Head),
	fail
    ;	true
    ),
    (	fsa_interpreter:apply_wt_transition(
	    St0,St,S0,S,T0,T,Fa,Sts0,Sts,B0,B,L0,L,Q0,Q,G0,[],W0,W1),
	fun(wt_accepts,St0,F0),
	(   nonvar(S0)
	->  S0=[S0h|S0t],
	    Head =.. [F0,S0h,S0t,T0,B0,Sts0,L0,Q0,W0,W]
	;   Head =.. [F0,S0,T0,B0,Sts0,L0,Q0,W0,W]
	),
	fun(wt_accepts,St,F),
	Lgoal =.. [F,S,T,B,Sts,L,Q,W1,W],
	list_to_conj(G0,Lgoal,Body),
	portray_clause((Head :- Body)),
	fail
    ;   true
    ).

fsa_compile_to_prolog_w(Fa0) :-
    fsa_regex:rx(weighted_recognizer(fa(Fa0)),Fa),
    comment_message(w),
    (	fsa_data:fsa_start_state(Fa,X),
	fun(w_accepts,X,Fun),
	Body =.. [Fun,S,0,T,[X]],
	portray_clause((w_accepts(S,T) :- Body)),
	fail
    ;	true
    ),
    (	fsa_data:fsa_final_state(Fa,X),
	fun(w_accepts,X,Fun),
	Head =.. [Fun,[],S,S,_],
	portray_clause(Head),
	fail
    ;	true
    ),
    (	fsa_data:fsa_state(Fa,S),
	fun(w_accepts,S,Fun),
	Head =.. [Fun,[Sym|T],X0,X,Sts],
	Body =.. [Fun,Sym,T,X0,X,Sts],
	portray_clause((Head :- Body)),
	fail
    ;	true
    ),  %% to do: use fsa_interpreter here...
    (	fsa_interpreter:apply_w_transition(
	    St0,St,S0,S,T0,T1,Fa,Sts0,Sts,G0,[]),
	fun(w_accepts,St0,F0),
	(   nonvar(S0)
	->  S0=[S0h|S0t],
	    Head =.. [F0,S0h,S0t,T0,T1,Sts0]
	;   Head =.. [F0,S0,T0,T1,Sts0]
	),
	fun(w_accepts,St,F),
	Lgoal =.. [F,S,T1,T,Sts],
	list_to_conj(G0,Lgoal,Body),
	portray_clause((Head :- Body)),
	fail
    ;	true
    ).

