:- module( fsa_data,
    [ fsa_symbols_decl/2,
      fsa_symbols_set/2,
      fsa_symbols_number/2,
      fsa_state/2,
      fsa_states_number/2,
      fsa_states_set/2,
      fsa_start_state/2,
      fsa_start_states/2,
      fsa_final_state/2,
      fsa_final_states/2,
      fsa_transition/4,
      fsa_transitions/2,
      fsa_jump/3,
      fsa_jumps/2,
      fsa_type/2,
      fsa_components/7,                  % fsa_components
      fsa_construct/5,
      fsa_construct/6,
      fsa_construct/7,
      fsa_construct_rename_states/5,
      fsa_construct_rename_states/6,
      fsa_copy_except/5,
      fsa_count_file/2,
      fsa_count/1,
      fsa_count/2,
      fsa_density_file/2,
      fsa_density/1,
      fsa_density/2,
      fsa_identical/2,
      fsa_identical_files_report/2
    ] ).

#swi# :- style_check([-atom]).
#!swi# :- use_module(library(ordsets), [ ord_member/2 ]).
#swi# :- use_module(ordsets, [ ord_member/2 ]).
#yap# :- use_module(library(lists),   [ member/2, memberchk/2, append/3 ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

user:help_info(io,normal,"Internal Format of Finite Automata",
"This module provides a consistent interface to the internal format of\
 \n finite automata. A finite automaton is a term \
 \n \
 \n     fa(Symbols,States,Starts,Finals,Transs,Jumps)\
 \n \
 \n Symbols is a term r(Sig) (for recognizers) or t(SigD,SigR) for\
 \n transducers. Weighted automata have r(Sig,fsa_frozen) and\
 \n t(SigD,SigR,fsa_frozen). Here Sig, SigD, SigR are the predicate modules.\
 \n \
 \n States is an integer indicating the number of states in the automaton.\
 \n   \
 \n Starts is an ordered list of integers indicating the start states of\
 \n the automaton. \
 \n \
 \n Finals is an ordered list of integers indicating the final states of\
 \n the automaton. \
 \n \
 \n Transs is an ordered list of triples trans(A,B,C) where A and C are\
 \n integers indicating source and target state, and B is the symbol\
 \n part. The symbol part is different for the different types of \
 \n automata:\
 \n \
 \n recognizers: P where P is a predicate\
 \n \
 \n weighted recognizers: P/W where P is a predicate and W is a number\
 \n \
 \n letter transducer: P/Q where P and Q is a predicate or the empty list\
 \n \
 \n sequence transducer: P/Q where P is a predicate or the empty list, and\
 \n Q is a list of predicates\
 \n \
 \n weighted letter transducer: P/(Q/W) where P and Q is a predicate or \
 \n the empty list, and W is a number\
 \n \
 \n weighted sequence transducer: P/(Q/W) where P is a predicate or the empty \
 \n list, and Q is a list of predicates\
 \n \
 \n In addition, transducers allow a term '$@'(P) anywhere where a predicate\
 \n is allowed (P a predicate).\
 \n \
 \n Jumps is an ordered list of pairs jump(A,B) where A and B are integers\
 \n indicating source and target state. This implies there is an epsilon\
 \n transition from A to B.").


%%%%%%%%%%%%%%%%%
%%%% symbols %%%%
%%%%%%%%%%%%%%%%%

%%  the data-structure that represents everything to do with symbols
fsa_symbols_decl(fa(Symbols,_,_,_,_,_),Symbols).

user:help_info(xpred,fsa_states_number,"fsa_states_number(?Fa,?Integer)",
"The number of states in Fa is Integer.").
fsa_states_number(fa(_,States,_,_,_,_),States).

user:help_info(xpred,fsa_states_set,"fsa_states_set(+Fa,?States)",
	  "States is an ordered list of integers: all states in Fa. ").
fsa_states_set(Fa,AllStates):-
    fsa_states_number(Fa,States),
    generate_list(0,States,AllStates).

generate_list(N0,N,States) :-
    (	N0=:=N        %% number of states is 1+largest state number
    ->  States=[]
    ;   States=[N0|States0],
	N1 is N0+1, 
	generate_list(N1,N,States0) 
    ).

user:help_info(xpred,fsa_state,"fsa_state(+Fa,?State)",
	  "State is a state in Fa.").
fsa_state(Fa,State) :-
    %% if there are 10 states, then there are states 0 .. 9
    fsa_states_number(Fa,States0),
    States is States0-1,
    fsa_util:between(0,States,State).

user:help_info(xpred,fsa_start_states,"fsa_start_states(?Fa,?StartStates)",
	  "StartStates is the ordered list of start states of Fa. ").
fsa_start_states(fa(_,_,Starts,_,_,_),Starts).

user:help_info(xpred,fsa_start_state,"fsa_start_state(+Fa,?StartState)",
	  "StartState is a start states of Fa. ").
fsa_start_state(Fa,Start) :-
    fsa_start_states(Fa,Starts),
    (	nonvar(Start)
    ->  ord_member(Start,Starts)
    ;   member(Start,Starts)
    ).

user:help_info(xpred,fsa_final_states,"fsa_final_states(?Fa,?FinalStates)",
 	  "FinalStates is the ordered list of final states of Fa. ").
fsa_final_states(fa(_,_,_,Finals,_,_),Finals).

user:help_info(xpred,fsa_final_state,"fsa_final_state(+Fa,?FinalState)",
	  "FinalState is a final states of Fa. ").
fsa_final_state(Fa,Final) :-
    fsa_final_states(Fa,Finals),
    (	integer(Final)
    ->  ord_member(Final,Finals)
    ;	member(Final,Finals)
    ).

user:help_info(xpred,fsa_transitions,"fsa_transitions(?Fa,?Trans)",
	  "Trans is the ordered list of transitions of Fa.").
fsa_transitions(fa(_,_,_,_,Transs,_),Transs).

user:help_info(xpred,fsa_transition,"fsa_transition(+Fa,?P,?Sym,?Q)",
	  "In Fa there is a transition from P to Q with symbol(pair) Sym.").
fsa_transition(Fa,P,Sym,Q) :-
    fsa_transitions(Fa,Transs),
    (	ground(trans(P,Sym,Q))
    ->  ord_member(trans(P,Sym,Q),Transs)
    ;	member(trans(P,Sym,Q),Transs)
    ).

    
user:help_info(xpred,fsa_jumps,"fsa_jumps(?Fa,?Jumps)",
	  "Jumps is the ordered list of jumps of Fa.").
fsa_jumps(fa(_,_,_,_,_,Jumps),Jumps).

user:help_info(xpred,fsa_jump,"fsa_jump(+Fa,?P,?Q)",
	  "In Fa there is a jump from P to Q.").
fsa_jump(Fa,P,Q) :-
    fsa_jumps(Fa,Jumps),
    (	ground(jump(P,Q))
    ->  ord_member(jump(P,Q),Jumps)
    ;	member(jump(P,Q),Jumps)
    ).
    
user:help_info(xpred,fsa_construct,
"fsa_construct([[+Symbols,]+NumberStates,]+Starts,+Finals,+Trans,+Jumps,-Fa)","\
 \n Predicate to construct a finite automaton on the basis of lists of\
 \n start states, final states, transitions and jumps. These lists need\
 \n not be ordered. It's somewhat more efficient to specify the number\
 \n of states, if known. It's even more efficient if you also know the\
 \n symbols data-structure you want for Fa. "). 

%% this needs argument telling it *what type* of thing to construct
fsa_construct(Starts0,Finals0,Transs0,Jumps0,FsaOut) :-
    sort(Starts0,Starts),
    fsa_start_states(FsaOut,Starts),
    sort(Finals0,Finals),
    fsa_final_states(FsaOut,Finals),
    sort(Transs0,Transs),
    fsa_transitions(FsaOut,Transs),
    sort(Jumps0,Jumps),
    fsa_jumps(FsaOut,Jumps),
    collect_states_and_symbols(Starts,Finals,Transs,Jumps,Length,Syms),
    fsa_states_number(FsaOut,Length),
    fsa_symbols_decl(FsaOut,Syms).

%% this needs argument telling it *what type* of thing to construct
fsa_construct(Length,Starts0,Finals0,Transs0,Jumps0,FsaOut) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    sort(Starts0,Starts),
    fsa_start_states(FsaOut,Starts),
    sort(Finals0,Finals),
    fsa_final_states(FsaOut,Finals),
    sort(Transs0,Transs),
    fsa_transitions(FsaOut,Transs),
    sort(Jumps0,Jumps),
    fsa_jumps(FsaOut,Jumps),
    fsa_states_number(FsaOut,Length),
    (   Transs=[trans(_,_/_,_)|_]
    ->  Sigma=t(Module,Module)
    ;   Sigma=r(Module)
    ),
    fsa_symbols_decl(FsaOut,Sigma).

user:help_info(xpred,fsa_components,"fsa_components(?Symbols,?Length,?Starts,?Finals,?Trans,?Jumps,?Fa)",
"Predicate to construct an automaton on the basis of its components, \
 \n or to query the components of a given automaton. The difference with \
 \n fsa_construct/7 is that Starts, Finals, Trans and Jumps must be sorted \
 \n already. ").
fsa_components(Symbols,Length,Starts,Finals,Transs,Jumps,
       fa(Symbols,Length,Starts,Finals,Transs,Jumps)).


fsa_construct(Symbols,Length,Starts0,Finals0,Transs0,Jumps0,FsaOut) :-
    sort(Starts0,Starts),
    fsa_start_states(FsaOut,Starts),
    sort(Finals0,Finals),
    fsa_final_states(FsaOut,Finals),
    sort(Transs0,Transs),
    fsa_transitions(FsaOut,Transs),
    sort(Jumps0,Jumps),
    fsa_jumps(FsaOut,Jumps),
    fsa_states_number(FsaOut,Length),
    fsa_symbols_decl(FsaOut,Symbols).

user:help_info(xpred,fsa_construct_rename_states,
"fsa_construct_rename_states([+Symbols,]+Starts,+Finals,+Trans,+Jumps,-Fa)","\
 \n Predicate to construct a finite automaton on the basis of lists of\
 \n start states, final states, transitions and jumps. These lists need\
 \n not be ordered. Moreover, state names can be arbitrary Prolog\
 \n terms. These state names will be renamed to integers. Symbol list is\
 \n computed on the basis of Trans. Sigma is determined by the current\
 \n default predicate module (i.e. by flag *pred_module*).. It's more\
 \n efficient if you also know the symbols data-structure you want for\
 \n Fa. Some checking on these symbols is performed nevertheless."). 

%% version of fsa_construct in which state names are perhaps not integers yet.
fsa_construct_rename_states(Starts0,Finals0,Transitions0,Jumps0,FsaOut) :-
    fsa_hash:fsa_hash_new(Hash),
    rename_states(Starts0,Starts,Hash,0,Len0),
    rename_states(Finals0,Finals,Hash,Len0,Len1),
    rename_items(Transitions0,Transitions,Hash,Len1,Len2),
    rename_items(Jumps0,Jumps,Hash,Len2,Len),
    fsa_construct(Len,Starts,Finals,Transitions,Jumps,FsaOut).

%% version of fsa_construct in which state names are perhaps not integers yet.
fsa_construct_rename_states(Symbols,Starts0,Finals0,Transitions0,Jumps0,FsaOut) :-
    fsa_hash:fsa_hash_new(Hash),
    rename_states(Starts0,Starts,Hash,0,Len0),
    rename_states(Finals0,Finals,Hash,Len0,Len1),
    rename_items(Transitions0,Transitions,Hash,Len1,Len2),
    rename_items(Jumps0,Jumps,Hash,Len2,Len),
    fsa_construct(Symbols,Len,Starts,Finals,Transitions,Jumps,FsaOut).

rename_states([],[],_,L,L).
rename_states([H0|T0],[H|T],Hash,Len0,Len) :-
    rename_state(H0,H,Hash,Len0,Len1),
    rename_states(T0,T,Hash,Len1,Len).

rename_items([],[],_,L,L).
rename_items([H0|T0],[H|T],Hash,Len0,Len) :-
    rename_item(H0,H,Hash,Len0,Len1),
    rename_items(T0,T,Hash,Len1,Len).

rename_item(trans(A0,B,C0),trans(A,B,C),Hash,Len0,Len) :-
    rename_state(A0,A,Hash,Len0,Len1),
    rename_state(C0,C,Hash,Len1,Len).

rename_item(jump(A0,C0),jump(A,C),Hash,Len0,Len) :-
    rename_state(A0,A,Hash,Len0,Len1),
    rename_state(C0,C,Hash,Len1,Len).

rename_state(A0,A,Hash,Len0,Len) :-
    fsa_hash:fsa_hash_access(A0,A,Hash),
    (	nonvar(A)
    ->	Len0=Len
    ;	Len0=A,
	Len is Len0+1
    ).

collect_states_and_symbols(Starts,Finals,Trans,Jumps,Nr,RT) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    (   Trans=[trans(_,_/_,_)|_]
    ->  RT=t(Module,Module),
	collect_states(Starts,Finals,Trans,Jumps,Nr)
    ;   RT=r(Module),
	collect_states(Starts,Finals,Trans,Jumps,Nr)
    ).

collect_states(Starts,Finals,Trans,Jumps,Nr) :-
    collect_states0(Starts,1,Nr0),
    collect_states0(Finals,Nr0,Nr1),
    collect_states_tr(Trans,Nr1,Nr2),
    collect_states_j(Jumps,Nr2,Nr).  % Nr is 1+max(state)

collect_states0([],N,N).
collect_states0([H|T],N0,N) :-
    add_state(H,N0,N1),
    collect_states0(T,N1,N).

collect_states_tr([],N,N).
collect_states_tr([trans(A,_Sym,C)|T],N0,N) :-
    add_state(A,N0,N1),
    add_state(C,N1,N2),
    collect_states_tr(T,N2,N).

collect_states_j([],N,N).
collect_states_j([jump(A,C)|T],N0,N) :-
    add_state(A,N0,N1),
    add_state(C,N1,N2),
    collect_states_j(T,N2,N).

add_state(State,N0,N) :-
    (   State >= N0
    ->  N is State+1
    ;   N0=N
    ).

check_fa(Fa0,Fa):-
    (	fa_type(Fa0,FaType),
	check_fa(FaType,Fa0,Fa)
    ->	true
    ;	user:throw(fsa_error(
      "Huh? What's this? It doesn't look like a finite automaton:~n~w~n",
			      [Fa0]))
    ).

fa_type(fa(_,_,_,_,_,_,_),fsa3).
fa_type(fa(A,_,_,_,_,_),Type):-
       fa_type0(A,Type).

fa_type0([],fsa4).
fa_type0([_|_],fsa4).
fa_type0(s(_,_),fsa5).
fa_type0(r(_),fsa6).
fa_type0(r(_,_),fsa6).
fa_type0(t(_,_),fsa6).
fa_type0(t(_,_,_),fsa6).

check_fa(fsa3,fa(A,B,C,D,E,F,_),Fa):-
    check_fa(fsa4,fa(A,B,C,D,E,F),Fa).

check_fa(fsa4,Fa0,Fa) :-
    (	fsa_transition(Fa0,_,_/_,_)
    ->  fsa_copy_except(symbols,Fa0,Fa1,List,s(List,r(open)))
    ;   fsa_copy_except(symbols,Fa0,Fa1,List,s(List,t(open,open)))
    ),
    check_fa(fsa5,Fa1,Fa).

check_fa(fsa5,Fa0,Fa) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    fsa_copy_except(symbols,Fa0,Fa1,s(List,RT0),RT),
    convert_rt(RT0,RT,Module),
    fsa_type(Fa1,Type),
    fsa_copy_except(transitions,Fa1,Fa2,Trans0,Trans),
    convert_trans(Trans0,Trans1,Type,Module,List),
    sort(Trans1,Trans),    
    check_fa(fsa6,Fa2,Fa).

check_fa(fsa6,Fa,Fa).

convert_rt(r(_),r(M),M).
convert_rt(t(_,_),t(M,M),M).

convert_trans([],[],_,_,_).
convert_trans([trans(P,Sym0,Q)|Trans0],[trans(P,Sym,Q)|Trans],Type,Module,List):-
    convert_sym(Type,Sym0,Sym,Module,List),
    convert_trans(Trans0,Trans,Type,Module,List).

convert_sym(recognizer,Sym0,Sym,Module,List) :-
    convert_sym(Sym0,Sym,Module,List).
convert_sym(transducer(Sub),Sym0,Sym,Module,List) :-
    convert_sym(Sub,Sym0,Sym,Module,List).

convert_sym(letter,$@ / $@,'$@'(Sym)/'$@'(Sym),Module,List) :-
    !,
    convert_sym($? ,Sym,Module,List).

convert_sym(letter,Syma0/Symb0,Syma/Symb,Module,List) :-
    (	Syma0==[]
    ->	Syma=[]
    ;	convert_sym(Syma0,Syma,Module,List)
    ),
    (	Symb0==[]
    ->	Symb=[]
    ;	convert_sym(Symb0,Symb,Module,List)
    ).

convert_sym(sequence,Syma0/List0,Syma/List,Module,SList) :-
    (	Syma0==[]
    ->	Syma=[]
    ;	convert_sym(Syma0,Syma,Module,SList)
    ),
    convert_sym_list(List0,List,Module,SList).

convert_sym_list([],[],_,_).
convert_sym_list([H0|T0],[H|T],Module,List) :-
    convert_sym(H0,H,Module,List),
    convert_sym_list(T0,T,Module,List).

convert_sym(Sym0,Sym,Module,List) :-
    (	Sym0 == $?
    ->	negate(List,Sym,Module)
    ;	fsa_pred_module:regex_notation_to_predicate(Module,Sym0,Sym)
    ).

negate(List,Sym,Module) :-
    fsa_pred_module:true(Module,True),
    negate(List,True,Sym,Module).

negate([],T,T,_).
negate([H|T],P0,P,Module):-
    (	H == $?
    ->  P1=P0
    ;	H == $@
    ->  P1=P0
    ;   fsa_pred_module:negation(Module,H,Neg),
	fsa_pred_module:conjunction(Module,Neg,P0,P1)
    ),
    negate(T,P1,P,Module).

user:help_info(xpred,fsa_copy_except,"fsa_copy_except(+Key,?Fa0,?Fa1,?Part0,?Part1)","\
 \n This predicate unifies Fa0 and Fa1 except for the part specified by\
 \n Key. Part must be one of the atoms symbols, states, \
 \n start_states, final_states, transitions, jumps. Part0 and Part1 \
 \n are the corresponding parts in Fa0 and Fa1."). 
fsa_copy_except(symbols,         fa(A0,B,C,D,E,F),fa(A,B,C,D,E,F),A0,A).
fsa_copy_except(states,          fa(A,B0,C,D,E,F),fa(A,B,C,D,E,F),B0,B).
fsa_copy_except(start_states,    fa(A,B,C0,D,E,F),fa(A,B,C,D,E,F),C0,C).
fsa_copy_except(final_states,    fa(A,B,C,D0,E,F),fa(A,B,C,D,E,F),D0,D).
fsa_copy_except(transitions,     fa(A,B,C,D,E0,F),fa(A,B,C,D,E,F),E0,E).
fsa_copy_except(jumps,           fa(A,B,C,D,E,F0),fa(A,B,C,D,E,F),F0,F).

user:help_info(xpred,fsa_type,"fsa_type(+Fa,?Type)","\
 \n Type is the type of the automaton Fa, where type is one of recognizer,\
 \n w_recognizer, transducer(Sub), w_transducer(Sub). Sub is one of\
 \n letter or sequence."). 
fsa_type(Fa,Type) :-
    fsa_symbols_decl(Fa,Sigma),
    subtype(Sigma,Fa,Type).

subtype(r(_),     _,recognizer).
subtype(r(_,_),   _,w_recognizer).

subtype(t(_,_),Fa,transducer(SubType)):-
    fsa_transitions(Fa,Trans),
    subtype_trs(Trans,SubType).

subtype(t(_,_,_),Fa,w_transducer(SubType)) :-
    fsa_transitions(Fa,Trans),
    subtype_trs_w(Trans,SubType).
    
subtype_trs([],letter).
subtype_trs([trans(_,_/List,_)|Tail],SubType) :-
    (	List == []
    ->	subtype_trs(Tail,SubType)
    ;	List=[_|_]
    ->	SubType=sequence
    ;   SubType=letter
    ).

subtype_trs_w([],letter).
subtype_trs_w([trans(_,_/(List/_),_)|Tail],SubType) :-
    (	List == []
    ->	subtype_trs_w(Tail,SubType)
    ;	List=[_|_]
    ->	SubType=sequence
    ;   SubType=letter
    ).


%% fsa_symbols_set(+Fa,?Syms)
fsa_symbols_set(Fa,Syms) :-
    fsa_transitions(Fa,Trans),
    fsa_type(Fa,Type),
    transition_symbols(Trans,Type,Syms0),
    sort(Syms0,Syms).    

transition_symbols([],_,[]).
transition_symbols([trans(_,S,_)|Tr],Type,Syms):-
    transition_symbol(Type,S,Syms,Syms1),
    transition_symbols(Tr,Type,Syms1).

transition_symbol(recognizer,S,[S|Syms],Syms).
transition_symbol(w_recognizer,S/T,[T|Syms0],Syms) :-
    (	S == []
    ->	Syms0=Syms
    ;   Syms0=[S|Syms]
    ).
transition_symbol(transducer(Sub),Sa/Sb,Syms0,Syms):-
    transition_symbol(Sub,Sa,Sb,Syms0,Syms).
transition_symbol(w_transducer(Sub),Sa/(Sb/Sc),[Sc|Syms0],Syms):-
    transition_symbol(Sub,Sa,Sb,Syms0,Syms).

transition_symbol(letter,Sa,Sb,Syms0,Syms):-
    (	Sa==[]
    ->	Syms0=Syms1
    ;   Syms0=[Sa|Syms1]
    ),
    (	Sb==[]
    ->	Syms1=Syms
    ;   Syms1=[Sb|Syms]
    ).
transition_symbol(sequence,Sa,Sb,Syms0,Syms):-
    (	Sa==[]
    ->	Syms0=Syms1
    ;   Syms0=[Sa|Syms1]
    ),
    append(Sb,Syms,Syms1).

%% fsa_symbols_number(+Fa,?Integer)
fsa_symbols_number(Fa,No):-
    fsa_symbols_set(Fa,Syms),
    length(Syms,No).

index_jumps(Fa,JT):-
    fsa_states_number(Fa,Size),
    fsa_jumps(Fa,Jumps),
    fsa_array:fsa_array_new(JT,Size),
    jumps_to_thing(Jumps,JT).

jumps_to_thing([],_).
jumps_to_thing([jump(P,Q)|Tail],Thing):-
    fsa_array:fsa_array_access(P,[Q|Targets],Thing),
    jumps_to_thing1(Tail,P,Targets,Thing).

jumps_to_thing1([],_,[],_).
jumps_to_thing1([jump(P0,Q)|Tail],P,Targets,Thing) :-
    (	P0==P
    ->	Targets=[Q|Targets1],
	jumps_to_thing1(Tail,P,Targets1,Thing)
    ;	Targets=[],
	fsa_array:fsa_array_access(P0,[Q|Targets1],Thing),
	jumps_to_thing1(Tail,P0,Targets1,Thing)
    ).

index_finals(Fa,Table):-
    fsa_states_number(Fa,Size),
    fsa_final_states(Fa,Finals),
    fsa_array:fsa_array_new(Table,Size),
    finals_to_thing(Finals,Table).

index_starts(Fa,Table):-
    fsa_states_number(Fa,Size),
    fsa_start_states(Fa,Finals),
    fsa_array:fsa_array_new(Table,Size),
    starts_to_thing(Finals,Table).

finals_to_thing([],_).
finals_to_thing([H|T],Table):-
    fsa_array:fsa_array_access(H,[],Table),
    finals_to_thing(T,Table).

starts_to_thing([],_).
starts_to_thing([H|T],Table):-
    fsa_array:fsa_array_access(H,[],Table),
    starts_to_thing(T,Table).

index_transitions(Fa,Table) :-
    fsa_states_number(Fa,Size),
    fsa_transitions(Fa,Trans),
    fsa_array:fsa_array_new(Table,Size),
    transitions_to_thing(Trans,Table).

%% like transitions_to_graph, but we don't construct a graph,
%% but a Thing directly.
transitions_to_thing([],_).
transitions_to_thing([trans(P,Sym,Q)|Tail],Thing) :-
    fsa_array:fsa_array_access(P,[Sym-[Q|Targets]|Syms],Thing),
    transitions_to_thing1(Tail,P,Sym,Targets,Syms,Thing).

transitions_to_thing1([],_,_,[],[],_).
transitions_to_thing1([trans(P0,Sym0,Q)|Tail],P,Sym,Targets,Syms,Thing):-
    (	P0 == P
    ->	transitions_to_thing11(P,Sym0,Sym,Q,Tail,Targets,Syms,Thing)
    ;	Targets=[],
	Syms=[],
	fsa_array:fsa_array_access(P0,[Sym0-[Q|Targets1]|Syms1],Thing),
	transitions_to_thing1(Tail,P0,Sym0,Targets1,Syms1,Thing)
    ).

transitions_to_thing11(P,Sym0,Sym,Q,Tail,Targets,Syms,Thing) :-
    (	Sym0==Sym
    ->	Targets=[Q|Targets1],
	transitions_to_thing1(Tail,P,Sym,Targets1,Syms,Thing)
    ;	Targets=[],
	Syms=[Sym0-[Q|Targets1]|Syms1],
	transitions_to_thing1(Tail,P,Sym0,Targets1,Syms1,Thing)
    ).

%% indexed_jump(+State,+JT,-Target)
%% generates all jumps originating from State
indexed_jump(State,JT,Target):-
    fsa_array:fsa_array_access(State,TargetList,[],JT),
    (	ground(Target)
    ->	memberchk(Target,TargetList)
    ;   member(Target,TargetList)
    ).

%% indexed_transition(+State,+TT,-Sym,-Target)
%% generates all transitions originating from State
indexed_transition(State,TT,Sym,Target):-
    fsa_array:fsa_array_access(State,Keylist,[],TT),
    (	ground(Sym)
    ->	memberchk(Sym-TargetList,Keylist)
    ;   member(   Sym-TargetList,Keylist)
    ),
    (	ground(Target)
    ->	memberchk(Target,TargetList)
    ;   member(Target,TargetList)
    ).

%% indexed_final(+State,+FT,?YesNo)
indexed_final(State,FT,Bool):-
    fsa_array:fsa_array_access(State,Val,FT),
    (	var(Val)
    ->	Bool=no
    ;   Bool=yes
    ).

%% indexed_start(+State,+FT,?YesNo)
indexed_start(State,FT,Bool):-
    fsa_array:fsa_array_access(State,Val,FT),
    (	var(Val)
    ->	Bool=no
    ;   Bool=yes
    ).

:- public index_inverted_transitions/2, index_inverted_jumps/2.
index_inverted_transitions(Fa,Thing):-
    fsa_states_number(Fa,Size),
    fsa_transitions(Fa,Trans0),
    inverse_trans(Trans0,Trans1),
    sort(Trans1,Trans),
    fsa_array:fsa_array_new(Thing,Size),
    transitions_to_thing(Trans,Thing).

inverse_trans([],[]).
inverse_trans([trans(P,Sym,Q)|T0],[trans(Q,Sym,P)|T]):-
    inverse_trans(T0,T).

index_inverted_jumps(Fa,Thing):-
    fsa_states_number(Fa,Size),
    fsa_jumps(Fa,Jumps0),
    inverse_jumps(Jumps0,Jumps1),
    sort(Jumps1,Jumps),
    fsa_array:fsa_array_new(Thing,Size),
    jumps_to_thing(Jumps,Thing).

inverse_jumps([],[]).
inverse_jumps([jump(P,Q)|T0],[jump(Q,P)|T]):-
    inverse_jumps(T0,T).

fsa_count_file(File0,File) :-
    fsa_io:fsa_read_file(File0,FaIn),
    fsa_count(File,FaIn).

fsa_count(File,FaIn) :-
    fsa_util:to_file(fsa_data:fsa_count(FaIn),File).

fsa_count(FaIn):-
    fsa_globals:fsa_global_get(count,OnOff),
    fsa_count0(OnOff,FaIn).

fsa_count_data(Fa,[StatesC,StartsC,FinalsC,TransitionsC,JumpsC,SymbolsC]) :-
    no_states(Fa,StatesC),
    no_symbols(Fa,SymbolsC),
    no_starts(Fa,StartsC),
    no_finals(Fa,FinalsC),
    no_trans(Fa,TransitionsC),
    no_jumps(Fa,JumpsC).

no_states(Fa,"%% ~w states; "/[Count]) :-
    fsa_states_number(Fa,Count).

no_finals(Fa,"~w finals; "/[Count]) :-
    fsa_final_states(Fa,FS),
    length(FS,Count).

no_starts(Fa,"~w starts; "/[Count]) :-
    fsa_start_states(Fa,FS),
    length(FS,Count).

no_trans(Fa,"~w transitions; "/[Count]) :-
    fsa_transitions(Fa,Ts),
    length(Ts,Count).

no_jumps(Fa,"~w jumps; "/[Count]) :-
    fsa_jumps(Fa,Js),
    length(Js,Count).

no_symbols(Fa,"~w symbols~n"/[Count]) :-
    fsa_symbols_number(Fa,Count).

fsa_count0(long,FaIn):-
    fsa_count_data(FaIn,Counts),
    format_counts_l(Counts).

fsa_count0(short,FaIn):-
    fsa_count_data(FaIn,Counts),
    format_counts_s(Counts),
    nl.

format_counts_l([]).
format_counts_l([Count/Vars|T]):-
    format(Count,Vars),
    format_counts_l(T).

format_counts_s([]).
format_counts_s([_/[Var]|T]):-
    format("~w ",[Var]),
    format_counts_s(T).

fsa_density_file(File0,File) :-
    fsa_io:fsa_read_file(File0,FaIn),
    fsa_util:to_file(fsa_data:fsa_density(FaIn),File).

fsa_density(FaIn) :-
    fsa_density(FaIn,[DetDens,AbsDens,JDens]),
    format("%% det. density: ~6f; ",[DetDens]),
    format("abs. density: ~6f; ",[AbsDens]),
    format("jump density: ~6f~n",[JDens]).

fsa_density(Fa,[Det,Abs,J]) :-
    fsa_states_number(Fa,StatesNo),
    fsa_symbols_number(Fa,SymbolsNo),
    fsa_transitions(Fa,Transitions), length(Transitions,TransitionsNo),
    fsa_jumps(Fa,Jumps), length(Jumps,JumpsNo),
    Det is TransitionsNo/(StatesNo*SymbolsNo),
    Abs is Det/StatesNo,
    J is JumpsNo/(StatesNo*StatesNo).


fsa_identical(A0,B0) :-
    (	check_deterministic(A0)
    ->	true
    ;   format(user_error,
           "fsa_identical: warning: automaton is not deterministic.~n",
           [])
    ),
    (	check_deterministic(B0)
    ->	true
    ;   format(user_error,
               "fsa_identical: warning: automaton is not deterministic.~n",
           [])
    ),  %%% eigenlijk ook niet goed: want dit *maakt* hem accessible,
        %%% maar we nemen aan dat-i deterministic is, dus dat gebeurt nooit
    fsa_regex:fsa_regex_compile(accessible(freeze(fa(A0))),A),
    fsa_regex:fsa_regex_compile(accessible(freeze(fa(B0))),B),
    A=B.

fsa_identical_files_report(FileA,FileB):-
    fsa_io:fsa_read_file(FileA,FaA),
    fsa_io:fsa_read_file(FileB,FaB),
    (	fsa_identical(FaA,FaB)
    ->	format("~w and ~w are identical~n",[FileA,FileB])
    ;   format("~w and ~w are not identical~n",[FileA,FileB])
    ).

%% this is different from fsa_determinizer:fsa_deterministic, in that
%% we do NOT care about predicates here, nor about pairs etc. If
%% it LOOKS deterministic, we are happy.
check_deterministic(Fa) :-
    fsa_data:fsa_components(_,_Nr,[_],_,Transitions,[],Fa),
    is_deterministic(Transitions).

is_deterministic([]).
is_deterministic([trans(P,Sym,_)|Tail]) :-
    is_deterministic(Tail,P,Sym).

is_deterministic([],_,_).
is_deterministic([trans(P0,Sym0,_)|Tail],P,Sym):-
    (	P0 == P
    ->	Sym0 \== Sym
    ;   true
    ),
    is_deterministic(Tail,P0,Sym0).

