:- module(fsa_globals, [ fsa_global_get/2,
			 fsa_global_set/2,
			 fsa_global_decl/5,
			 fsa_global_list/0,
			 fsa_global_list/1,
			 fsa_host_prolog/1,
			 fsa_version/0
		       ]).

#swi# :- style_check([-atom]).

#yap# :- use_module(library(charsio), [ format_to_chars/3 ]).
#!yap# :- use_module(charsio, [ format_to_chars/3 ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

user:help_info(class,global,"Global Variables",
"This section lists the global variables and documents their\
 \n effect. Global variables can be set from the command line and the\
 \n command interpreter using Var=Val. You can also set variables using\
 \n the Settings menu of the graphical user interface. Note that global\
 \n variables use the blackboard primitives; all keys are module expanded\
 \n using the *fsa* module."). 

user:help_info(global,Var,VarString,ExplString):-
    fsa_global_decl(Var,Expl,Default,Possible,_Check),
    (	user:bb_get(fsa:Var,Current)
    ->  true
    ;   Current=Default
    ),
    atom_codes(Var,VarString),
    charsio:format_to_chars("~s~nCurrent value is ~w. Default value is ~w. Typical values are ~w~n",[Expl,Current,Default,Possible],ExplString).


fsa_global_set(Key,Val) :-
    fsa_global_decl(Key,_,_,_,Val^Call),
    !,
    (	call(Call)
    ->	user:bb_put(fsa:Key,Val)
    ;   user:throw(fsa_error("Value not appropriate:~w = ~w~n",[Key,Val]))
    ).
fsa_global_set(Key,Val) :-
    format(user_error,"Warning: unknown flag ~w~n",[Key]),
    user:bb_put(fsa:Key,Val).	   

user:help_info(xpred,fsa_global_set,"fsa_global_set(+Key,?Val)",
"Predicate to set the global variable with name *Key* to *Val*.").

user:help_info(xpred,fsa_global_get,"fsa_global_get(+Key,?Val)",
"Predicate to query the value of the global variable with name *Key*.\
 \n If the value is undefined then *Val* is unified to a default value.\
 \n These default values are available as the third argument of the\
 \n *fsa_global_decl* predicate.").

fsa_global_get(Key,Val) :-
    (	user:bb_get(fsa:Key,Val)
    ->	true
    ;   fsa_global_decl(Key,_,Val,_,_)
    ).

user:help_info(xpred,fsa_global_decl,
"fsa_global_decl(?Key,?Help,?Default,?Typical,Val^Goal)",
"*Key* is a global variable with default value *Default*. Some typical\
 \n values are given in the list *Typical*. *Help* is a string explaining the\
 \n variable. Val^Goal can be used to check that Val is an appropriate\
 \n value for this flag.").

user:help_info(xpred,fsa_global_list,
"fsa_global_list[-List]",
"*List* will be unified with a keylist of all the global variables \
 \n with their associated values. If no argument is given, then this list\
 \n is written to standard output"). 

fsa_global_list :-
    fsa_global_list(GlobalVals),
    format_all_globals(GlobalVals).

format_all_globals([]).
format_all_globals([Global-Val|Tail]):-
    format(user_error,"~w:~w~n",[Global,Val]),
    format_all_globals(Tail).

%% only globals in fsa module...
fsa_global_list(GlobalVals):-
    findall(Global-Val,( fsa_global_decl(Global,_,Val0,_,_),
		       (   user:bb_get(fsa:Global,Val)
		       ->  true
		       ;   Val=Val0
		       )), GlobalVals0),
    keysort(GlobalVals0,GlobalVals).

#sic# fsa_global_decl(tkconsol,
#sic# "Boolean flag which determines whether *library(tkconsol)* is used for\
#sic# \n standard output. Note that support for tkconsol is very experimental.",
#sic# 	        off,[on,off],V^(lists:member(V,[on,off]))). 

#sic# fsa_global_decl(tk_fsa_add_help_menu,
#sic# "Boolean flag which determines whether on-line help information is added\
#sic# \n the menu. Since this takes quite a bit of band-width, you might want to\
#sic# \n turn it off for slow internet connections.",on,[on,off],
#sic#          V^(lists:member(V,[on,off]))).

#sic# fsa_global_decl(fsa_tcl_directory,
#sic# "Path to the directory in which the FSA tcl scripts are installed",
#sic# none,[],V^atom(V)).

fsa_global_decl(pred_module,
"Default module for interpreting predicates on symbols. ",
		fsa_preds,[fsa_preds,fsa_frozen],_^true).

fsa_global_decl(regex,"Used internally",[],[],_^true).
fsa_global_decl(fa,   "Used internally",[],[],_^true).

fsa_global_decl(hash_size,
"Default size for hashes (refer to library *fsa_arrays* for detail).",
		65025,
		[500,1000,5000,10000,50000,100000,250000,500000,1000000],
		X^integer(X)).

fsa_global_decl(interactive,
"This flag can be used to indicate that you want to run FSA interactively,\
 \n even if you provide a command-line argument which would normally cause\
 \n non-interactive usage. The value *cmdint* also implies interactivity but\
 \n in addition the command interpreter is started.",
		off,[on,off,cmdint],X^(lists:member(X,[on,off,cmdint,halt]))). 

%% fsa_global_decl(time_out,
%% "Number of seconds for timeout. If FSA is started non-interactively, then
%% the main command is performed with a time-out set; however this doesn't 
%% seem to be very robust (sometimes the process continues nonetheless).",
%% 	        2147483646,
%%		[500,1000,10000,100000,2147483646],X^integer(X)). 

fsa_global_decl(pstricks_style,
"Determines what kind of pstricks picture is contructed; at the moment\
 \n *fancy* and *plain* are equivalent except that *fancy* implies that \
 \n colors are used.",
	        plain,[fancy,plain],X^(lists:member(X,[fancy,plain]))). 

fsa_global_decl(v_algorithm,
"One of *dot* or *tree*. The first uses AT&T's *dot* program (from the\
 \n GraphViz package) to compute geometry of states. The latter uses a \
 \n built-in method which works reasonable for small graphs.",
	        tree,[tree,dot],X^(lists:member(X,[tree,dot]))). 

fsa_global_decl(v_tree_depth,
"Is used by the tree algorithm for visualisation. Its effect has been\
 \n forgotten by the author.",
		off,[on,off],X^(lists:member(X,[on,off]))). 

fsa_global_decl(v_angle,
"Angle of edges in visualization of automata on Tk Canvas, as well\
 \n as for postscript and latex output. A value of 0 implies straight lines\
 \n between nodes. For larger values the lines that are drawn between nodes\
 \n will move further away from the straight line.",
		0.15,
		[0.1,0.15,0.2,0.3,0.4,0.5,1.0],X^number(X)). 

fsa_global_decl(v_xdist,
"Horizontal distance of states in visualization of automata on Tk Canvas, \
 \n as well as for postscript and latex output. ",
	        120,[40,60,80,100,120,150,200],X^number(X)). 

fsa_global_decl(v_ycoord,"Used internally by the *tree* algorithm for\
 \n visualization. I don't think it matters.",
	        200,[],X^number(X)). 

fsa_global_decl(display_unused_states,
"This boolean variable determines whether states should be visualized which\
 \n have no outgoing or incoming transitions, and which are neither a\
 \n start state. ",
	        on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(fl_arbitrary_symbol,
"This flag determines which atomic symbol should be used if transducers\
 \n are compiled to C, C++ or Java, and the transducer contains don't care\
 \n outputs. The value must be a Prolog atom. For example, if the transducer\
 \n is given by [c,?,a:?]* and the value of this flag is '?', then the input\
 \n \
 \n         cxacyacza\
 \n \
 \n will be mapped to \
 \n \
 \n         cx?cy?cz?", '?',['?','0','#'],X^atom(X)).

fsa_global_decl(symbol_separator,
"This flags determines which character is used to separate sequences of\
 \n symbols that are accepted/transduced. For instance, if the value is 32\
 \n (for space) then you can type \
 \n \
 \n          a b a bb b aaa\
 \n \
 \n to indicate the sequence of six symbols a, b, a, bb, b, and aaa. If the\
 \n value is 44 (for comma) the same sequence is written/read as \
 \n \
 \n          a,b,a,bb,b,aaa\
 \n \
 \n As a special case, a value of 0 indicates that a sequence is written\
 \n without a separator; every single letter is assumed to be a symbol.\
 \n For instance,  \
 \n \
 \n          ababba\
 \n \
 \n represents the sequence of symbols a, b, a, b, b and a.",
	       0,[0,32,43,44,45],X^(integer(X),X<256)).

fsa_global_decl(symbol_separator_out,
"As global variable *symbol_separator*, but only for output. If this\
 \n variable is undefined, then the value of the global variable \
 \n *symbol_separator* is used instead.", 
	        undefined,[undefined,0,32,43,44,45],
		X^(integer(X),X<256)).		

fsa_global_decl(symbol_separator_in,
"As global variable *symbol_separator*, but only for input. If this \
 \n variable is undefined, then the value of the global variable \
 \n *symbol_separator* is used instead.", 
	        undefined,[undefined,0,32,43,44,45],
		X^(integer(X),X<256)).

fsa_global_decl(nr_sol_max,
"For the *produce* and the *transduce* options this global variable \
 \n determines how many transductions for each input string should be given \
 \n at most.",
	        25,
		[1,5,10,25,50,100,1000,10000],X^integer(X)). 

fsa_global_decl(length_max,
"For the produce options this global variable  determines the\
 \n maximum length of strings that should be produced. In  the case of\
 \n transducer the variable determines maximum length of left  string. A value\
 \n of 0 indicates no restriction (in that case strings are not\
 \n produced in order of length).",30,[0,5,10,25,50,100,1000,10000],
          X^integer(X)). 

fsa_global_decl(interpreter,
"This boolean flag indicates whether input automata for *fsa_interpreter*\
 \n are compiled (by fsa_compiler_to_prolog) or interpreted.",
                on,[on,off],X^(lists:member(X,[on,off]))). 

fsa_global_decl(debug,
"A value 0 indicates no continuation messages at all. A value of 1\
 \n will give cputime of operation. A level of 2 will give cputime of all\
 \n intermediate operations too. Finally, level 3 is used for detailed\
 \n continuation messages",0,[0,1,2,3,4],X^(integer(X),X<8)).

fsa_global_decl(regex_cache,
"This global variable determines whether regular expression compilations\
 \n  are cached or not. If the value *selective* is used, then only \
 \n those operators are cached for which \
 \n \
 \n     bb_get(fsa_regex_cache:Fun)\
 \n \
 \n succeeds.",
                selective,[on,off,selective],
		X^(lists:member(X,[on,off,selective]))). 

fsa_global_decl(determinize_preds_cache,
"This global variable determines whether during the determinization\
 \n algorithm the determinize_preds predicate should be cached. This is\
 \n typically much faster, but requires much more memory in some cases.",
                on,[on,off],
		X^(lists:member(X,[on,off]))). 

fsa_global_decl(cleanup_list_cache,
"This global variable determines whether during the cleanup\
 \n algorithm the determinize_preds predicate should be cached. This is\
 \n typically faster, but requires more memory in some cases.",
                on,[on,off],
		X^(lists:member(X,[on,off]))). 

fsa_global_decl(set_random,
"This boolean global variable indicates whether the random generator\
 \n should start with a new seed or not. If *off* the sequence of randomly\
 \n generated automata will be the same for different FSA incarnations.",
	        off,[on,off],X^(lists:member(X,[on,off]))). 

fsa_global_decl(w_determinizer_minimum,
"This flag determines whether the *t_determinizer* applied to transducers\
 \n using the *fsa_weights* semiring should only consider paths with lowest \
 \n scores.",
	        on,[on,off],X^(lists:member(X,[on,off]))). 

fsa_global_decl(read,
"This global variable determines the format of input automata. The \
 \n formats are explained in *module(fsa_io)*.",
	        normal, [normal,old,fast,compact,fsm,grail],
		X^(lists:member(X,[normal,old,fast,compact,fsm,grail]))).

fsa_global_decl(write,
"This global variable determines the format of output automata. The \
 \n formats are explained in *module(fsa_io)*.",
	        normal, 
		[normal,old,fast,compact,postscript,vcg,davinci,
		 dot,pstricks,latex,prolog,c,count,fsm,java,cpp,grail],
		X^(lists:member(X,[normal,old,fast,compact,postscript,
             vcg,davinci,dot,pstricks,latex,prolog,c,count,
	     fsm,java,cpp,grail]))).

fsa_global_decl(count,
"This global variable determines if results are displayed in\
 \n long or short format for the *count* output format, the *-count* option\
 \n and the fsa_count predicate.",
	        long,[short,long],X^(lists:member(X,[short,long]))).

fsa_global_decl(postscript_res,
"This variable determines which version of postscript output is used.\
 \n *lowres* is better used for conversion to pngs, *normal* is better \
 \n used for printing postscript",
	        normal,[normal,lowres],X^(lists:member(X,[normal,lowres]))). 

fsa_global_decl(no_display_beyond,
"Integer which determines a maximum number of states for automata that\
 \n are displayed by any of the visualization tools. Automata with\
 \n more states are not displayed; in such cases a small automaton is\
 \n displayed indicating that the maximum was reached.",
	        50,[30,40,50,100,1000],X^integer(X)). 

fsa_global_decl(use_graphviz,
"Render graphs using graphviz instead of the built-in renderer.",
	        off,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(c_with_main,
"This variable has effect for compilation of automata to C. If *on*, then\
 \n the resulting C program will contain a main procedure. If *off* no \
 \n such main procedure will be created.",
	        on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(java_with_main,
"This variable has effect for compilation of automata to JAVA. If *on*, then\
 \n the resulting JAVA program will contain a main procedure. If *off* no \
 \n such main procedure will be created.",
	        on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(cpp_with_main,
"This variable has effect for compilation of automata to C++. If *on*, then\
 \n the resulting C++ program will contain a main procedure. If *off* no \
 \n such main procedure will be created.",
	        on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(to_c_conversion,
"Boolean variable which has effect for compilation of automata to C, cf \
 \n the *fsa_compiler_to_c* module for details. If on, the automaton is \
 \n converted first; otherwise it's assumed the input is already converted.",
                on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(to_java_conversion,
"Boolean variable which has effect for compilation of automata to JAVA, cf \
 \n the *fsa_java* module for details. If on, the automaton is \
 \n converted first; otherwise it's assumed the input is already converted.",
                on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(to_cpp_conversion,
"Boolean variable which has effect for compilation of automata to C++, cf \
 \n the *fsa_cpp* module for details. If on, the automaton is \
 \n converted first; otherwise it's assumed the input is already converted.",
                on,[on,off],X^(lists:member(X,[on,off]))).

fsa_global_decl(fl_multiple_symbol_start,
"If transducers are compiled to C, C++ or Java, then if they contain final\
 \n states with multiple outputs these multiple outputs are combined into\
 \n a single output. This single output starts with a special symbol: the\
 \n value of the f_multiple_symbol_start flag. The single output is constructed\
 \n by concatenating each of the possible outputs, separated by the value of\
 \n the fl_multiple_symbol_separator flag. For example, the transducer\
 \n defined by the expression {a:b,a:c,a:d} will write out the symbol #d|c|b\
 \n assuming that fl_multiple_symbol_start is # and fl_multiple_symbol_sep is |.",
'#',['#'],X^atom(X)).
fsa_global_decl(fl_multiple_symbol_separator,
"If transducers are compiled to C, C++ or Java, then if they contain final\
 \n states with multiple outputs these multiple outputs are combined into\
 \n a single output. This single output starts with a special symbol: the\
 \n value of the f_multiple_symbol_start flag. The single output is constructed\
 \n by concatenating each of the possible outputs, separated by the value of\
 \n the fl_multiple_symbol_separator flag. For example, the transducer\
 \n defined by the expression {a:b,a:c,a:d} will write out the symbol #d|c|b\
 \n assuming that fl_multiple_symbol_start is # and fl_multiple_symbol_sep is |.",
'|',['|'],X^atom(X)).

user:help_info(xpred,fsa_version,"fsa_version",
"FSA version information is displayed on standard error. Note that the\
 \n version information is available through the *fsa_version* global variable.").

fsa_version:-
    user:bb_get(fsa:fsa_version,Atom),
    format(user_error,"FSA utilities ~w~n",[Atom]).

user:help_info(xpred,fsa_host_prolog,"fsa_host_prolog(?Atom)",
"Atom is an atom indicating the current Prolog system. At the\
 \n moment Atom is one of *sicstus*, *yap*, or *swi*.").
	  
#sic# fsa_host_prolog(sicstus).
#yap# fsa_host_prolog(yap).
#swi# fsa_host_prolog(swi).
