:- module(fsa_hash, [ %% non-updatable hashes
		      fsa_hash_new/1,
		      fsa_hash_new/2,
		      fsa_hash_access/3,
		      fsa_hash_access/4,
		      fsa_hash_to_keylist/2,
		      fsa_hash_get/3
		    ]).

#swi# :- style_check([-atom]).

#!swi# :- use_module(library(terms), [ term_hash/4 ]).
#swi# :- use_module(terms, [ term_hash/4 ]).
#yap# :- use_module(library(lists), [ memberchk/2 ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.
:- public help_info/4.

user:help_info(module,fsa_hash,"Non-updatable Hashes (N+K trees)",
"This module provides a non-updatable hash datastructure, on top of\
 \n the fsa_array module.\
 \n \
 \n Here's an overview of the predicates provided:  \
 \n \
 \n * fsa_hash_new/[1,2]          create a new non-updatable hash\
 \n * fsa_hash_access/[3,4]       access a value in a non-updatable hash\
 \n * fsa_hash_get/3              get a value in a non-updatable hash\
 \n \
 \n The hash function is taken from library(terms). The default size of the\
 \n hashes is determined by the global variable *hash_size*.").


help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

help_info(pred,fsa_hash_new,"fsa_hash_new(-FsaHash[,Size])",
"Initializes a new FsaHash with size Size; or default size if there is\
 \n no second argument. The default size is given by the global variable\
 \n *hash_size*. ").

fsa_hash_new(hash(Assoc,Size)) :-
    fsa_globals:fsa_global_get(hash_size,Size),
    fsa_array:fsa_array_new(Assoc,Size).

fsa_hash_new(hash(Assoc,Size),Size) :-
    fsa_array:fsa_array_new(Assoc,Size).

help_info(pred,fsa_hash_access,
"fsa_hash_access(+Key,?Val[,?Default],+FsaHash)",
"Unifies Val with the value associated with Key in FsaHash. Note that\
 \n keys must be ground Prolog terms. For the 4-ary form, if Key had no\
 \n associated value, then Default is unified with Val (and Key is not added\
 \n to the table).").
fsa_hash_access(Key,Val,hash(Thing,Size)):-
    term_hash(Key,-1,Size,Index),
    fsa_array:fsa_array_access(Index,VarList,Thing),
    key_member_or_add_tailvar_list(Key,Val,VarList).

fsa_hash_get(Key,Val,hash(Thing,Size)):-
    term_hash(Key,-1,Size,Index),
    fsa_array:fsa_array_get(Index,VarList,Thing),
    key_member_tailvar_list(Key,Val,VarList).

fsa_hash_access(Key,Val,Default,Hash):-
    fsa_hash_access(Key,Val0,Hash),
    (	var(Val0)
    ->	Val=Default
    ;   Val0=Val
    ).

key_member_or_add_tailvar_list(Term,Val,Var) :-
    (	var(Var)   
    ->	Var=[Term-Val|_]
    ;	Var=[Key-Val1|Tail],
	(   Term==Key  
	->  Val1=Val
	;   key_member_or_add_tailvar_list(Term,Val,Tail)
	)
    ).

key_member_tailvar_list(Key,Val,Var) :-
    nonvar(Var),
    Var=[Key0-Val0|Tail],
    (	Key==Key0
    ->	Val=Val0
    ;   key_member_tailvar_list(Key,Val,Tail)
    ).

help_info(pred,fsa_hash_to_keylist,
"fsa_hash_to_keylist(+HashedFsaArray,-Keylist)",
"Keylist is a list of all the Key-Value pairs in HashedFsaArray."). 

fsa_hash_to_keylist(hash(Thing,_),List):-
    fsa_array:fsa_array_to_list(Thing,List0),
    list_to_keylist(List0,List).

list_to_keylist([],[]).
list_to_keylist([H|T],List):-
    tailvar_list_to_keylist(H,List,List0),
    list_to_keylist(T,List0).


tailvar_list_to_keylist(V,List0,List):-
    (	var(V)
    ->  List0=List
    ;   V==[]
    ->	List0=List
    ;   V=[H|V0],
	List0=[H|List1],
	tailvar_list_to_keylist(V0,List1,List)
    ).
