%%% -*- Module: fsa_interpreter; -*-
:- module( fsa_interpreter,
   [ fsa_regex_accepts/2,
     fsa_regex_transduces/3,
     fsa_regex_transduces_w/3,
     fsa_regex_transduces_wt/4,
     fsa_regex_accept_all/3,

     fsa_accepts/2,
     fsa_accepts_file/2,
     fsa_transduces/3,
     fsa_transduces_w/3,
     fsa_transduces_wt/4,
     fsa_accept_all/3,

     fsa_produces/2,
     fsa_produces/3,

     fsa_sample/2,
     fsa_sample/1,

     fsa_approx/2,
     fsa_approx_file/2,
     fsa_regex_approx_accepts/3,
     fsa_regex_approx_transduces/3,
     fsa_regex_approx_transduces_w/3,
     fsa_regex_approx_transduces_wt/4,

     fsa_approx_accepts/3,
     fsa_approx_transduces/3,
     fsa_approx_transduces_w/3,
     fsa_approx_transduces_wt/4
   ]).

#swi# :- style_check([-atom]).

#yap# :- use_module(library(charsio), [ format_to_chars/3 ]).
#!yap# :- use_module(charsio, [ format_to_chars/3 ]).
#yap# :- use_module(library(lists),   [ memberchk/2, append/3, member/2, reverse/2 ]).
#sic# :- use_module(library(lists),   [ reverse/2 ]).

#sic# :- use_module(library(random), [ setrand/1, random/1, random/3 ]).

:- meta_predicate interpreted_or_compiled(?,:,:),
                  atmost(:,:,?), atmost(?,?,:,?).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

:- public fsa_interpreter_aux:unknown_predicate_handler/2.
fsa_interpreter_aux:unknown_predicate_handler(accepts(_),true) :-
    !,
    user:throw(fsa_error(
	"accepts/1 does not exist.\
 \n Maybe your compiled automaton wasn't a recognizer?~n",[])).
fsa_interpreter_aux:unknown_predicate_handler(t_accepts(_,_),true) :-
    !,
    user:throw(fsa_error(
	"t_accepts/2 does not exist.\
 \n Maybe your compiled automaton wasn't a transducer?~n",[])).
fsa_interpreter_aux:unknown_predicate_handler(w_accepts(_,_),true) :-
    !,
    user:throw(fsa_error(
	"w_accepts/2 does not exist.\
 \n Maybe your compiled automaton wasn't a weighted recognizer?~n",[])).

fsa_interpreter_aux:unknown_predicate_handler(wt_accepts(_,_,_),true) :-
    !,
    user:throw(fsa_error(
	"wt_accepts/3 does not exist.\
 \n Maybe your compiled automaton wasn't a weighted transducer?~n",[])).

fsa_interpreter_aux:unknown_predicate_handler(_,fail).

user:help_info(xpred,fsa_regex_accepts,"fsa_regex_accepts(+Atom,+String)",
"Succeeds if String is accepted by the regular expression in Atom. For\
 \n example: \
 \n \
 \n     fsa_regex_accepts('[{a,b}*,b,a,b,{a,b}*]',\"abbbbabababa\").").

fsa_regex_accepts(Regex,String0):-
    string_to_list_of_symbols(String0,String),
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_accepts(String,Fa).

user:help_info(xpred,fsa_regex_transduces,"fsa_regex_transduces(+Atom,+String0,?String)",
"String is a transduction of String0 according to the regular\
 \n expression in Atom. Example:\
 \n \
 \n     fsa_regex_transduces('a:b',\"a\",L).\
 \n \
 \n     L = [98] ? ").

fsa_regex_transduces(Regex,String0,String):-
    string_to_list_of_symbols(String0,String1),
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_transduces(String1,String2,Fa),
    list_of_symbols_to_string(String2,String,out).

user:help_info(xpred,fsa_regex_transduces_w,
"fsa_regex_transduces_w(+Atom,+String0,?Weight)",
"String is a transduction of String0 according to the regular\
 \n expression in Atom. Example:\
 \n \
 \n     fsa_regex_transduces_w('[a:3,b:1*]',\"abbb\",L).\
 \n \
 \n     L = 6 ? ").

fsa_regex_transduces_w(Regex,String0,Weight):-
    string_to_list_of_symbols(String0,String1),
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_transduces_w(String1,Weight,Fa).

fsa_regex_transduces_wt(Regex,String0,Out,Weight):-
    string_to_list_of_symbols(String0,String1),
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_transduces_wt(String1,Out0,Weight,Fa),
    list_of_symbols_to_string(Out0,Out,out).

user:help_info(xpred,fsa_accepts,"fsa_accepts(+String,+Fa)",
"This predicate can be used both to recognize a given string\
 \n or to produce a string according to Fa. This is why we use dif/2\
 \n below. We prefer functionality over efficiency here; note that\
 \n the compiler-to-prolog implements the same functionality.\
 \n \
 \n Making this code faster could be done for instance by indexing on source\
 \n state and symbol. For deterministic automata, use the compiler-to-c\
 \n for a fast and compact recognizer.\
 \n \
 \n Since input can be non-deterministic we check for epsilon-cycles\
 \n (the fifth argument of accepts/6 is a list of states visited\
 \n after last consumption of input). Again, there are cases where it\
 \n would make more sense to pre-compute efree automata first. But if\
 \n that's the case you could do it, right?").
fsa_accepts(String,Fa0) :-
    fsa_regex:fsa_regex_compile(recognizer(fa(Fa0)),Fa),
    fsa_data:fsa_start_state(Fa,State),
    accepts(State,String,Fa,[State]).

%% accepts/5 does the real work of accepts/2
%% accepts(+State
%%         +Input
%%         +Fa
%%         +VisitedStatesSinceLastInput (unordered list)
%%         +QmAtom
accepts(State,[],Fa,_) :-
    fsa_data:fsa_final_state(Fa,State).
accepts(State0,S0,Fa,States0):-
    apply_transition(Fa,State0,State,S0,S,States0,States,Goals,[]),
    call_list(Goals),
    accepts(State,S,Fa,States).

%% goals that should be called while interpreting a given string
%% are collected in Goals0-Goals. The reason is that this predicate is
%% also used in fsa_compiler.pl
apply_transition(Fa,State0,State,S0,S,States0,States,Goals0,Goals) :-
    transition(Fa,State0,Sym,State),
    match_input(Sym,Fa,S0,S,_,_,Goals0,Goals1),
    (	Sym == []
    ->	Goals1= [\+ memberchk(State,States0)|Goals],
	States = [State|States0]
    ;   States = [State], Goals1=Goals
    ).

transition(Fa,State0,In,State) :-
    fsa_data:fsa_transition(Fa,State0,In,State).
transition(Fa,State0,[],State) :-
    fsa_data:fsa_jump(Fa,State0,State).

user:help_info(xpred,fsa_transduces,"fsa_transduces(+StringIn,?StringOut,+Fa)",
"StringOut is a transduction of StringIn according to transducer Fa.\
 \n \
 \n This predicate employs a meta-notation in cases where loop-checking\
 \n encounters a cycle. In that case the notation  \
 \n \
 \n       |N+|\
 \n \
 \n is written into the output string indicating that the previous N\
 \n symbols could be repeated here as many times as desired. For example,\
 \n consider the simple regular expression mapping an a to one or more b's: \
 \n \
 \n          a:(b+)\
 \n \
 \n If a transduction is request for input string a, then the following\
 \n outputs occur:\
 \n \
 \n     b\
 \n     bb|1+|\
 \n \
 \n In the second output string the meta-notation indicates that the\
 \n second b could be repeated multiple times.\
 \n  \
 \n  \
 \n Many of the same remarks wrt fsa_accepts/2 apply here: works for\
 \n non-instantiated input (lists with variable elements work OK, but\
 \n variable length lists typically don't). Also takes care of identity\
 \n constraints in the transducer, including delayed identity constraints and\
 \n too early identity constraints, by using a non-proper implementation\
 \n of queues which allow dequeue-ing before enqueue-ing! Examples to try,\
 \n using the fsa_preds predicate module:\
 \n \
 \n         t_minimize([a:b,class(a..f)])\
 \n \
 \n         t_minimize({[a:b,?,?,?,?,?,b],[a:c,?,?,?,?,?,c]})\
 \n \
 \n I think this is neat.").

fsa_transduces(String,Output,Fa0) :-
    fsa_regex:fsa_regex_compile(sequence_transducer(fa(Fa0),silent),Fa),
    fsa_data:fsa_start_state(Fa,State),
    empty_q(EmptyQ),
    t_accepts(State,String,Output,Fa,[],[State],0,EmptyQ).

%% t_accepts(+State,+String,-Output,+Fa,
%%           +yes/no: is the next transition allowed to be on [] input?
%%           +States0: list of states visited since last input consumption
%%                     (always includes current state) for loop-checking
%%           +OutLL: integer indicating the number of output symbols added
%%                     since last input consumption (used for meta-output
%%                     in case of loops on [] input)
%%           +Queue: queue indicating the sequence of input symbols matched
%%                   with $@ which haven't yet been outputted (required to
%%                   treat the result of t_determinizing identity(?) automs.)
t_accepts(State,[],[],Fa,_,_,_,_Emptyq) :-   
    fsa_data:fsa_final_state(Fa,State).
t_accepts(State0,S,T,Fa,Bool0,States0,OutLL0,QmStack0) :-
    apply_t_transition(State0,State,S,S1,T,T1,Fa,States0,States,
		     Bool0,Bool,OutLL0,OutLL,QmStack0,QmStack,Goals,[]),
    call_list(Goals),
    t_accepts(State,S1,T1,Fa,Bool,States,OutLL,QmStack).

t_transition(Fa,State0,In/Out,State) :-
    fsa_data:fsa_transition(Fa,State0,In/Out,State).
t_transition(Fa,State0,[]/[],State) :-
    fsa_data:fsa_jump(Fa,State0,State).

%% goals that should be called while interpreting a given string
%% are collected in Gs0-Gs. The reason is, that this predicate is
%% also used in fsa_compiler.pl
apply_t_transition(State0,State,S0,S,T0,T,Fa,States0,States,
	Bool0,Bool,LL0,LL,Qs0,Qs,Gs0,Gs) :-
    t_transition(Fa,State0,In/OutList,State),
    match_input(In,Fa,S0,S,Qs0,Qs1,Gs0,Gs1),
    match_output(OutList,Fa,T0,T1,Qs1,Qs,LL0,LL1,Gs1,Gs2),
    (	In == []                % loop checking...
    ->	Gs2 = [(   memberchk(State,Bool0)
	       ->  fail
	       ;   memberchk(State,States0)
	       ->  LL1 > 0,	% LOOP with output!
				% Loop without output simply fails here
		   T1=['|',LL1,'+|'|T],	% introduce meta notation
		   Bool=[State|Bool0],	% disallow [] input in next steps
		   LL=0
	       ;   %% no loop (yet)
		   States=[State|States0], T1=T, Bool0=Bool, LL=LL1   
	       )|Gs]
    ;	%% no [] input, forget old states
	Gs2=Gs, States=[State], T1=T, Bool=[], LL=0         
    ).

match_input(Sym,Fa,S0,S,Qs0,Qs,Goals0,Goals) :-
    (	Sym == []
    ->  S0 = S, Qs0=Qs, Goals0=Goals
    ;	S0=[Seen|S], 
	(   Sym = $@(Pred)
	->  enq(Seen,Qs0,Qs)
	;   Sym=Pred, Qs0=Qs
	),
	domain_module(Fa,Module),
	Goals0=[fsa_pred_module:evaluate_predicate(Module,Pred,Seen)|Goals]
    ).


%% here: from symbol to type?
%% LL1, LL are used to keep track of the length of possible output
%% cycles (used for meta-notation in output).
match_output([],_Fa,S,S,Qs,Qs,LL,LL,Gs,Gs).
match_output([Sym|Syms],Fa,S0,S,Qs0,Qs,LL0,LL,Gs1,Gs) :-
    (	Sym = '$@dequeue'
    ->	deq(_,Qs0,Qs1),
	match_output(Syms,Fa,S0,S,Qs1,Qs,LL0,LL,Gs1,Gs)
    ;	S0=[Seen|S1],
	Gs1=[LL1 is LL0+1|Gs0],
	(   Sym = $@(Pred)
	->  deq(Seen,Qs0,Qs1)
	;   Qs0=Qs1, Sym=Pred
	),
	range_module(Fa,Module),
	Gs0=[fsa_pred_module:evaluate_predicate(Module,Pred,Seen)|Gs2],
	match_output(Syms,Fa,S1,S,Qs1,Qs,LL1,LL,Gs2,Gs)
    ).

user:help_info(xpred,fsa_transduces_w,"fsa_transduces_w(+String,?Weight,+Fa)",
"Weight is the weight assigned to String by Fa. \
 \n \
 \n Similar to fsa_accepts/2 and fsa_transduces/3 above. However, we assume that\
 \n there are no identity constraints. Loop-checking for [] input,\
 \n but no meta-notation in output: we simply produce the minimum in such\
 \n cases. That seems to be appropriate in most applications (?).").

%%
fsa_transduces_w(String,Output,Fa0) :-
    fsa_regex:fsa_regex_compile(weighted_recognizer(fa(Fa0)),Fa),
    fsa_data:fsa_start_state(Fa,State),
    fsa_weights:zero(Zero),
    w_accepts(State,String,Zero,Output,Fa,[State]).

%% w_accepts(+State,+Input,+Output0,?Output,+Fa,
%%           +StatesVisistedSinceLastInputConsumption)
w_accepts(State,[],O,O,Fa,_) :-
    fsa_data:fsa_final_state(Fa,State).
w_accepts(State0,S,Sc0,Sc,Fa,Sts0) :-
    apply_w_transition(State0,State,S,S1,Sc0,Sc1,Fa,Sts0,Sts,Goals,[]),
    call_list(Goals),
    w_accepts(State,S1,Sc1,Sc,Fa,Sts).

%% goals that should be called while interpreting a given string
%% are collected in G0-G. The reason is that this predicate is
%% also used in fsa_compiler.pl
apply_w_transition(State0,State,S0,S,T0,T,Fa,Sts0,Sts,G0,G) :-
    w_transition(Fa,State0,In/Out,State),
    match_input(In,Fa,S0,S,_,_,G0,G1),
    match_w_output(Out,T0,T,G2,G),
    (	In == []
    ->  G1 = [\+ memberchk(State,Sts0)|G2],
	Sts = [State|Sts0]
    ;   G1 = G2, Sts=[State]
    ).

w_transition(Fa,State0,In/Out,State) :-
    fsa_data:fsa_transition(Fa,State0,In/Out,State).
w_transition(Fa,State0,[]/0,State) :-
    fsa_data:fsa_jump(Fa,State0,State).

match_w_output(Int,Sc0,Sc,G0,G) :-
    (	number(Int)
    ->  G0 = [Sc is Int+Sc0|G]
    ;   user:throw(fsa_error(
	"output symbol in string-to-weight transducer is not a number: ~w~n",
		[Int]))
    ).

fsa_transduces_wt(String,Out,Weight,Fa0) :-
    fsa_regex:fsa_regex_compile(weighted_sequence_transducer(fa(Fa0)),Fa),
    fsa_data:fsa_start_state(Fa,State),
    fsa_weights:zero(Zero),
    empty_q(EmptyQ),
    wt_accepts(State,String,Out,Fa,[],[State],0,EmptyQ,Zero,Weight).

wt_accepts(State,[],[],Fa,_,_,_,_Emptyq,W,W) :-   
    fsa_data:fsa_final_state(Fa,State).
wt_accepts(State0,S,T,Fa,Bool0,States0,OutLL0,QmStack0,W0,W) :-
    apply_wt_transition(State0,State,S,S1,T,T1,Fa,States0,States,
		     Bool0,Bool,OutLL0,OutLL,QmStack0,QmStack,Goals,[],W0,W1),
    call_list(Goals),
    wt_accepts(State,S1,T1,Fa,Bool,States,OutLL,QmStack,W1,W).

wt_transition(Fa,State0,In/Out,W,State) :-
    fsa_data:fsa_transition(Fa,State0,In/(Out/W),State).
wt_transition(Fa,State0,[]/[],0,State) :-
    fsa_data:fsa_jump(Fa,State0,State).

%% goals that should be called while interpreting a given string
%% are collected in Gs0-Gs. The reason is, that this predicate is
%% also used in fsa_compiler.pl
apply_wt_transition(State0,State,S0,S,T0,T,Fa,States0,States,
	Bool0,Bool,LL0,LL,Qs0,Qs,Gs0,Gs,W0,W) :-
    wt_transition(Fa,State0,In/OutList,W1,State),
    match_input(In,Fa,S0,S,Qs0,Qs1,Gs0,Gs1),
    match_output(OutList,Fa,T0,T1,Qs1,Qs,LL0,LL1,Gs1,Gs2),
    match_w_output(W1,W0,W,Gs2,Gs3),
    (	In == []                % loop checking...
    ->	Gs3 = [(   memberchk(State,Bool0)
	       ->  fail
	       ;   memberchk(State,States0)
	       ->  LL1 > 0,	% LOOP with output!
				% Loop without output simply fails here
		   T1=['|',LL1,'+|'|T],	% introduce meta notation
		   Bool=[State|Bool0],	% disallow [] output in next step
		   LL=0
	       ;   %% no loop (yet)
		   States=[State|States0], T1=T, Bool0=Bool, LL=LL1   
	       )|Gs]
    ;	%% no [] input, forget old states
	Gs3=Gs, States=[State], T1=T, Bool=[], LL=0         
    ).

%% utils
%% normal queues:
%% empty_q(q(X,X,0)).
%%
%% enq(El,q(Hd,[El|T2],N),  q(Hd,T2,s(N))).
%%
%% deq(El,q([El|T0],T,s(N)),q(T0,T, N)).
%%
%% 
%% queues that allow to dequeue element before it's enqueued.
%% this is actually useful for examples where identity-constrained predicates
%% are pushed forward even before its corresponding input predicate has
%% been applied!
empty_q(X-X).

enq(El,Hd-[El|T2],Hd-T2).

deq(El,[El|T0]-T, T0-T).

call_list([]).
call_list([H|T]) :-
    call(H),
    call_list(T).

%% next: the interfaces to the (t_|w_)accepts relations above
%% first two auxiliaries to facilitate distinction between
%% interpreted and compiled automaton.
get_fa(File,Flag,Fa) :-
    fsa_globals:fsa_global_get(interpreter,OnOff),
    (   OnOff==on
    ->	Flag=interpreted,
	fsa_io:fsa_read_file(File,Fa)
    ;   Flag=compiled,
	Fa=compiled,  %% to prevent variable, should not be necc.
	(   File==user
	->  
#!swi#             compile(fsa_interpreter_aux:user)
#swi#             consult(fsa_interpreter_aux:user)
	;   ensure_loaded(fsa_interpreter_aux:File)
	)
    ).


interpreted_or_compiled(interpreted,Call,_) :-
    Call.
interpreted_or_compiled(compiled,_,Call) :-
    Call.

fsa_accepts_file(List0,File):-
    get_fa(File,Flag,Fa),
    fsa_data:fsa_type(Fa,Type),
    string_to_list_of_symbols(List0,List),
    optp1(Type,Flag,Fa,List).

fsa_regex_accept_all(Regex,FileIn,FileOut):-
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    accept_all0(Fa,FileIn,FileOut,interpreted).

%% only for web-demo
:- public fsa_interpreter:fsa_regex_accept_all_t/3.

fsa_regex_accept_all_t(Regex,FileIn,FileOut):-
    fsa_regex:fsa_regex_atom_compile(Regex,Fa0),
    fsa_regex:fsa_regex_compile(identity_if_recognizer(fa(Fa0)),Fa),
    accept_all0(Fa,FileIn,FileOut,interpreted).

fsa_accept_all(File0,FileIn,FileOut):-
    get_fa(File0,Flag,Fa),
    accept_all0(Fa,FileIn,FileOut,Flag).

accept_all0(Fa,FileIn,FileOut,Flag):-
    user:call_cleanup(
	fsa_interpreter:accept_all0_call(Fa,FileIn,FileOut,Flag),
	(   seen,  told  )
		     ).

accept_all0_call(Fa,FileIn,FileOut,Flag):-
    fsa_data:fsa_type(Fa,Type),
    tell(FileOut), 
    see(FileIn),
    user:bb_put(fsa_interpreter:lines_read,0),
    user:bb_put(fsa_interpreter:nrsol,0),
    repeat,
    get_line(Line0,End_input),
    (	End_input =:= -1
    ->	user:bb_get(fsa_interpreter:lines_read,Lines),
	user:bb_get(fsa_interpreter:nrsol,Sol),
	format(user_error,"Accepted ~w of ~w lines~n",[Sol,Lines])
    ;	inc(lines_read),
	string_to_list_of_symbols(Line0,Line),
	optp1(Type,Flag,Fa,Line),
	fail
    ),
    !.  % repeat cut

fsa_produces(File0,File) :-
    get_fa(File0,Flag,Fa),
    fsa_data:fsa_type(Fa,Type),
    fsa_util:to_file(fsa_interpreter:fsa_produces(Type,Flag,Fa),File).


fsa_produces(Type,Flag,Fa) :-
    (	gen_length(List),
	optp2(Type,Flag,Fa,List),
	fail
    ;	true
    ).

fsa_sample(File0,File) :-
    get_fa(File0,Flag,Fa),
    (	Flag == compiled
    ->	user:throw(fsa_error("fsa_sample: not available for compiled automata~n",[]))
    ;	fsa_util:to_file(fsa_interpreter:fsa_sample(Fa),File)
    ).

% todo: other types apart from w_recognizer...
% todo: check Fa0's weights are consistent, i.e. all final states
% have a weight. Currently we necc. stop at any final state we encounter
% ...
fsa_sample(Fa0) :-
    fsa_sample_conversion(Fa0,Fa,Type),
    fsa_data:fsa_symbols_decl(Fa,PM),
    fsa_regex:set_random,
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    fsa_data:index_transitions(Fa,Table),
    user:bb_put(fsa_interpreter:nrsol,0),
    repeat,
    fsa_sample(Fa,Table,List,Type),
    write_sample_output(Type,List,PM),
    inc(nrsol),
    user:bb_get(fsa_interpreter:nrsol,Sols),
    Sols >= Max,
    !.

write_sample_output(recognizer,List0,r(Module,_)) :-
    random_symbols_from_preds(List0,List,Module),
    write_list_of_symbols(List,out),nl.

write_sample_output(transducer,Pairs,t(ModuleA,ModuleB,_)) :-
    random_symbols_from_preds(Pairs,ListL,ListR,ModuleA,ModuleB),
    write_list_of_symbols(ListL,out), write(' --> '),
    write_list_of_symbols(ListR,out), nl.

random_symbols_from_preds([],[],[],_,_).
random_symbols_from_preds([H0|T0],SymsL,SymsR,Ma,Mb) :-
    random_symbol_from_pred(H0,SymsL,SymsL1,SymsR,SymsR1,Ma,Mb),
    random_symbols_from_preds(T0,SymsL1,SymsR1,Ma,Mb).

random_symbol_from_pred(Pred,SymsL0,SymsL,SymsR0,SymsR,Ma,Mb) :-
    (	Pred == []   % is this possible?
    ->	SymsL0 = SymsL, SymsR0=SymsR
    ;	Pred = LPred/RPred,
	(   LPred = '$@'(P),
	    RPred = '$@'(P),
	    Ma = Mb
	->  random_symbol_from_pred_(P,Symbol,Ma),
	    SymsL0 = [Symbol|SymsL],
	    SymsR0 = [Symbol|SymsR]
	;   random_symbol_from_pred(LPred,SymsL0,SymsL,Ma),
	    random_symbol_from_pred(RPred,SymsR0,SymsR,Mb)
	)
    ).

random_symbols_from_preds([],[],_).
random_symbols_from_preds([H0|T0],Syms,M) :-
    random_symbol_from_pred(H0,Syms,Syms1,M),
    random_symbols_from_preds(T0,Syms1,M).

random_symbol_from_pred_(Pred,Symbol,Module) :-
    findall(Sym,fsa_pred_module:evaluate_predicate(Module,Pred,Sym),Syms),
    fsa_regex:random_element(Symbol,Syms).

random_symbol_from_pred(Pred,Syms0,Syms,Module) :-
    (	Pred == []
    ->	Syms0 = Syms
    ;	Syms0 = [Symbol|Syms],
	random_symbol_from_pred_(Pred,Symbol,Module)
    ).

fsa_sample(Fa,Table,List,Type) :-
    fsa_data:fsa_start_states(Fa,[Start|Rest]),
    (	Rest == []
    ->	true
    ;	user:throw(fsa_error(
           "fsa_sample/3 does not accept multiple start states~n",[]))
    ),	
    fsa_sample(Fa,Start,Table,List,Type).

fsa_sample(Fa,State,Table,List,Type) :-
    fsa_array:fsa_array_access(State,KeyList,[],Table),
    (	fsa_data:fsa_final_state(Fa,State)
    ->	List=[]
    ;
#sic#     random(Random),
#yap#     Random is random,
#swi#     Random is random(100000)/100000,
	fsa_array:fsa_array_access(State,KeyList,[],Table),
	sample_tr(Type,KeyList,Random,0,Symbol,State1),
	List=[Symbol|List1],
	fsa_sample(Fa,State1,Table,List1,Type)
    ).

sample_tr(recognizer,[(Sym/Weight)-Sts|Tail],Ran,Tot,Sym1,St1) :-
    sample_tr_(Sts,Sym,Weight,Tail,Ran,Tot,Sym1,St1,recognizer).
sample_tr(transducer,[(SymL/(SymR/Weight))-Sts|Tail],Ran,Tot,Sym1,St1) :-
    sample_tr_(Sts,SymL/SymR,Weight,Tail,Ran,Tot,Sym1,St1,transducer).

sample_tr_([],_,_,Tail,Ran,Tot,Sym1,St1,Type) :-
    sample_tr(Type,Tail,Ran,Tot,Sym1,St1).
sample_tr_([St|Sts],Sym,Weight,Tail,Ran,Tot,Sym1,St1,Type) :-
    Tot1 is Weight+Tot,
    (	Ran < Tot1
    ->	St1=St, Sym1=Sym
    ;	sample_tr_(Sts,Sym,Weight,Tail,Ran,Tot1,Sym1,St1,Type)
    ).

weights_to_probs(Trans0,Trans,Type) :-
    % todo: do this efficiently
    findall(P-L,
	    setof(trans(Sym,Q),lists:member(trans(P,Sym,Q),Trans0),L),
	    Ls0),
    weights_to_probs_each(Ls0,Trans,[],Type).



weights_to_probs_each([],T,T,_).
weights_to_probs_each([P-L0|Ls],T0,T,Type) :-
    sum_exp(L0,L1,0,Sum,Type),
    normalize(L1,Sum,P,T0,T1,Type),
    weights_to_probs_each(Ls,T1,T,Type).

sum_exp([],[],S,S,_).
sum_exp([Trans0|Ts0],[Trans|Ts],S0,S,Type) :-
    sum_exp_(Type,Trans0,Trans,S0,S1),
    sum_exp(Ts0,Ts,S1,S,Type).

sum_exp_(recognizer,trans(Sym/W0,Q),trans(Sym/W,Q),S0,S1) :-
    W is 2 ** (-W0),
    S1 is S0+W.
	    
sum_exp_(transducer,trans(SymL/(SymR/W0),Q),trans(SymL/(SymR/W),Q),S0,S1) :-
    W is 2 ** (-W0),
    S1 is S0+W.
	    
normalize([],_,_,T,T,_).
normalize([Trans0|Ts0],Sum,P,[Trans|T0],T,Type):-
    normalize_(Type,Trans0,Trans,Sum,P),
    normalize(Ts0,Sum,P,T0,T,Type).

normalize_(recognizer,trans(Sym/W0,Q),trans(P,Sym/W,Q),Sum,P) :-
    W is W0/Sum.
normalize_(transducer,trans(SymL/(SymR/W0),Q),trans(P,SymL/(SymR/W),Q),Sum,P):-
    W is W0/Sum.

%% variant 1: succeeds always, prints yes/no
%% variant 2: succeeds if accepted, prints string
optp1(recognizer,Flag,Fa,ListA) :-
    (	interpreted_or_compiled(Flag,
		fsa_interpreter:fsa_accepts(ListA,Fa),
		fsa_interpreter_aux:accepts(ListA)
			       )
        %% first solution is good enough for recognizers
    ->	inc(nrsol),
	format(user_error,"yes~n",[])
    ;   format(user_error,"no~n",[])
    ).

optp1(transducer(_),Flag,Fa,ListA) :-
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    atmost(fsa_interpreter:interpreted_or_compiled(Flag,
		         fsa_interpreter:fsa_transduces(ListA,ListB,Fa),
	                 fsa_interpreter_aux:t_accepts(ListA,ListB)),
		    (    fsa_interpreter:write_list_of_symbols(ListB,out),
			nl
		    ),
		    Max
		   ).


optp1(w_recognizer,Flag,Fa,ListA) :-
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    atmost(fsa_interpreter:interpreted_or_compiled(Flag,
	                fsa_interpreter:fsa_transduces_w(ListA,Score,Fa),
	                fsa_interpreter_aux:w_accepts(ListA,Score)),
		    format("~w~n",[Score]),
		    Max
		   ).

optp1(w_transducer(_),Flag,Fa,ListA) :-
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    atmost(fsa_interpreter:interpreted_or_compiled(Flag,
                fsa_interpreter:fsa_transduces_wt(ListA,ListB,Score,Fa),
                fsa_interpreter_aux:wt_accepts(ListA,ListB,Score)
						  ),
	   (   format("~w:",[Score]),
	       fsa_interpreter:write_list_of_symbols(ListB,out),
	       nl
	   ),
	   Max
	  ).

optp2(recognizer,Flag,Fa,ListA) :-
    interpreted_or_compiled(Flag,
		fsa_interpreter:fsa_accepts(ListA,Fa),
		fsa_interpreter_aux:accepts(ListA)
			   ),
    write_list_of_symbols(ListA,in),
    nl.

optp2(transducer(_),Flag,Fa,ListA) :-
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    atmost(fsa_interpreter:interpreted_or_compiled(Flag,
	             fsa_interpreter:fsa_transduces(ListA,ListB,Fa),
	             fsa_interpreter_aux:t_accepts(ListA,ListB)),
		(   fsa_interpreter:write_list_of_symbols(ListA,in),
		    write(' <==>'),
		    nl,
		    fsa_interpreter:write_list_of_symbols(ListB,out),
		    nl
		),
		Max ).

optp2(w_recognizer,Flag,Fa,ListA) :-
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    atmost(fsa_interpreter:interpreted_or_compiled(Flag,
	             fsa_interpreter:fsa_transduces_w(ListA,ListB,Fa),
                     fsa_interpreter_aux:w_accepts(ListA,ListB)),
		(   fsa_interpreter:write_list_of_symbols(ListA,in),
		    write(' <==>'),
		    nl,
		    write(ListB),
		    nl
		),
		Max).

optp2(w_transducer(_),Flag,Fa,ListA) :-
    fsa_globals:fsa_global_get(nr_sol_max,Max),
    atmost(fsa_interpreter:interpreted_or_compiled(Flag,
	        fsa_interpreter:fsa_transduces_wt(ListA,ListB,Score,Fa),
                fsa_interpreter_aux:wt_accepts(ListA,ListB,Score)
						  ),
	   (   fsa_interpreter:write_list_of_symbols(ListA,in),
	       write(' <==>'),
	       nl,
	       format("~w: ",[Score]),
	       fsa_interpreter:write_list_of_symbols(ListB,out),
	       nl
	   ),
	   Max ).


gen_length(List) :-
    fsa_globals:fsa_global_get(length_max,Max),
    (   Max =:= 0
    ->  true
    ;   gen_length(List,0,Max)
    ).

gen_length(List,L,Max) :-
    (	L > Max
    ->	fsa_util:debug_message(0,
                     "that's enough of that (flag(length_max,~w)~n",
	       [Max]),
	fail
    ;	length(List,L)
    ;	L1 is L+1,
	gen_length(List,L1,Max)
    ).

write_list_of_symbols(L,InOut) :-
    list_of_symbols_to_string(L,Str,InOut),
    format(Str,[]).

%% string_to_list_of_symbols(+String,-Symbols)
%% list_of_symbols_to_string(+Symbols,-String,InOut)
%% flag(symbol_separator,Char)
string_to_list_of_symbols([],[]).
string_to_list_of_symbols([Sh|St],Symbols) :-
    symbol_separator(in,Sep),
    (	Sep==0
    ->	string_to_list_of_symbols_e([Sh|St],Symbols)
    ;   string_to_list_of_symbols([Sh|St],Sep,Symbols,[])
    ).

string_to_list_of_symbols_e([],[]).
string_to_list_of_symbols_e([Str|String],[Sym|Symbols]) :-
    atom_codes(Sym,[Str]),
    string_to_list_of_symbols_e(String,Symbols).

string_to_list_of_symbols([],_,Syms,Syms).
string_to_list_of_symbols([H|T],Sep,[Symbol|Syms0],Syms) :-
    string_to_symbol(Symbol,Sep,S0,[],S0,[H|T],Rest),
    string_to_list_of_symbols(Rest,Sep,Syms0,Syms).

%% note: a_symbol is different from symbol_to_string
%% since there are symbols we can write (integers) but which are
%% read in differently...
string_to_symbol(Symbol,Sep,Sym0,Sym,SymbolStr,Rest0,Rest) :-
    (   Rest0==[]
    ->  Sym0=Sym,
	atom_codes(Symbol,SymbolStr),
	Rest=[]
    ;   Rest0=[H|T],
	(   Sep=H
	->  Rest=T,
	    Sym0=Sym,
	    atom_codes(Symbol,SymbolStr)
	;   Sym0=[H|Sym1],
	    string_to_symbol(Symbol,Sep,Sym1,Sym,SymbolStr,T,Rest)
	)
    ).

list_of_symbols_to_string([],"[]",_).
list_of_symbols_to_string([SymbolsH|SymbolsT],String,InOut) :-
    symbol_separator(InOut,Sep),
    list_of_symbols_to_string([SymbolsH|SymbolsT],Sep,String,[]).

list_of_symbols_to_string([],_,Str,Str).
list_of_symbols_to_string([H|T],Sep,Str0,Str) :-
    symbol_to_string(H,Str0,Str1),
    list_of_symbols_to_string0(T,Sep,Str1,Str).

list_of_symbols_to_string0([],_,Str,Str).
list_of_symbols_to_string0([H|T],Sep,Str0,Str) :-
    separator_to_string(Sep,Str0,Str1),
    symbol_to_string(H,Str1,Str2),
    list_of_symbols_to_string0(T,Sep,Str2,Str).

separator_to_string(Sep,Str0,Str) :-
    (  Sep==0
    -> Str0=Str
    ;  Str0=[Sep|Str]
    ).

symbol_to_string(Symbol,Str0,Str) :-
    (	atom(Symbol)
    ->	atom_codes(Symbol,String)
    ;   number(Symbol)
    ->	number_codes(Symbol,String)
    ;   var(Symbol)
    ->	frozen(Symbol,Cons),
	charsio:format_to_chars("|~w :- ~w|",[Symbol,Cons],String)
    ;   charsio:format_to_chars("~w",[Symbol],String)
    ),
    append(String,Str,Str0).

get_line(String,End) :-
    get_code(Char),
    (	(   Char < 0
	;   Char =:= 10
	)
    ->  End=Char, String=[]
    ;   String=[Char|Rest],
	get_line(Rest,End)
    ).

symbol_separator(in,Sep) :-
    (	fsa_globals:fsa_global_get(symbol_separator_in,Sep), Sep \== undefined
    ->  true
    ;   fsa_globals:fsa_global_get(symbol_separator,Sep)
    ->  true
    ;   fsa_globals:fsa_global_decl(symbol_separator,_,Sep,_,_)
    ).

symbol_separator(out,Sep) :-
    (	fsa_globals:fsa_global_get(symbol_separator_out,Sep), Sep \== undefined
    ->  true
    ;   fsa_globals:fsa_global_get(symbol_separator,Sep)
    ->  true
    ;   fsa_globals:fsa_global_decl(symbol_separator,_,Sep,_,_)
    ).



inc(Flag):-
    (	user:bb_get(fsa_interpreter:Flag,N0)
    ->  N is N0+1
    ;   N is 1
    ),
    user:bb_put(fsa_interpreter:Flag,N).

%% for the first flag(nr_sol_max) solutions to call(Call1), do call(Call2)
%% always succeeds.
atmost(Call1,Call2,Max):-
    (	(  user:bb_get(fsa_interpreter_unique,Ref0)  % get unique id
	-> true
	;  Ref0=0
	),
	Ref is Ref0+1,
	user:bb_put(fsa_interpreter_unique,Ref),
	user:bb_put(fsa_nr_sol:Ref,0),
	user:call_cleanup(fsa_interpreter:atmost_call(Call1,Ref,Max,Call2),
		     (user:bb_delete(fsa_nr_sol:Ref,Val),
		      format(user_error,"(~w outputs)~n",[Val])
		     )
		    )
    ->	true			% as soon as atmost/3 succeeds, don-t try
				% to re-satisfy Call1
    ;   true
    ).  

atmost_call(Call1,Ref,Max,Call2):-
    call(Call1),
    user:bb_get(fsa_nr_sol:Ref,N0),
    N is N0+1,
    atmost(N,Max,Call2,Ref).		     

atmost(N,Max,Call2,Ref) :-
    call(Call2),
    (	N < Max
    ->  user:bb_put(fsa_nr_sol:Ref,N),
        fail
    ;   user:bb_put(fsa_nr_sol:Ref,N),
        fsa_util:debug_message(1,
	       "that's enough of that (flag(nr_sol_max,~w))~n",
	       [Max])
    ).

user:help_info(xpred,fsa_regex_approx_accepts,
"fsa_regex_approx_accepts(+String,+Regex,-Recipe)",
"String is a Prolog string, and Regex is an atom that will be\
 \n parsed as a regular expression. The system will match String\
 \n approximately to that regular expression, returning each of \
 \n the matches which require a minimal number of substitutions,\
 \n insertions, deletions, and transpositions. A match is given by \
 \n a recipe which is a list of Prolog terms as follows:\
 \n \
 \n         P:d        deletion at position P\
 \n         P:i(Pred)  insertion of symbol for which Pred is true, at P\
 \n         P:s(Pred)  substitution of symbol for which Pred is true, at P\
 \n         P:t        transposition at position P\
 \n \
 \n where P refers to the position in the sequence of symbols extracted\
 \n from String where the corresponding edit operation takes place.").
fsa_regex_approx_accepts(String0,Regex,String) :-
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_approx_accepts(String0,Fa,String).

user:help_info(xpred,fsa_approx_accepts,
"fsa_approx_accepts(+String,+Fa,-Recipe)",
"String is a Prolog string, and Fa is a finite automaton.\
 \n The system will match String approximately to this Fa, returning each of \
 \n the matches which require a minimal number of substitutions,\
 \n insertions, deletions, and transpositions. A match is given by \
 \n a recipe which is a list of Prolog terms as follows:\
 \n \
 \n         P:d        deletion at position P\
 \n         P:i(Pred)  insertion of symbol for which Pred is true, at P\
 \n         P:s(Pred)  substitution of symbol for which Pred is true, at P\
 \n         P:t        transposition at position P\
 \n \
 \n where P refers to the position in the sequence of symbols extracted\
 \n from String where the corresponding edit operation takes place.").
%% String,Fa,Recipe
fsa_approx_accepts(String0,Fa0,String):-
    fsa_approx_accepts(String0,Fa0,String,_).
    
fsa_approx_accepts(String0,Fa0,String,String1):-
    string_to_list_of_symbols(String0,String1),
    fsa_regex:fsa_regex_compile(determinize(fa(Fa0)), Fa),
    fsa_data:fsa_symbols_decl(Fa,r(Module)),
    fsa_data:index_transitions(Fa,Trans),
    fsa_data:fsa_start_state(Fa,State),
    fsa_hash:fsa_hash_new(Done),
    string_to_array(String1,Array,Len),
    approx_trans([p(State,0,[])],Items,Module,Trans,Array,Done),
    sols(Items,Sols0,Fa,Len,Done),
    approx_closure(Sols0,Items,Sols,Module,Trans,Fa,Array,Len,Done),
    member(String2,Sols),
    reverse(String2,String).

user:help_info(xpred,fsa_regex_approx_transduces,
"fsa_regex_approx_transduces(+String0,+Regex,-String)",
"String0 and String is a Prolog string, Regex is an atom that will be\
 \n parsed as a regular expression denoting a string to string transducer. \
 \n The system will match String0 approximately to the domain of the regular \
 \n expression, and return each of the transductions of these approximate \
 \n matches.").
fsa_regex_approx_transduces(String0,Regex,String) :-
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_approx_transduces(String0,Fa,String).

user:help_info(xpred,fsa_approx_transduces,
"fsa_approx_transduces(+String0,+Fa,-String)",
"String0 and String is a Prolog string, Fa a string to string transducer.\
 \n The system will match String0 approximately to the domain of Fa, and return\
 \n each of the transductions of these approximate matches.").
fsa_approx_transduces(String0,Fa0,String):-
    fsa_regex:fsa_regex_compile(domain(fa(Fa0)), Fa1),
    fsa_approx_accepts(String0,Fa1,Recipe,String1),
    recipe_to_string(Recipe,String1,String2,Fa1),
    fsa_transduces(String2,String3,Fa0),
    list_of_symbols_to_string(String3,String,out).

user:help_info(xpred,fsa_regex_approx_transduces_w,
"fsa_regex_approx_transduces_w(+String0,+Regex,-Weight)",
"String0 is a Prolog string, Regex is an atom that will be\
 \n parsed as a regular expression denoting a weighted recognizer.\
 \n The system will match String0 approximately to the domain of the regular \
 \n expression, and return each of the weights of these approximate \
 \n matches.").
fsa_regex_approx_transduces_w(String0,Regex,String) :-
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_approx_transduces_w(String0,Fa,String).

user:help_info(xpred,fsa_regex_approx_transduces_wt,
"fsa_regex_approx_transduces_wt(+String0,+Regex,-String,-Weight)",
"String0 and String is a Prolog string, Regex is an atom that will be\
 \n parsed as a regular expression denoting a weighted transducer. \
 \n The system will match String0 approximately to the domain of the regular \
 \n expression, and return each of the transductions of these approximate \
 \n matches (a String and a Weight).").
fsa_regex_approx_transduces_wt(String0,Regex,String,Weight) :-
    fsa_regex:fsa_regex_atom_compile(Regex,Fa),
    fsa_approx_transduces_wt(String0,Fa,String,Weight).

user:help_info(xpred,fsa_approx_transduces_w,
"fsa_approx_transduces_w(+String0,+Fa,-Weight)",
"String0 is a Prolog string, Fa a weighted recognizer.\
 \n The system will match String0 approximately to the domain of Fa, and return\
 \n each of the weights of these approximate matches.").
fsa_approx_transduces_w(String0,Fa0,Weight):-
    fsa_regex:fsa_regex_compile(domain(fa(Fa0)), Fa1),
    fsa_approx_accepts(String0,Fa1,Recipe,String1),
    recipe_to_string(Recipe,String1,String2,Fa1),
    fsa_transduces_w(String2,Weight,Fa0).

user:help_info(xpred,fsa_approx_transduces_wt,
"fsa_approx_transduces_wt(+String0,+Fa,-String,-Weight)",
"String0 and String is a Prolog string, Fa a weighted transducer.\
 \n The system will match String0 approximately to the domain of Fa, and return\
 \n each of the transductions of these approximate matches (a String and a\
 \n Weight).").
fsa_approx_transduces_wt(String0,Fa0,String,Weight):-
    fsa_regex:fsa_regex_compile(domain(fa(Fa0)), Fa1),
    fsa_approx_accepts(String0,Fa1,Recipe,String1),
    recipe_to_string(Recipe,String1,String2,Fa1),
    fsa_transduces_wt(String2,String3,Weight,Fa0),
    list_of_symbols_to_string(String3,String,out).

%% approx_closure(+Sols0,+Items,-Sols,+Mod,+Trans,+Fa,+Array,+Len,?Done)
%% Sols:  list of solutions at current penalty level (if it's
%%        non-empty, we're finished)
%% Items: list of triples p(State,Pos,Recipe) indicating the possibilities
%%        for the current penalty level, i.e. each triple indicates that
%%        there is a path from the start state to State, consuming the
%%        input up to Pos, with Recipe.
%% Sols:  list of solutions
%% Mod:   predicate module
%% Trans: indexed transitions
%% Array: input string as an array
%% Len:   length of input string
%% Done:  hash table which maintains all Pos/State pairs which have
%%        been encountered at previous penalty levels (so these must
%%        not be generated at this or later levels)
approx_closure([H|T],_,[H|T],_,_,_,_,_,_).
approx_closure([],Items0,Sols,Mod,Trans,Fa,Array,Len,Done):-
    fsa_util:debug_message(2,"approx: increasing penalty level~n",[]),
    % extend each item in all possible ways, by applying penalty
    % transformation (ins,del,subs,transp)
    add_one(Items0,Items1,Mod,Trans,Array),
    % compute transitive closure of all items (i.e. taking transitions)
    approx_trans(Items1,Items,Mod,Trans,Array,Done),
    % extract list of items which are solutions (State final, Pos final)
    sols(Items,Sols0,Fa,Len,Done),
    % recurse
    approx_closure(Sols0,Items,Sols,Mod,Trans,Fa,Array,Len,Done).

add_one([],[],_,_,_).
add_one([p(State0,Str0,RStr0)|T],Items,Mod,Trans,Ar):-
    findall(Item,pen_step(State0,Str0,RStr0,Trans,Item,Mod,Ar),Items,Items0),
    add_one(T,Items0,Mod,Trans,Ar).

%% approx_trans(Items0,Items,Sols,Mod,Trans)
%% generate all items that are reachable from any item in Items0,
%%   final items go in Sols. NB: automaton is deterministic!
approx_trans([],[],_,_,_,_).
approx_trans([p(State0,P0,Rec)|T],I,Mod,Trans,Array,Done):-
    transition_seq(State0,P0,Rec,Mod,Trans,I,I0,Array,Done),
    approx_trans(T,I0,Mod,Trans,Array,Done).

transition_seq(State0,P0,Rec0,M,T,Items0,Items,Array,Done):-
    fsa_hash:fsa_hash_access(P0/State0,Var,Done),
    (	nonvar(Var)  % this Pos/State pair has already been
    ->  Items0=Items % found with better distance
    ;	Items0=[p(State0,P0,Rec0)|Items1],
	findall(Item,transition(State0,P0,Rec0,T,Item,M,Array),Next),
	transition_seq_next(Next,Items1,Items,M,T,Array,Done)
    ).

transition_seq_next([],I,I,_,_,_,_).
transition_seq_next([p(State0,P0,Rec0)],Items0,Items,M,T,Array,Done):-
    transition_seq(State0,P0,Rec0,M,T,Items0,Items,Array,Done).

sols([],[],_,_,_).
sols([p(State,Str,Rec)|Tail],Sols,Fa,Len,Done):-
    sols0(Str,State,Rec,Fa,Sols0,Sols,Len,Done),
    sols(Tail,Sols0,Fa,Len,Done).

sols0(Pos,State,Rec,Fa,Sols0,Sols,Len,Done):-
    (	Pos=:=Len,
	fsa_data:fsa_final_state(Fa,State)
    ->	Sols=[Rec|Sols0]
    ;   Sols=Sols0
    ),
    fsa_hash:fsa_hash_access(Pos/State,[],Done).

%% each of the operations
transition(State0,P0,Rec,Trans,p(State,P,Rec),Module,Array):-
    P is P0+1,
    fsa_array:fsa_array_get(P0,Sym,Array),
    fsa_data:indexed_transition(State0,Trans,SymPred,State),
    fsa_pred_module:evaluate_predicate(Module,SymPred,Sym).

pen_step(State0,P0,Rec,_Trans,p(State0,P,[P0:d|Rec]),_M,_A) :-
    P is P0+1.

pen_step(State0,P0,Rec,Trans,p(State,P0,[P0:i(SymPred)|Rec]),_M,_A):-
    fsa_data:indexed_transition(State0,Trans,SymPred,State).

pen_step(State0,P0,Rec,Trans,p(State,P,[P0:s(SymPred)|Rec]),_M,_A):-
    fsa_data:indexed_transition(State0,Trans,SymPred,State),
    P is P0 + 1.

pen_step(State0,P0,Rec,Trans,p(State,P,[P0:t|Rec]),Module,Array) :-
    fsa_array:fsa_array_get(P0,Sym0,Array), P1 is P0 + 1,
    fsa_array:fsa_array_get(P1,Sym1,Array), P is P1 + 1,
    fsa_data:indexed_transition(State0,Trans,SymP1,State1),
    fsa_pred_module:evaluate_predicate(Module,SymP1,Sym1),
    fsa_data:indexed_transition(State1,Trans,SymP0,State),
    fsa_pred_module:evaluate_predicate(Module,SymP0,Sym0).

fsa_approx_file(File,W):-
    fsa_io:fsa_read_file(File,Fa),
    fsa_approx(Fa,W).

fsa_approx(Fa,W) :-
    fsa_data:fsa_type(Fa,Type),
    fsa_approx(Type,Fa,W).

fsa_approx(recognizer,Fa,W) :-
    (	fsa_approx_accepts(W,Fa,Rec),
	format_recipe(Rec),
	fail
    ;   true
    ).

fsa_approx(transducer(_),Fa,W) :-
    (	fsa_approx_transduces(W,Fa,Str),
	format("~s~n",[Str]),
	fail
    ;   true
    ).

fsa_approx(w_recognizer,Fa,W) :-
    (	fsa_approx_transduces_w(W,Fa,Weight),
	format("~w~n",[Weight]),
	fail
    ;   true
    ).

fsa_approx(w_transducer(_),Fa,W) :-
    (	fsa_approx_transduces_wt(W,Fa,Str,Weight),
	format("~w:~s~n",[Weight,Str]),
	fail
    ;   true
    ).

string_to_array(L,Array,Size) :-
    length(L,Size),
    fsa_array:fsa_array_new(Array,Size),
    string_to_array_(L,0,Array).

string_to_array_([],_,_).
string_to_array_([H|T],N0,Ar):-
    fsa_array:fsa_array_access(N0,H,Ar),
    N is N0+1,
    string_to_array_(T,N,Ar).


recipe_to_string(Recipe,String0,String,Fa) :-
    fsa_data:fsa_symbols_decl(Fa,r(Module)),
    recipe_to_string(Recipe,0,String0,String,Module).

recipe_to_string([],_,Str,Str,_).
recipe_to_string([P0:H|T],P,StrA,StrB,Mod):-
    (	P0 =:= P
    ->	tr_recipe(H,StrA,StrA0,StrB,StrB0,Mod),
	Rec=T
    ;   StrA=[Hd|StrA0],
	StrB=[Hd|StrB0],
	Rec=[P0:H|T]
    ),
    P1 is P+1,
    recipe_to_string(Rec,P1,StrA0,StrB0,Mod).

tr_recipe(t,[X,Y|A],A,[Y,X|B],B,_).
tr_recipe(s(Pred),[_Hd|A],A,[Hd1|B],B,Mod):-
    fsa_pred_module:evaluate_predicate(Mod,Pred,Hd1).
tr_recipe(i(Pred),A,A,[Hd1|B],B,Mod) :-
    fsa_pred_module:evaluate_predicate(Mod,Pred,Hd1).
tr_recipe(d,[_|A],A,B,B,_).

%% write: swi puts spaces in between list elements :-(
format_recipe([]) :-
    format("Accepted!",[]).
format_recipe([H|T]):-
    format("Approximate Match Recipe: ",[]),
    format_recipe0(T,H).

format_recipe0([],H) :-
    format("~w~n",[H]).
format_recipe0([H|T],N) :-
    format("~w ",[N]),
    format_recipe0(T,H).

domain_module(Fa,Pm) :-
    fsa_data:fsa_symbols_decl(Fa,Term),
    arg(1,Term,Pm).

range_module(Fa,Pm) :-
    fsa_data:fsa_symbols_decl(Fa,Term),
    arg(2,Term,Pm).


fsa_sample_conversion(Fa0,Fa,Type) :-
    weighted_type(Fa0,Fa1,Type),
    fsa_data:fsa_components(Syms,Len0,Starts0,Finals0,Trans0,Jumps0,Fa1),
    fsa_data:fsa_components(Syms,Len, Starts, Finals, Trans, [],Fa),
    sample_jumps_to_transitions(Jumps0,Trans1,Trans0,Type),
    sample_starts_to_transitions(Starts0,Starts,Len0,Len1,Trans2,Trans1,Type),
    sample_finals_to_transitions(Finals0,Finals,Len1,Len,Trans3,Trans2,Type),
    weights_to_probs(Trans3,Trans,Type).

% fsa_regex:fsa_regex_compile(weighted_recognizer(fa(Fa0)),Fa1),

weighted_type(Fa0,Fa,Type) :-
    fsa_data:fsa_type(Fa0,Type0),
    weighted_type(Type0,Fa0,Fa,Type).

weighted_type(w_recognizer,Fa,Fa,recognizer).
weighted_type(recognizer,Fa0,Fa,recognizer) :-
    fsa_regex:fsa_regex_compile(zero_weights(fa(Fa0)),Fa).
weighted_type(w_transducer(_),Fa0,Fa,transducer) :-
    fsa_regex:fsa_regex_compile(weighted_transducer(fa(Fa0)),Fa).
weighted_type(transducer(_),Fa0,Fa,transducer) :-
    fsa_regex:fsa_regex_compile(weighted_transducer(fa(Fa0)),Fa).


sample_jumps_to_transitions([],T,T,_).
sample_jumps_to_transitions([jump(P,Q)|Tail],[Jump|T0],T,Type):-
    sample_jump(Type,P,Q,Jump),
    sample_jumps_to_transitions(Tail,T0,T,Type).

sample_jump(recognizer,P,Q,trans(P,[]/Zero,Q)) :-
    fsa_semiring:zero(fsa_weights,Zero).
sample_jump(transducer,P,Q,trans(P,[]/Zero,Q)) :-
    fsa_semiring:zero(fsa_weighted_strings,Zero).

sample_starts_to_transitions([],[],L,L,T,T,_).
sample_starts_to_transitions([S0|Ss],[Len0],Len0,Len,T0,T,Type):-
    Len is Len0+1,
    sample_starts_to_transitions([S0|Ss],Len0,T0,T,Type).

sample_starts_to_transitions([],_,T,T,_).
sample_starts_to_transitions([Q|Qs],P,[Jump|T0],T,Type) :-
    sample_jump(Type,P,Q,Jump),
    sample_starts_to_transitions(Qs,P,T0,T,Type).

sample_finals_to_transitions([],[],L,L,T,T,_).
sample_finals_to_transitions([S0|Ss],[Len0],Len0,Len,T0,T,Type):-
    Len is Len0+1,
    sample_finals_to_transitions([S0|Ss],Len0,T0,T,Type).

sample_finals_to_transitions([],_,T,T,_).
sample_finals_to_transitions([P|Ps],Q,[Jump|T0],T,Type) :-
    sample_jump(Type,P,Q,Jump),
    sample_finals_to_transitions(Ps,Q,T0,T,Type).


