%%% -*- Module: fsa_io; -*-
:- module( fsa_io,
	   [ fsa_read_file/2,
	     fsa_read_file/3,
	     fsa_write_file/1,
	     fsa_write_file/2,
	     fsa_write_file/3,
	     fsa_copy_file/2
	   ]).

#swi# :- style_check([-atom]).

#sic# :- use_module(library(fastrw),  [ fast_read/1,
#sic# 				        fast_write/1        ]).
#yap# :- use_module(library(charsio), [ read_from_chars/2,format_to_chars/3 ]).
#!yap# :- use_module(charsio, [ read_from_chars/2,format_to_chars/3 ]).
#yap# :- use_module(library(lists),   [ append/3, select/3  ]).
#sic# :- use_module(library(lists),   [ select/3  ]).

:- use_module(fsa_util).

:- public fsa_converse_in:unknown_predicate_handler/2,
          fsa_io_in:unknown_predicate_handler/2.

fsa_converse_in:unknown_predicate_handler(_,fail).
fsa_io_in:unknown_predicate_handler(_,fail).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous fsa_write_file/3.
:- discontiguous help_info/4.

user:help_info(class,io,"Formats of Finite State Automata","\
 \n FSA is capable of reading and writing finite automata in a number                \
 \n of different formats. The defaul format for reading is determined by the         \
 \n global variable *read*. The default format for writing is determined             \
 \n by the global variable *write*.                                                  \
 \n                                                                                  \
 \n The following formats are available both for reading and writing:                \
 \n                                                                                  \
 \n * fast. Binary format of the *normal* format. Uses library(fastrw).              \
 \n   Much faster reading and writing of automata. Drawback: binary files.           \
 \n * normal. Internal representation  (single Prolog term).                         \
 \n * old. Prolog program defining clauses start/1, final/2, trans/3,                \
 \n   jump/2. A variant of this was used by FSA2 and FSA3, but it is                 \
 \n   still useful, for instance, if you want to input automata directly,            \
 \n   rather than by means of regular expressions.                                   \
 \n * fsm. Format of automata as used in AT&T's fsm library.                         \
 \n * compact. text format, fairly compact. Slow (especially for output).            \
 \n                                                                                  \
 \n For writing, the following additional formats are available:                     \
 \n                                                                                  \
 \n * ps. PostScript.                                                                \
 \n * vcg. Input to the Xvcg graph visualization tool.                               \
 \n * davinci. Input to the DaVinci graph visualization tool.                        \
 \n * tk. Starts a interactive tcl/tk widget.                                        \
 \n * dot. Input for the GraphViz visualization tools dot and dotty.                 \
 \n * pstricks. LaTeX code to be included in a document; requires pstricks           \
 \n   macro's.                                                                       \
 \n * latex. LaTeX document; requires pstricks macro's.                              \
 \n * prolog. Prolog program; interface to fsa_compiler module.                      \
 \n * c. C program; interface to fsa_compiler_to_c module.                           \
 \n * java. JAVA program; interface to fsa_java module.                              \
 \n * cpp. C++ program; interface to fsa_cpp module.                                 \
 \n * fsm. Format of automata as used in AT&T's fsm library.                         \
 \n * grail. Format of automata as used in the Grail programme.                      \
 \n                                                                                  \
 \n The *normal* format is the internal format used in FSA6. The module              \
 \n *fsa_data* provides a consistent interface to this format.                       \
 \n                                                                                  \
 \n Here is a table indicating the relative speed of the standard input              \
 \n and output formats:                                                              \
 \n                                                                                  \
 \n     format      compact    fast     normal                                       \
 \n     writing          20       1          4                                       \
 \n     reading           5       1          4                                       \
 \n                                                                                  \
 \n Here is a table indicating the relative size of the standard input               \
 \n and output formats (measured in bytes):                                          \
 \n                                                                                  \
 \n                 compact    fast     normal                                       \
 \n                       1     1.8        1.6                                       ").

user:help_info(io,old,"The *old* format",
  " In the *old* format a finite automaton is given as a Prolog program.      \
 \n The automaton is defined by clauses for the predicates:                   \
 \n                                                                           \
 \n * start(State)                                                            \
 \n * final(State)                                                            \
 \n * trans(State0,Sym,State)                                                 \
 \n * jump(State0,State)                                                      \
 \n                                                                           \
 \n Note that in this format states can be are arbitrary ground Prolog        \
 \n terms (these will be converted to integers). In the case of               \
 \n transducers, Sym is a pair Left/Right. The empty list [] is used to       \
 \n indicate the empty string.  In the case of sequential transducers,        \
 \n Right must be a list of symbols.                                          ").

user:help_info(io,compact,"The *compact* format",
  " The *compact* format fairly closely follows the *normal* format. See the            \
 \n documentation on the *normal* format in the *fsa_data* module for more           \
 \n information. In this format a file is an ordinary text file. The                 \
 \n format is intended to be used for machines only, and is not very                 \
 \n helpful for human consumption.                                                   \
 \n                                                                                  \
 \n * The first line of the file is the string *fsa6*. This is used to               \
 \n differentiate the file from the pre-fsa6 compact formats (which can              \
 \n still be read-in).                                                               \
 \n                                                                                  \
 \n * The second line is the letter *r* for recognizers or *t* for                   \
 \n transducers.                                                                     \
 \n                                                                                  \
 \n * For recognizers, the third line is the name of the predicate module.           \
 \n                                                                                  \
 \n * For transducers, there are two such lines. The first line defines              \
 \n the domain predicate module, the second line the range predicate module.         \
 \n                                                                                  \
 \n * The next line is an integer indicating the number of states                    \
 \n                                                                                  \
 \n * The next line is an ordered sequence of integers, separated by tabs,           \
 \n indicating the start states                                                      \
 \n                                                                                  \
 \n * The next line is an ordered sequence of integers, separated by tabs,           \
 \n indicating the final states                                                      \
 \n                                                                                  \
 \n * The next lines each represent a transition, until an empty line is             \
 \n encountered. The transitions are ordered. Each transition is a triple            \
 \n State Symbol State. Seperator is the tab again. States are                       \
 \n integers. Symbols are readable as Prolog terms (recognizers) or pairs            \
 \n of In/Out, where In is a term and Out is either a single term or a list of       \
 \n terms.  If the source state is identical to the source state of the              \
 \n previous line, it can be left out. If the symbol is identical as well,           \
 \n then it can be left out as well.                                                 \
 \n                                                                                  \
 \n * The next lines are jumps. Jumps are ordered. Each line consists of two         \
 \n states separated by a tab. If the source state is identical to the               \
 \n source state of the previous line, it can be left out. If the symbol             \
 \n is identical as well, then it can be left out as well.                           \
 \n                                                                                  \
 \n Example:                                                                         \
 \n                                                                                  \
 \n     fsa write=compact -r '[class(a..f),{g,h}]'                                   \
 \n                                                                                  \
 \n     fsa6                                                                         \
 \n     r                                                                            \
 \n     fsa_preds                                                                    \
 \n     3                                                                            \
 \n     0                                                                            \
 \n     1                                                                            \
 \n     0       in([a,b,c,d,e,f])       2                                            \
 \n     2       g       1                                                            \
 \n             h       1                                                            ").

user:help_info(io,fast,"The *fast* format",
  " The *fast* format uses the same Prolog term representation as the           \
 \n *normal* format, except that library(fastrw) is used to read and            \
 \n write the Prolog term. This implies that reading and writing of             \
 \n automata in this format is very fast; the disadvantage is that *fast*       \
 \n is a binary format and therefore cannot be (easily) treated by other        \
 \n programs. Only supported by SICStus Prolog.                                 ").


user:help_info(xpred,copy_fa,"copy_fa(+File0,+File1).","The automaton in        \
 \n File0 is copied to File1. Useful to convert between different               \
 \n formats, by setting the *read* and *write* global variables.                ").

fsa_copy_file(File0,File) :-
    fsa_read_file(File0,Fa),
    fsa_write_file(File,Fa).

user:help_info(xpred,fsa_read_file,"fsa_read_file([+Format,]+File,?Fa)",
"Fa is read from File. If Format is unspecified the value of the global  \
 \n variable *read* is taken as the input format.").

fsa_read_file(File,Fa) :-
    fsa_globals:fsa_global_get(read,ReadFlag),
    fsa_util:time(fsa_io:fsa_read_file(ReadFlag,File,Fa)).

fsa_read_file(normal,File,Fa):-
    see(File),
    prompt(OldPrompt,''),
    user:call_cleanup( read(Fa1),
               (   seen,  prompt(_,OldPrompt) )  ),
    fsa_data:check_fa(Fa1,Fa).

#sic# fsa_read_file(fast,File,Fa):-
#sic#     see(File),
#sic#     user:call_cleanup(fastrw:fast_read(Fa1),seen),
#sic#     fsa_data:check_fa(Fa1,Fa).

#!sic# fsa_read_file(fast,_,_) :-
#!sic#    user:throw(fsa_error("fast format not supported by host Prolog~n",[])).

fsa_read_file(old,File,Fa) :-
#!swi#    compile(fsa_io_in:File),
#swi#    consult(fsa_io_in:File),
    findall(S,fsa_io_in:start(S),Ss),
    findall(F,fsa_io_in:final(F),Fs),
    findall(trans(P,Sym,Q),fsa_io_in:trans(P,Sym,Q),Ts),
    findall(jump(I,J),fsa_io_in:jump(I,J),Js),
    fsa_data:fsa_construct_rename_states(Ss,Fs,Ts,Js,Fa).

fsa_read_file(compact,File,Fa) :- 
    get_file_words(File,Lines),
    fa_compact_read(Fa1,Lines),
    fsa_data:check_fa(Fa1,Fa).

fsa_read_file(grail,File,Fa) :-
    get_file_words(File,Lines),
    fa_grail_read(Fa1,Lines),
    fsa_data:check_fa(Fa1,Fa).

fsa_read_file(fsm,File,Fa) :-
    get_file_words(File,FsmLines),
    fsm2fsa(FsmLines,Fa).

user:help_info(xpred,fsa_write_file,"fsa_write_file([+Format,]+File,+Fa)",
"Fa is written to File. If Format is unspecified the value of the global \
 \n variable *write* is taken as the input format. ").

fsa_write_file(File,Fa) :-
    fsa_globals:fsa_global_get(write,WriteFlag),
    fsa_util:time(fsa_io:fsa_write_file(WriteFlag,File,Fa)).

#sic# fsa_write_file(fast,File,Fa) :-
#sic#     fsa_util:to_file(fastrw:fast_write(Fa),File).

#!sic# fsa_write_file(fast,_,_) :-
#!sic#    user:throw(fsa_error("fast format not supported by host Prolog~n",[])).

fsa_write_file(fa(A,B,C,D,E,F)):-
    header,
    format("fa(~n",[]),
    format_symbols(A,"sigma and symbols",","), 
    format_integer(B), format(", % number of states~n",[]),
    format_list(C,"start states",","),
    format_list(D,"final states",","),
    format_list(E,"transitions",","),
    format_list(F,"jumps",")."), 
    format("~n",[]).

fsa_write_file(normal,File,fa(A,B,C,D,E,F)) :-
    fsa_util:to_file(fsa_io:fsa_write_file(fa(A,B,C,D,E,F)),File).

fsa_write_file(old,File,Fa) :-
    fsa_util:to_file(fsa_io:fsa_write_file_old(Fa),File).

fsa_write_file_old(Fa):-
    header,
    (   fsa_data:fsa_start_state(Fa,State),
        format("start(~w).~n",[State]),
        fail
    ;   fsa_data:fsa_final_state(Fa,Final),
	format("final(~w).~n",[Final]),
	fail
    ;	fsa_data:fsa_transition(Fa,P,Sym,Q),
	format("trans(~w,~q,~w).~n",[P,Sym,Q]),
	fail
    ;	fsa_data:fsa_jump(Fa,P,Q),
	format("jump(~w,~w).~n",[P,Q]),
	fail
    ;	true
    ).

fsa_write_file(compact,File,Fa) :-
    fa_compact_write(Fa,LinesOfTokens),
    fsa_util:to_file(fsa_io:format_lines_of_tokens(LinesOfTokens), File).

fsa_write_file(grail,File,Fa) :-
    fa_grail_write(Fa,LinesOfTokens),
    fsa_util:to_file(fsa_io:format_lines_of_tokens(LinesOfTokens,32), File).

fsa_write_file(ps,File,Fa) :-
    fsa_visualization:fsa_postscript(Fa,File,[fsa_tk]).

fsa_write_file(postscript,File,Fa) :-
    fsa_visualization:fsa_postscript(Fa,File,[fsa_tk]).

fsa_write_file(vcg,File,Fa) :-
    fsa_util:to_file(fsa_visualization:fsa_vcg(Fa), File).

fsa_write_file(davinci,File,Fa) :-
    fsa_util:to_file(fsa_visualization:fsa_davinci(Fa), File).

fsa_write_file(dot,File,Fa) :-
    fsa_util:to_file(fsa_visualization:fsa_dot(Fa), File).

fsa_write_file(pstricks,File,Fa):-
    fsa_visualization:fsa_pstricks(picture,Fa,File,[fsa_tk]).

fsa_write_file(pstricks_tex,File,Fa):-
    fsa_visualization:fsa_pstricks(tex,Fa,File,[fsa_tk]).

fsa_write_file(latex,File,Fa):-
    fsa_visualization:fsa_pstricks(tex,Fa,File,[fsa_tk]).

fsa_write_file(prolog,File,Fa) :-
    fsa_util:to_file(fsa_compiler:fsa_compile_to_prolog(Fa),File).

fsa_write_file(c,File,Fa):-
    fsa_compiler_to_c:fsa_compile_to_c_fa(Fa,File).

fsa_write_file(java,File,Fa):-
    fsa_java:fsa_java_fa(Fa,File).

fsa_write_file(cpp,File,Fa):-
    fsa_cpp:fsa_cpp_fa(Fa,File).

fsa_write_file(count,File,Fa):-
    fsa_data:fsa_count(File,Fa).

fsa_write_file(fsm,File,Fa):-
    fsa2fsm(Fa,Transitions,Finals,Type),
    fsa_util:to_file(fsa_io:format_fsm(Type,Transitions,Finals), File).

format_lines_of_tokens(L) :-
    format_lines_of_tokens(L,9).   % default: tab is separator

format_lines_of_tokens([],_Sep).
format_lines_of_tokens([H|T],Sep) :-
    format_tokens(H,Sep),
    nl,
    format_lines_of_tokens(T,Sep).

format_tokens([],_).
format_tokens([H|T],Sep) :-
    format_token(H),
    format_tokens0(T,Sep).

format_tokens0([],_).
format_tokens0([H|T],Sep) :-
    put_code(Sep),
    format_token(H),
    format_tokens0(T,Sep).

format_token(X) :-
    format("~s",[X]).

format_integer(Arg) :-
    format("~d",[Arg]).

format_symbols(Symbols,Comment,End):-
    format("%begin ~s~n",[Comment]),
    format("~q~s~n",[Symbols,End]),
    format("%end ~s~n",[Comment]).

format_list([],Comment,End):-
    format("[]~s % ~s~n",[End,Comment]).
format_list([H|T],Comment,End) :-
    format("[ % begin ~s~n",[Comment]),
    format_list0(T,H,"]~s % end ~s~n",[End,Comment]).

format_list0([],H,LastF,LastArgs) :-
    format("~q~n",[H]),
    format(LastF,LastArgs).

format_list0([X|T],H,LastF,LastArgs) :-
    format("~q,~n",[H]),
    format_list0(T,X,LastF,LastArgs).

header :-
    format("%% Automatically generated by FSA Utilities.~n",[]),
    format("%% For more info, cf. http://www.let.rug.nl/~avannoord/Fsa/~n~n",['~']).    

% get_file(File,Lines) as in Quintus library getfile, except that we tokenize
% each line 
get_file_words(File, Lines) :-
    see(File),
    user:call_cleanup(fsa_io:get_file_words(Lines), seen ).

%% character number 9, \t, TAB, C-i, is used for separating items in
%% compact format. 
get_file_words(Lines) :-
    get_code(Char),
    (   Char  <  0
    ->	Lines = []
    ;   Char =:= 9
    ->	Lines=[[[],Word|Line]|Rest],
	get_code(Next),
	get_file_words(Next,Word,Line,Rest)
    ;   Char =:= 32
    ->	Lines=[[[],Word|Line]|Rest],
	get_code(Next),
	get_file_words(Next,Word,Line,Rest)
    ;   Char =:= 10
    ->	Lines = [[]|Rest],
	get_file_words(Rest)
    ;	Lines = [[Word|Line]|Rest], 
	get_file_words(Char, Word, Line, Rest)
    ).

get_file_words(Char,Word2,Line,Lines) :-
    (   Char =:= 9
    ->  Word2=[],
        Line=[Word|Line1],
	get_code(Next1),
	get_file_words(Next1,Word,Line1,Lines)
    ;   Char =:= 10
    ->  Word2=[],
	Line=[],
	get_file_words(Lines)
    ;   Word2=[Char|Word],
	get_code(Next),
	(   Next =:= 9
	->  Word=[], 
	    Line=[Word1|Line1],
	    get_code(Next1),
	    get_file_words(Next1,Word1,Line1,Lines)
	;   Next =:= 32
	->  Word=[], 
	    Line=[Word1|Line1],
	    get_code(Next1),
	    get_file_words(Next1,Word1,Line1,Lines)
	;   Next  >  10
	->  get_file_words(Next, Word, Line, Lines)
	;   Next =:= 10
	->  Line = [], Word=[],
	    get_file_words(Lines)
	;   Next  <   0
	->  Line = [],
	    Word=[],
	    Lines = []
	;   get_file_words(Next, Word, Line, Lines)
	)
    ).


ints_to_range([A0|T],A,B,Rest) :-
    list_to_pairs(T,A0,B0,Rest),
    number_codes(A0,A),
    number_codes(B0,B).

range_to_ints([A0|T],A,B,Rest):-
    number_codes(A0,A),
    number_codes(B0,B),
    pairs_to_list(A0,B0,T,Rest).

list_to_pairs([H|T],A,B,Rest) :-
    H =:= A+1,
    list_to_pairs0(T,H,B,Rest).

list_to_pairs0([],B,B,[]).
list_to_pairs0([H|T],A,B,Rest) :-
    (	H =:= A+1
    ->	list_to_pairs0(T,H,B,Rest)
    ;   A=B,
	Rest=[H|T]
    ).

pairs_to_list(A0,A,L0,L) :-
    (	A0==A
    ->  L0=L
    ;   A1 is A0+1,
	L0=[A1|L1],
	pairs_to_list(A1,A,L1,L)
    ).

fa_compact_read(fa(Symbols,States,Starts,Finals,Transs,Jumps),Ln0) :-
    fa_compact_fsa6(Ln0,Ln1),
    fa_compact_read_sigma(Ln1,Ln2,Symbols),
    fa_compact_read_states_number(Ln2,Ln3,States),
    fa_compact_read_list_states(Ln3,Ln4,Starts),
    fa_compact_read_list_states(Ln4,Ln5,Finals),
    fa_compact_read_trans(Ln5,Ln6,Transs),
    fa_compact_read_jumps(Ln6,Jumps).

fa_compact_write(fa(Symbols,States,Starts,Finals,Transs,Jumps),Ln0) :-
    fa_compact_fsa6(Ln0,Ln1),
    fa_compact_write_sigma(Symbols,Ln1,Ln2),
    fa_compact_write_states_number(States,Ln2,Ln3),
    fa_compact_write_list_states(Starts,Ln3,Ln4),
    fa_compact_write_list_states(Finals,Ln4,Ln5),
    fa_compact_write_trans(Transs,Ln5,Ln6),
    fa_compact_write_jumps(Jumps,Ln6).

fa_compact_fsa6([["fsa6"]|Ln],Ln).

fa_compact_write_sigma(r(Module),[["r"],[M]|Ln],Ln):-
    atom_codes(Module,M).

fa_compact_write_sigma(t(ModuleA,ModuleB),[["t"],[Ma],[Mb]|Ln],Ln):-
    atom_codes(ModuleA,Ma),
    atom_codes(ModuleB,Mb).

fa_compact_read_sigma([["r"],[M]|Ln],Ln,r(Module)):-
    !,
    atom_codes(Module,M).

fa_compact_read_sigma([["t"],[Ma],[Mb]|Ln],Ln,t(ModuleA,ModuleB)):-
    !,
    atom_codes(ModuleA,Ma),
    atom_codes(ModuleB,Mb).

fa_compact_write_states_number(States,[[States0]|Ln],Ln) :-
    number_codes(States,States0).

fa_compact_read_states_number([[States0]|Ln],Ln,States) :-
    number_codes(States,States0).

fa_compact_write_list_states(List,[Line|Lines],Lines):-
    fa_compact_write_integer_list(List,Line).

fa_compact_write_integer_list([],[]).
fa_compact_write_integer_list([Int0|Ints],[IntA,"-",IntB|L]):-
    ints_to_range([Int0|Ints],IntA,IntB,Rest),!,
    fa_compact_write_integer_list(Rest,L).
fa_compact_write_integer_list([Int|Ints],[Chars|L]):-
    number_codes(Int,Chars),
    fa_compact_write_integer_list(Ints,L).

fa_compact_read_list_states([Line|Lines],Lines,List):-
    fa_compact_read_integer_list(Line,List).

fa_compact_read_integer_list([],[]).
fa_compact_read_integer_list([IntA,"-",IntB|L],[Int0|Ints]):-
    !,
    range_to_ints([Int0|Ints],IntA,IntB,Rest),
    fa_compact_read_integer_list(L,Rest).
fa_compact_read_integer_list([Chars|L],[Int|Ints]):-
    number_codes(Int,Chars),
    fa_compact_read_integer_list(L,Ints).

fa_compact_write_trans([],[[]|T],T).
fa_compact_write_trans([trans(A,B,C)|Tail],[[H,I,J]|Ln0],Ln):-
    number_codes(A,H),
    fa_compact_write_symbol(B,I),
    number_codes(C,J),
    fa_compact_write_trans0(Tail,A,B,Ln0,Ln).

fa_compact_write_trans0([],_,_,[[]|T],T).
fa_compact_write_trans0([trans(A,B,C)|Tail],A0,B0,[Trans|Ln0],Ln):-
    (	A0==A,
	B0==B
    ->	Trans=[[],[],J],
	number_codes(C,J)
    ;   A0==A
    ->	Trans=[[],I,J],
	fa_compact_write_symbol(B,I),
	number_codes(C,J)
    ;   Trans=[H,I,J],
	number_codes(A,H),
	fa_compact_write_symbol(B,I),
	number_codes(C,J)
    ),
    fa_compact_write_trans0(Tail,A,B,Ln0,Ln).

fa_compact_read_trans([Next|Lines0],Lines,Trans):-
    fa_compact_read_trans4(Next,Lines0,Lines,Trans).

fa_compact_read_trans4([],L,L,[]).
fa_compact_read_trans4([H,I,J],Ln0,Ln,[trans(A,B,C)|Trans]):-
    number_codes(A,H),
    fa_compact_read_symbol(B,I),
    number_codes(C,J),
    fa_compact_read_trans5(Ln0,Ln,A,B,Trans).

fa_compact_read_trans5([],[],_,_,[]).
fa_compact_read_trans5([First|Ln0],Ln,A,B,Trans):-
    fa_compact_read_trans6(First,Ln0,Ln,A,B,Trans).

fa_compact_read_trans6([],L,L,_,_,[]).
fa_compact_read_trans6([H,I,J],L0,L,A0,B0,[trans(A,B,C)|Trans]):-
    (	H==[],
	I==[]
    ->	A0=A, B0=B,
	number_codes(C,J)
    ;   H==[]
    ->	A0=A,
	fa_compact_read_symbol(B,I),
	number_codes(C,J)
    ;   number_codes(A,H),
	fa_compact_read_symbol(B,I),
	number_codes(C,J)
    ),
    fa_compact_read_trans5(L0,L,A,B,Trans).

fa_compact_write_symbol(Sym,Chars) :-
    codesio:format_to_chars("~q",[Sym],Chars).

fa_compact_read_symbol(Sym,Chars) :-
    append(Chars," .",Chars1),
    codesio:read_from_chars(Chars1,Sym).

fa_compact_write_jumps([],[]).
fa_compact_write_jumps([jump(A,C)|Tail],[[H,J]|Ln0]):-
    number_codes(A,H),
    number_codes(C,J),
    fa_compact_write_jumps0(Tail,A,Ln0).

fa_compact_write_jumps0([],_,[]).
fa_compact_write_jumps0([jump(A,C)|Tail],A0,[Jump|Ln0]):-
    (	A0==A
    ->	Jump=[[],J],
	number_codes(C,J)
    ;   Jump=[H,J],
	number_codes(A,H),
	number_codes(C,J)
    ),
    fa_compact_write_jumps0(Tail,A,Ln0).

fa_compact_read_jumps([],[]).
fa_compact_read_jumps([[H,J]|Ln0],[jump(A,C)|Tail]):-
    number_codes(A,H),
    number_codes(C,J),
    fa_compact_read_jumps0(Ln0,A,Tail).

fa_compact_read_jumps0([],_,[]).
fa_compact_read_jumps0([[H,J]|Ln0],A0,[jump(A,C)|Tail]):-
    (	H==[]
    ->	A0=A,
	number_codes(C,J)
    ;   number_codes(A,H),
	number_codes(C,J)
    ),
    fa_compact_read_jumps0(Ln0,A,Tail).

%%%%%%%%%%%%%%%%%%%%%%%
%%%%% AT&T's FSM %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

%% unstable; doesn't treat sequence transducers, nor predicates.
%%
%% in the fsm format symbols are replaced by integers; there can
%% be a separate symbols file which `remembers' the integer <-> symbol
%% mapping. In the fsm output format, however, we simply forget this
%% mapping. Be warned: this clearly loses information.

:- public fsa2fsm_file/2, fsa2fsm_file/3, fsm2fsa_file/2.
fsa2fsm_file(FsaFile,Fsm) :-
    fsa_read_file(FsaFile,Fa),
    fsa2fsm(Fa,Transitions,Finals,Type),
    fsa_util:to_file(fsa_io:format_fsm(Type,Transitions,Finals),Fsm).

fsa2fsm_file(FsaFile,FsmSymbols,Fsm) :-
    fsa_read_file(FsaFile,Fa),
    fsa2fsm_with_symbols(Fa,Symbols,Transitions,Finals,Type),
    fsa_util:to_file(fsa_io:format_symbol_pairs(Symbols),FsmSymbols),
    fsa_util:to_file(fsa_io:format_fsm(Type,Transitions,Finals),Fsm).

%% if there is more than a single start state -> add jumps
%% if there is no start state -> add one (sink).
%% if start state has no outgoing trans -> add one (epsilon to sink)
format_symbol_pairs([]).
format_symbol_pairs([Sym-No|Tail]) :-
    format("~w ~w~n",[Sym,No]),
    format_symbol_pairs(Tail).

fsa2fsm_with_symbols(Fa0,Symbols,Transitions,Finals,Type):-
    fsa_regex:expand_predicates_fa(Fa0,Fa1,_),
    fsa_data:fsa_type(Fa1,Type),
    fsm_treat_start_state(Fa1,Fa2,Type), % ensure there is a single start state
				     % with at least one outgoing transition
    fsm_number_symbols(Fa2,Fa3,Symbols,_Thing,Type),
    fsa_data:fsa_transitions(Fa3,Transitions),
    fsa_data:fsa_final_states(Fa3,Finals).

fsa2fsm(Fa0,Transitions,Finals,Type):-
    fsa_regex:expand_predicates_fa(Fa0,Fa1,_),
    fsa_data:fsa_type(Fa1,Type),
    fsm_treat_start_state(Fa1,Fa2,Type), % ensure there is a single start state
				     % with at least one outgoing transition
    fsm_number_symbols(Fa2,Fa3,_Symbols,Thing,Type),
    fsa_data:fsa_final_states(Fa3,Finals),
    fsa_data:fsa_transitions(Fa3,Transitions0),
    fsm_alter_transitions(Transitions0,Transitions,Thing,Type).

fsm_treat_start_state(Fa0,Fa,Type):-
    fsm_add_start(Fa0,Fa1),
    fsm_add_jumps(Fa1,Fa2),
    fsm_add_trans(Fa2,Fa,Type).

fsm_add_start(Fa0,Fa) :-
    fsa_data:fsa_start_states(Fa0,Starts),
    fsm_add_start(Starts,Fa0,Fa).

%% if there is no start state, simply add a new state, and make it the
%% start state
fsm_add_start([],fa(A,B0,[],D,E,F),fa(A,B,[B0],D,E,F)) :-
    B is B0+1.
%% already ok
fsm_add_start([_|_],Fa,Fa).

%% if there are multiple start states, add jumps
fsm_add_jumps(Fa0,Fa) :-
    fsa_data:fsa_start_states(Fa,[H|T]),
    fsm_add_jumps(T,H,Fa0,Fa).

fsm_add_jumps([],_,Fa,Fa).
fsm_add_jumps([H|T],Start,fa(A,B,_,D,E,F0),fa(A,B,[H],D,E,F)):-
    fsm_jmps([H|T],Start,F,F0).

fsm_jmps([],_,E,E).
fsm_jmps([H|T],Start,[jump(Start,H)|E0],E):-
    fsm_jmps(T,Start,E0,E).

%% if there is no outgoing transition from the start state, then
%% add a jump-like transition to self, and ensure it is first in the list
%% if there is an outgoing transition from the start state, then
%% move it to the beginning of the list of transitions
fsm_add_trans(fa(A,B,[C],D,E0,F),fa(A,B,[C],D,E,F),Type):-
    (	select(trans(C,Sym,Targ),E0,E1)
    ->	E = [trans(C,Sym,Targ)|E1]
    ;   E = [trans(C,Eps,C)|E0],
	eps_symbol(Type,Eps)
    ).

eps_symbol(recognizer,[]).
eps_symbol(w_recognizer,[]/0).
eps_symbol(transducer(_),[]/[]).
eps_symbol(w_transducer(_),[]/([]/0)).

%% epsilon symbol, [], is always numbered 0
fsm_number_symbols(Fa0,Fa,[[]-0|Symbols],Thing,Type):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Tr0,Tr),
    fsa_data:fsa_copy_except(jumps,Fa1,Fa,Jumps0,[]),
    fsa_hash:fsa_hash_new(Thing),
    fsa_hash:fsa_hash_access([],0,Thing),
    fsm_number_transitions(Tr0,Tr,Jumps,Thing,Symbols,[],1,Type),
    fsm_number_jumps(Jumps0,Jumps,Type).

fsm_number_jumps([],[],_).
fsm_number_jumps([jump(A,B)|Js0],[Trans|Js],Type) :-
    fsm_number_jump(Type,A,B,Trans),
    fsm_number_jumps(Js0,Js,Type).

fsm_number_jump(Type,A,B,trans(A,Eps,B)) :-
    eps_symbol(Type,Eps).

%% transitions aren't changed - but we add tail var for jumps.
fsm_number_transitions([],Tr,Tr,_,Syms,Syms,_,_Type).
fsm_number_transitions([Trans|Tr0],[Trans|Tr1],Tr,Thing,Syms0,Syms,Nr0,Type) :-
    fsm_number_transition(Type,Trans,Thing,Syms0,Syms1,Nr0,Nr),
    fsm_number_transitions(Tr0,Tr1,Tr,Thing,Syms1,Syms,Nr,Type).

fsm_number_transition(recognizer,trans(_,Sym,_),Thing,Syms0,Syms,N0,N) :-
   fsm_number_symbol(Sym,Thing,Syms0,Syms,N0,N).
fsm_number_transition(w_recognizer,trans(_,Sym/_,_),Thing,Syms0,Syms,N0,N) :-
   fsm_number_symbol(Sym,Thing,Syms0,Syms,N0,N).
fsm_number_transition(transducer(_),trans(_,L/R,_),Thing,Syms0,Syms,N0,N) :-
   fsm_number_symbol(L,Thing,Syms0,Syms1,N0,N1),
   fsm_number_symbol(R,Thing,Syms1,Syms,N1,N).
fsm_number_transition(w_transducer(_),trans(_,L/(R/_),_),Thing,Syms0,Syms,N0,N) :-
   fsm_number_symbol(L,Thing,Syms0,Syms1,N0,N1),
   fsm_number_symbol(R,Thing,Syms1,Syms,N1,N).

fsm_number_symbol(Sym,Thing,Syms0,Syms,M0,M):-
    fsa_hash:fsa_hash_access(Sym,Nr,Thing),
    (	var(Nr)
    ->	Nr=M0,
	M is M0+1,
	Syms0=[Sym-Nr|Syms]
    ;	Syms0=Syms,
	M0=M
    ).

format_fsm(Type,Transitions,Finals) :-
    format_fsm_transs(Transitions,Type),
    format_fsm_finals(Finals,Type).

format_fsm_transs([],_).
format_fsm_transs([Trans|Tr],Type):-
    format_fsm_trans(Type,Trans),
    format_fsm_transs(Tr,Type).

format_fsm_trans(recognizer,trans(A,B,C)) :-
    format("~w ~w ~w~n",[A,C,B]).
format_fsm_trans(w_recognizer,trans(A,B/W,C)) :-
    format("~w ~w ~w ~w~n",[A,C,B,W]).
format_fsm_trans(transducer(_),trans(A,B/D,C)) :-
    format("~w ~w ~w ~w~n",[A,C,B,D]).
format_fsm_trans(w_transducer(_),trans(A,B/(D/W),C)) :-
    format("~w ~w ~w ~w ~w~n",[A,C,B,D,W]).

format_fsm_finals([],_Type).
format_fsm_finals([H|T],Type) :-
    format_fsm_final(Type,H),
    format_fsm_finals(T,Type).

format_fsm_final(recognizer,H) :-
    format("~w~n",[H]).
format_fsm_final(w_recognizer,H) :-
    format("~w 0~n",[H]).
format_fsm_final(transducer(_),H) :-
    format("~w~n",[H]).
format_fsm_final(w_transducer(_),H) :-
    format("~w 0~n",[H]).

fsm_alter_transitions([],[],_,_Type).
fsm_alter_transitions([trans(A,Sym0,B)|Ts0],[trans(A,Sym,B)|Ts],Thing,Type):-
    fsm_alter_symbol(Type,Sym0,Sym,Thing),
    fsm_alter_transitions(Ts0,Ts,Thing,Type).

fsm_alter_symbol(recognizer,Sym0,Sym,Thing) :-
    fsa_hash:fsa_hash_access(Sym0,Sym,Thing).
fsm_alter_symbol(w_recognizer,Sym0/W,Sym/W,Thing) :-
    fsa_hash:fsa_hash_access(Sym0,Sym,Thing).
fsm_alter_symbol(transducer(_),SymL0/SymR0,SymL/SymR,Thing) :-
    fsa_hash:fsa_hash_access(SymL0,SymL,Thing),
    fsa_hash:fsa_hash_access(SymR0,SymR,Thing).
fsm_alter_symbol(w_transducer(_),SymL0/(SymR0/W),SymL/(SymR/W),Thing) :-
    fsa_hash:fsa_hash_access(SymL0,SymL,Thing),
    fsa_hash:fsa_hash_access(SymR0,SymR,Thing).

fsm2fsa_file(FsmFile,FsaFile) :-
    get_file_words(FsmFile,Lines),
    fsm2fsa(Lines,Fa),
    fsa_write_file(FsaFile,Fa).

fsm2fsa(Lines,Fa) :-
    parse_fsm(Start,Transitions,Jumps,Finals,Lines,[]),
    fsa_data:fsa_construct([Start],Finals,Transitions,Jumps,Fa).

%% this little grammar is reversible, and inefficient.
parse_fsm(Start,Tr,Js,Fs,[[Word1|Words]|Lines],Rest):-
    number_codes(Start,Word1),
    parse_fsm(Tr,Js,Fs,[[Word1|Words]|Lines],Rest).

parse_fsm(Trs,[J0|Js],Fs,X0,X) :-
    j(J0,X0,X1),!,
    parse_fsm(Trs,Js,Fs,X1,X).

parse_fsm([Tr0|Trs],Js,Fs,X0,X) :-
    tr(Tr0,X0,X1),!,
    parse_fsm(Trs,Js,Fs,X1,X).

parse_fsm(Trs,Js,[F0|Fs],X0,X) :-
    f(F0,X0,X1),!,
    parse_fsm(Trs,Js,Fs,X1,X).

parse_fsm([],[],[],X,X).

tr(trans(A,Sym,B),[[A0,B0,Sym0]|X],X):-
    number_codes(A,A0),
    number_codes(B,B0),
    number_codes(Sym,Sym0).

tr(trans(A,Sym,B),[[A0,B0,Sym0,[]]|X],X):-
    !,
    number_codes(A,A0),
    number_codes(B,B0),
    number_codes(Sym,Sym0).

tr(trans(A,SymA/SymB,B),[[A0,B0,SymA0,SymB0]|X],X) :-
    number_codes(A,A0),
    number_codes(B,B0),
    number_codes(SymA,SymA0),
    number_codes(SymB,SymB0).

j(jump(A,B),[[A0,B0,Sym0]|X],X) :-
    number_codes(A,A0),
    number_codes(B,B0),
    number_codes(0,Sym0).

f(A,[[A0]|X],X) :-
    number_codes(A,A0).

fa_grail_read(Fa,Ln0) :-
    fsa_grail_read(Ln0,Start,[],Final,[],Trans,[]),
    fsa_data:fsa_construct(Start,Final,Trans,[],Fa).

fsa_grail_read([],S,S,F,F,T,T).
fsa_grail_read([Line|Lines],S0,S,F0,F,T0,T) :-
    fsa_grail_read_line(Line,S0,S1,F0,F1,T0,T1),
    fsa_grail_read(Lines,S1,S,F1,F,T1,T).

fsa_grail_read_line([Pstr,Symstr,Qstr],S0,S,F0,F,T0,T):-
    (	Pstr == "(START)",
	Symstr == "|-"
    ->	F0=F, T0=T, S0=[Q|S], number_chars(Q,Qstr)
    ;	Qstr == "(FINAL)",
	Symstr == "-|"
    ->	S0=S, T0=T, F0=[P|F], number_chars(P,Pstr)
    ;	S0=S, F0=F, T0=[trans(P,Pred,Q)|T],
	number_chars(P,Pstr),
	number_chars(Q,Qstr),
	atom_chars(Sym,Symstr),
	fsa_globals:fsa_global_get(pred_module,M),
	fsa_pred_module:regex_notation_to_predicate(M,Sym,Pred)
    ).


%% only for efree recognizer
fa_grail_write(Fa,Ln0) :-
    fsa_data:fsa_type(Fa,Type),
    (	Type == recognizer
    ->	true
    ;   user:throw(fsa_error(
           "grail output: must be recognizer~n",[]))
    ),
    fsa_data:fsa_jumps(Fa,Js),
    (	Js == []
    ->	true
    ;	user:throw(fsa_error(
           "grail output: must not have jumps~n",[]))
    ),
    fsa_data:fsa_symbols_decl(Fa,r(Mod)),
    (	Mod == fsa_frozen
    ->	true
    ;   user:throw(fsa_error(
           "grail output: requires fsa_frozen pred module~n",[]))
    ),
    fsa_data:fsa_start_states(Fa,Starts),
    fsa_data:fsa_final_states(Fa,Finals),
    fsa_data:fsa_transitions(Fa,Trans),
    fsa_grail_write_starts(Starts,Ln0,Ln1),
    fsa_grail_write_finals(Finals,Ln1,Ln2),
    fsa_grail_write_transitions(Trans,Ln2,[]).

fsa_grail_write_starts([],L,L).
fsa_grail_write_starts([H|T],L0,L) :-
    fsa_grail_write_start(H,L0,L1),
    fsa_grail_write_starts(T,L1,L).

fsa_grail_write_finals([],L,L).
fsa_grail_write_finals([H|T],L0,L) :-
    fsa_grail_write_final(H,L0,L1),
    fsa_grail_write_finals(T,L1,L).

fsa_grail_write_transitions([],L,L).
fsa_grail_write_transitions([H|T],L0,L) :-
    fsa_grail_write_transition(H,L0,L1),
    fsa_grail_write_transitions(T,L1,L).

fsa_grail_write_start(State,[["(START)","|-",StateString]|Ln],Ln) :-
    number_codes(State,StateString).

fsa_grail_write_final(State,[[StateString,"-|","(FINAL)"]|Ln],Ln) :-
    number_codes(State,StateString).

fsa_grail_write_transition(trans(P,Sym,Q),[[Pstr,SymChars,Qstr]|Ln],Ln) :-
    number_codes(P,Pstr),
    number_codes(Q,Qstr),
    codesio:format_to_chars("~q",[Sym],SymChars),
    length(SymChars,Len),
    (	Len =:= 1
    ->	true
    ;	user:throw(fsa_error(
           "grail output: must be one letter symbols~n",[])
		   )
    ).

