%% -*- Module: fsa_minimizer; -*-
:- module(fsa_minimizer,[]).

#swi# :- style_check([-atom]).
#yap# :- use_module(library(lists),   [ member/2, select/3 ]).
#!swi# :- use_module(library(ordsets), [ ord_subtract/3, ord_union/3 ]).
#swi# :- use_module(ordsets, [ ord_subtract/3, ord_union/3 ]).

minimize_hopcroft(Fa0,Fa) :-
    fsa_data:fsa_states_number(Fa0,Size),
    fsa_data:fsa_symbols_decl(Fa0,r(Module)),
    fsa_data:index_inverted_transitions(Fa0,ITT),
    fsa_data:fsa_final_states(Fa0,Fstates),
    fsa_data:fsa_states_set(Fa0,All),
    fsa_m_array:fsa_m_array_new(StCl,Size),   %% state -> class array
    add_state_class(All,0,StCl),
    fsa_m_array:fsa_m_array_new(ClSts,Size),  %% class -> class contents
    fsa_data:fsa_states_number(Fa0,AllL),
    length(Fstates,InL),
    OutL is AllL-InL,
    (	OutL =:= AllL   %% if no states are final, then states all equivalent
    ->	true
    ;	to_set(AllL,All,AllA),
	fsa_m_array:fsa_m_array_put(0,z(AllA,AllL,All),ClSts),
	split_class(0,AllA,AllL,Fstates,InL,StCl,ClSts,1,Nr,Ag,[0]),
	partition(Ag,ITT,StCl,ClSts,Nr,Module,Size)
    ),
    single_states(StCl,Fa0,Starts,Finals,Trans,Size1),
    fsa_data:fsa_construct(r(Module),Size1,Starts,Finals,Trans,[],Fa).

partition([],_,_,_,_,_,_).
partition([Q1|Agenda0],ITT,StCl,ClSts,Nr0,Module,Size) :-
    fsa_m_array:fsa_m_array_get(Q1,z(Q1A,_,Q1States),ClSts),
    (	var(Q1States)
    ->	to_list(Q1A,Q1States)
    ;   true
    ),
    collect_transitions(Q1States,Syms0,ITT),
    fsa_determinizer:merge_transitions(Syms0,Syms1),
    determinize_preds(Module,Syms1,Syms3),
    sort(Syms3,Syms),
    symbols_loop(Syms,Agenda,Agenda0,StCl,ClSts,Nr0,Nr,Size),
    partition(Agenda,ITT,StCl,ClSts,Nr,Module,Size).

vals([],[]).
vals([_-V|Vs0],[V|Vs]) :-
    vals(Vs0,Vs).

%% symbols_loop(+Syms,?Agenda0,?Agenda,+StCl,+ClSts,+Nr0,?Nr)
%% Syms: list of Sources
%% Agenda0-Agenda: maintains agenda (classes that need to be
%%                 used for splitting later)
%% StCl: mapping State->Class
%% ClSts: mapping Class->Triple
%%         where Triple=z(Bbb,Nr,List)
%%         Bbb:representing the elements of Class in an assoc,
%%         Nr: number of elements in Class
%%         List: ordered list of elements in Class
%% Nr0-Nr: maintains available number for new classes
symbols_loop([],Ag,Ag,_,_,Nr,Nr,_).
symbols_loop([Sources|Syms],Ag0,Ag,StCl,ClSts,Nr0,Nr,Size):-
    single_step(Sources,Ag0,Ag1,StCl,ClSts,Nr0,Nr1,Size),
    symbols_loop(Syms,Ag1,Ag,StCl,ClSts,Nr1,Nr,Size).

%% single step(+Sources,?Ag0,?Ag,+StCl,+ClSts,+Nr0,?Nr)
%% Sources is the set of states which have a transition over
%% some symbol into the current target subset.
%% All the existing classes are split (if possible) using Sources
single_step(Sources,Ag0,Ag,StCl,ClSts,Nr0,Nr,Size):-
    fsa_m_array:fsa_m_array_new(Array,Size), %% class -> states in Sources
    %% for each state in Sources, we lookup in which class it belongs;
    %% we then have for each class a subset of Sources (epsilon subsets
    %% are ignored)
    divide_sources_to_classes(Sources,Classes0,Array,StCl,ClSts),
    sort(Classes0,Classes),
    %% all classes which have some elements from Sources are now split
    split_classes(Classes,Ag0,Ag,Array,StCl,ClSts,Nr0,Nr).

%% split_classes(+ClassList,?Ag0,?Ag,+Array,+StCl,+ClSts,+Nr0,?Nr)
%% Array: maps Class->subset of Sources
%% split each class wrt the corresponding subset in Array
split_classes([],Ag,Ag,_,_,_,Nr,Nr).
split_classes([Class|Classes],Ag0,Ag,Array,StCl,ClSts,Nr0,Nr):-
    %% lookup contents of Class:
    fsa_m_array:fsa_m_array_get(Class,z(MembersA,MembersL,_),ClSts),
    %% lookup subset in Sources of Class:
    fsa_m_array:fsa_m_array_get(Class,In-[],Array),
    length(In,InL),
    %% split it:
    split_class(Class,MembersA,MembersL,In,InL,StCl,ClSts,Nr0,Nr1,Ag0,Ag1),
    split_classes(Classes,Ag1,Ag,Array,StCl,ClSts,Nr1,Nr).

%% split_class(+Class,+AllA,+AllL,+In,+InL,+StCl,+ClSts,+Nr0,?Nr,?Ag0,?Ag)
%% AllA: assoc of states of Class
%% AllL: number of states of Class
%% In: states of one part of Class
%% InL: length of In
split_class(Class,AllA,AllL,In,InL,StCl,ClSts,Nr0,Nr,Ag0,Ag):-
    (	%% check that class really has been split (as opposed that all
	%% elements have ended up in In)
	InL =:= AllL
    ->	Nr0=Nr, Ag0=Ag
    ;	%% a real split is possible
	Nr is Nr0+1,             % new class is Nr0, so increase Nr
	del(In,AllA,OutA), % OutA contains All-In elements
	to_set(InL,In,InA), % InA is required later
	OutL is AllL-InL,        % to see which is smaller
	(   InL =< OutL          % which is larger?
	->  %% forget about Large
	    %% since we don't have those here ready, and we often
	    %% won't need those (but sometimes we will, e.g. by the
	    %% next symbol)
	    In=Small, InA=SmallA, OutA=LargeA, InL=SmallL, OutL=LargeL
	;   to_list(OutA,Small),
	    %% now we have to compute Small, since we need those for
	    %% updating StCl
	    In=Large, InA=LargeA, OutA=SmallA, InL=LargeL, OutL=SmallL
	),
	%% the smallest is _always_the new one; this simplifies the
	%% lines 11-14 of Fig 4.35 of Aho, Hopcroft, Ullman (1974)
	%% considerably. 
	Ag0=[Nr0|Ag],		% add new class to Agenda
	fsa_m_array:fsa_m_array_put(Nr0,z(SmallA,SmallL,Small),ClSts),
	%% update Class
	fsa_m_array:fsa_m_array_put(Class,z(LargeA,LargeL,Large),ClSts),
	%% update StCl
	add_state_class(Small,Nr0,StCl)
    ).

add_state_class([],_,_).
add_state_class([H|T],Val,Array):-
    fsa_m_array:fsa_m_array_put(H,Val,Array),
    add_state_class(T,Val,Array).

divide_sources_to_classes([],[],_,_,_).
divide_sources_to_classes([H|T],Classes,Array,StCl,ClSts):-
    fsa_m_array:fsa_m_array_get(H,Class,StCl),
    fsa_m_array:fsa_m_array_get(Class,z(_,Size,_),ClSts),
    add_class(Size,H,Class,Classes,Classes0,Array),
    divide_sources_to_classes(T,Classes0,Array,StCl,ClSts).

add_class(Size,H,Class,Classes0,Classes,Array) :-
    (	Size > 1
    ->	Classes0=[Class|Classes],
	fsa_m_array:fsa_m_array_put(Class,Hd-[H|T2],E-E,Hd-T2,Array)
    ;	Classes0=Classes
    ).

%add_to_partition([],_,_).
%add_to_partition([State|States],Class,Array):-
%    fsa_m_array:fsa_m_array_put(State,Class,Array),
%    add_to_partition(States,Class,Array).

single_states(Parts,Fa,Starts,Finals,Trans,Nr) :-
    fsa_data:fsa_start_states(Fa,Starts0),
    fsa_data:fsa_final_states(Fa,Finals0),
    fsa_data:fsa_transitions(Fa,Trans0),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Table,Size),
    p_states(Starts0,Starts,Parts,Table,0,Nr0),
    p_states(Finals0,Finals,Parts,Table,Nr0,Nr1),
    p_trans(Trans0,Trans,Parts,Table,Nr1,Nr).

p_states([],[],_,_,Nr,Nr).
p_states([H0|T0],[H|T],Table0,Table,Nr0,Nr) :-
    get_state_number(H0,H,Table0,Table,Nr0,Nr1),
    p_states(T0,T,Table0,Table,Nr1,Nr).

p_trans([],[],_,_,Nr,Nr).
p_trans([trans(A0,Sym,B0)|T0],[trans(A,Sym,B)|T],Table0,Table,Nr0,Nr) :-
    get_state_number(A0,A,Table0,Table,Nr0,Nr1),
    get_state_number(B0,B,Table0,Table,Nr1,Nr2),
    p_trans(T0,T,Table0,Table,Nr2,Nr).

get_state_number(A0,A,Table0,Table,Nr0,Nr):-
    fsa_m_array:fsa_m_array_get(A0,Class,Table0),
    fsa_array:fsa_array_access(Class,A,Table),
    (	var(A)
    ->	A=Nr0,
	Nr is Nr0+1
    ;   Nr0=Nr
    ).

collect_transitions([],[],_).
collect_transitions([P|Tail],L0,TT) :-
    fsa_array:fsa_array_access(P,L1,TT),
    (	nonvar(L1)
    ->	L0=[L1|L2]
    ;	L2 = L0
    ),
    collect_transitions(Tail,L2,TT).

to_list(Set,List) :-
    set_bbbtree:set_bbbtree__to_sorted_list(Set,List).

to_set(Len,List,Set):-
    set_bbbtree:set_bbbtree__sorted_list_to_set_len(List,Set,Len).

del(List,Set0,Set):-
    set_bbbtree:set_bbbtree__delete_list(List,Set0,Set).


determinize_preds(Module,Syms0,Syms) :-
    (	atomic(Module)
    ->	fsa_pred_module:determinize_preds(Module,Syms0,Syms2),
	vals(Syms2,Syms)
    ;   Module=pair(Decl),
	pairs_determinize_preds(Decl,Syms0,Syms)
    ).

pairs_determinize_preds(r(M),Syms0,Syms):-   % is not supposed to happen
    determinize_preds(M,Syms0,Syms).
pairs_determinize_preds(r(Ma,Mb),Syms0,Syms):-
    pairs_determinize_preds(Syms0,Ma,Mb,Syms).
pairs_determinize_preds(t(Ma,Mb),Syms0,Syms):-
    pairs_determinize_preds(Syms0,Ma,Mb,Syms).
pairs_determinize_preds(t(Ma,Mb,_Mc),Syms0,Syms):-
    pairs_determinize_preds_w(Syms0,Ma,Mb,Syms).

%% K0 is a key-value lists where the key is a
%% SymL/SymR pair, and the value is a list of states.
%% K is a list of lists of states.
%%
%% could be worth memo-izing if the associated states could be somehow
%% computed separately
pairs_determinize_preds(K0,Ma,Mb,K) :-
    extract_id_kl(K0,K1),
    all_exclusive_preds(K1,K2,Ma,Mb),
    vals(K2,K).

extract_id_kl([],[]).
extract_id_kl([Key0-List|Tail0],[p(A,B,Id)-List|Tail]) :-
    extract_id(Key0,A,B,Id),
    extract_id_kl(Tail0,Tail).

extract_id('$@'(A)/'$@'(A),A0,A1,Id) :-
    !,
    A0=A, A1=A, Id=id.
extract_id(A/B,A,B,var).

all_exclusive_preds(KvList0,KvList,Ma,Mb):-
    findall(Key-Val,exclusive_preds(KvList0,Key,Val,Ma,Mb),KvList1),
    sort(KvList1,KvList).

exclusive_preds(KvList0,Key,Val,Ma,Mb):-
    lists:select(Key0-Val0,KvList0,KvList1),
    all_exclusive_preds1(KvList1,[Key0-Val0],Kv,Ma,Mb),
    lists:member(Key-Val,Kv).

all_exclusive_preds1([],Kv,Kv,_,_).
all_exclusive_preds1([Key0-Target0|T],Kv0,Kv,Ma,Mb):-
    findall(Key-Target,
	    add_exclusive_pred(Key0,Target0,Kv0,Key,Target,Ma,Mb),
	    Kv1
	   ),
    sort(Kv1,Kv2),
    all_exclusive_preds1(T,Kv2,Kv,Ma,Mb).

add_exclusive_pred(Key0,Target0,KvList0,Key,Target,Ma,Mb) :-
    member(Key1-Target1,KvList0),
    take_or_leave(Key0,Key1,Key,Target1,Target0,Target,Ma,Mb).

%% take
take_or_leave(Pred0,Pred1,Pred2,Target,Targets0,Targets,Ma,Mb) :-
    pair_conjunction(Ma,Mb,Pred0,Pred1,Pred2),
    ord_union(Targets0,Target,Targets).
%% leave
take_or_leave(Pred0,Pred1,Pred2,T,_,T,Ma,Mb) :-
    pair_negation(Ma,Mb,Pred0,NPred0),
    pair_conjunction(Ma,Mb,NPred0,Pred1,Pred2).

pair_negation(Ma,Mb,p(P,Q,Id),p(P1,Q1,Id1)) :-
    pair_negation(Id,P,Q,Id1,P1,Q1,Ma,Mb).

pair_negation(id,P,P,Id,Qa,Qb,M,M) :-
    fsa_pred_module:true(M,Var),
    (	fsa_pred_module:negation(M,P,Qa), Qb=Var, Id=var
    ;   fsa_pred_module:negation(M,P,Qb), Qa=Var, Id=var
    ;   Qa = Var,
	Qb = Var, Id=non_id
    ).
pair_negation(non_id,P,Q,Id,Qa,Qb,M,M) :-
    fsa_pred_module:true(M,Var),
    (	fsa_pred_module:negation(M,P,Qa), Qb=Var, Id=var
    ;   Qa=Var,  fsa_pred_module:negation(M,Q,Qb), Id=var
    ;   Qa=Var, Qb=Var, Id=id
    ).

pair_negation(var,P,_Q,var,NotP,Var,Ma,Mb) :-
    fsa_pred_module:negation(Ma,P,NotP),
    fsa_pred_module:true(Mb,Var).

pair_negation(var,_P,Q,var,Var,NotQ,Ma,Mb) :-
    fsa_pred_module:negation(Mb,Q,NotQ),
    fsa_pred_module:true(Ma,Var).


pair_conjunction(Ma,Mb,p(A0,A1,Ida),p(B0,B1,Idb),p(AB2,AB3,Id)):-
    fsa_pred_module:conjunction(Ma,A0,B0,AB0),
    fsa_pred_module:conjunction(Mb,A1,B1,AB1),
    conjoin_id(Ida,Idb,Id),
    add_id(Id,AB0,AB1,AB2,AB3,Ma,Mb).

conjoin_id(var,A,B) :-
    conjoin_id_var(A,B).
conjoin_id(id,A,B) :-
    conjoin_id_id(A,B).
conjoin_id(non_id,A,B) :-
    conjoin_id_non_id(A,B).

conjoin_id_var(var,var).
conjoin_id_var(id,id).
conjoin_id_var(non_id,non_id).

conjoin_id_id(var,id).
conjoin_id_id(id,id).

conjoin_id_non_id(var,non_id).
conjoin_id_non_id(non_id,non_id).


add_id(var,AB0,AB1,AB0,AB1,_Ma,_Mb).
add_id(id,A0,A1,A,A,M,M) :-
    fsa_pred_module:conjunction(M,A0,A1,A).
add_id(non_id,A,B,A,B,M,M) :-
    \+ empty_non_id(A,B,M).
		% check that A/B contains a non-identical pair...

%% if they are the same and singleton
%% this really would require `singleton' declaration.
empty_non_id(A,A,M) :-
    fsa_pred_module:identity(M,A,A).

%% and now for weights too! we treat weights here temporarily
%% through the fsa_preds module.
pairs_determinize_preds_w(K0,Ma,Mb,K) :-
    extract_id_kl_w(K0,K1),
    all_exclusive_preds_w(K1,K2,Ma,Mb),
    vals(K2,K).

extract_id_kl_w([],[]).
extract_id_kl_w([Key0-List|Tail0],[p(A,B,W,Id)-List|Tail]) :-
    extract_id_w(Key0,A,B,W,Id),
    extract_id_kl_w(Tail0,Tail).

extract_id_w('$@'(A)/('$@'(A)/W0),A1,A2,W,Id) :-
    !,
    A1=A, A2=A, Id=id,
    fsa_pred_module:regex_notation_to_predicate(fsa_preds,W0,W).
extract_id_w(A/(B/W0),A,B,W,var):-
    fsa_pred_module:regex_notation_to_predicate(fsa_preds,W0,W).

all_exclusive_preds_w(KvList0,KvList,Ma,Mb):-
    findall(Key-Val,exclusive_preds_w(KvList0,Key,Val,Ma,Mb),KvList1),
    sort(KvList1,KvList).

exclusive_preds_w(KvList0,Key,Val,Ma,Mb):-
    lists:select(Key0-Val0,KvList0,KvList1),
    all_exclusive_preds1_w(KvList1,[Key0-Val0],Kv,Ma,Mb),
    lists:member(Key-Val,Kv).

all_exclusive_preds1_w([],Kv,Kv,_,_).
all_exclusive_preds1_w([Key0-Target0|T],Kv0,Kv,Ma,Mb):-
    findall(Key-Target,
	    add_exclusive_pred_w(Key0,Target0,Kv0,Key,Target,Ma,Mb),
	    Kv1
	   ),
    sort(Kv1,Kv2),
    all_exclusive_preds1_w(T,Kv2,Kv,Ma,Mb).

add_exclusive_pred_w(Key0,Target0,KvList0,Key,Target,Ma,Mb) :-
    member(Key1-Target1,KvList0),
    take_or_leave_w(Key0,Key1,Key,Target1,Target0,Target,Ma,Mb).

%% take
take_or_leave_w(Pred0,Pred1,Pred2,Target,Targets0,Targets,Ma,Mb) :-
    pair_conjunction_w(Ma,Mb,Pred0,Pred1,Pred2),
    ord_union(Targets0,Target,Targets).
%% leave
take_or_leave_w(Pred0,Pred1,Pred2,T,_,T,Ma,Mb) :-
    pair_negation_w(Ma,Mb,Pred0,NPred0),
    pair_conjunction_w(Ma,Mb,NPred0,Pred1,Pred2).

pair_negation_w(Ma,Mb,p(P,Q,W,Id),p(P1,Q1,W1,Id1)) :-
    pair_negation_w(Id,P,Q,W,Id1,P1,Q1,W1,Ma,Mb).

pair_negation_w(id,P,P,W,Id,Qa,Qb,W1,M,M) :-
    fsa_pred_module:true(M,Var),
    fsa_pred_module:true(fsa_preds,WVar),
    (	fsa_pred_module:negation(M,P,Qa),
	Qb=Var,
	Id=var,
	W1=WVar
    ;   fsa_pred_module:negation(M,P,Qb),
	Qa=Var,
	Id=var,
	W1=WVar
    ;   Qa = Var,
	Qb = Var,
	Id=non_id,
	W1=WVar
    ;   Qa = Var,
	Qb = Var,
	Id=var,
	fsa_pred_module:negation(fsa_preds,W,W1)
    ).
pair_negation_w(non_id,P,Q,W,Id,Qa,Qb,W1,M,M) :-
    fsa_pred_module:true(fsa_preds,WVar),
    fsa_pred_module:true(M,Var),
    (	fsa_pred_module:negation(M,P,Qa),
	Qb=Var,
	Id=var,
	W1=WVar
    ;   Qa=Var,
	fsa_pred_module:negation(M,Q,Qb),
	Id=var,
	W1=WVar
    ;   Qa=Var,
	Qb=Var,
	Id=id,
	W1=WVar
    ;	Qa=Var,
	Qb=Var,
	Id=var,
	fsa_pred_module:negation(fsa_preds,W,W1)
    ).
pair_negation_w(var,P,Q,W,var,Qa,Qb,W1,Ma,Mb) :-
    fsa_pred_module:true(fsa_preds,WVar),
    fsa_pred_module:true(Mb,Var),
    (	fsa_pred_module:negation(Ma,P,Qa),
	Qb=Var,
	W1=WVar
    ;	Qa=Var,
	fsa_pred_module:negation(Mb,Q,Qb),
	W1=WVar
    ;	Qa=Var,
	Qb=Var,
	fsa_pred_module:negation(fsa_preds,W,W1)
    ).

pair_conjunction_w(Ma,Mb,p(A0,A1,W0,Ida),p(B0,B1,W1,Idb),p(AB2,AB3,W,Id)):-
    fsa_pred_module:conjunction(fsa_preds,W0,W1,W),
    fsa_pred_module:conjunction(Ma,A0,B0,AB0),
    fsa_pred_module:conjunction(Mb,A1,B1,AB1),
    conjoin_id(Ida,Idb,Id),
    add_id(Id,AB0,AB1,AB2,AB3,Ma,Mb).

