%%% -*- Module: fsa_minimum_path; -*-
:- module(fsa_minimum_path,
	  [ fsa_minimum_path/1,
	    fsa_minimum_path/2,
	    fsa_minimum_path_file/1,
	    fsa_minimum_path_array/3
	  ]).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

user:help_info(xpred,fsa_minimum_path_file,"fsa_minimum_path_file(+InFile)",
"Reports on standard output the minimum weight path in the transducer\
 \n read from InFile. \
 \n \
 \n This is implemented by a generalization of Dijkstra's algorithm to\
 \n find the minimum weight path in a given transducer. The algorithm for\
 \n transducers are implemented in a fully general way, i.e., for various\
 \n types of transducers (cf. the fsa_semiring module).\
 \n \
 \n The implementation is more general than the predicates provided by\
 \n e.g. the SICStus libraries for graphs. And *much* more efficient, even\
 \n though the agenda is not maintained as a heap (the latter decision\
 \n was caused by the fact that it would be hard to implement such\
 \n a heap efficiently for the various types of transducers with their\
 \n corresponding `minimum' definitions).").

fsa_minimum_path_file(File) :-
    fsa_io:fsa_read_file(File,Fa),
    fsa_minimum_path(Fa).

user:help_info(xpred,fsa_minimum_path,"fsa_minimum_path(+Fa[,-Path])",
"Reports on standard output (if Path is not present) or instantiates\
 \n Path to the minimum weight path in the transducer Fa.   ").

fsa_minimum_path(Fa) :-
    fsa_minimum_path(Fa,Score/Path),
    fsa_interpreter:write_list_of_symbols(Path,in),
    write(' <==>'),
    nl,
    write(Score),
    nl.

fsa_minimum_path(Fa0,Path):-
    to_flag(Fa0,Fa,Flag),
    fsa_minimum_path_array(Fa,PsArray,Flag),
    fsa_data:fsa_start_states(Fa,StartStates),
    get_minimum_array(StartStates,PsArray,Path,Flag).

to_flag(Fa0,Fa,Flag) :-
    fsa_data:fsa_type(Fa0,Type),
    type_to_flag(Type,Flag,Fa0,Fa).

type_to_flag(w_recognizer,    fsa_weights,Fa,Fa).
type_to_flag(w_transducer(_), fsa_weighted_strings,Fa0,Fa) :-
    fsa_regex:fsa_regex_compile(weighted_sequence_transducer(fa(Fa0)),Fa).
type_to_flag(transducer(_),   fsa_strings,Fa0,Fa) :-
    fsa_regex:fsa_regex_compile(sequence_transducer(fa(Fa0)),Fa).
type_to_flag(recognizer,_,_,_) :-
    user:throw(fsa_error(
      "fsa_minimum_path: expect weighted automaton, not recognizer!~n",
			 [])).

user:help_info(xpred,fsa_minimum_path_array,
"fsa_minimum_path_array(+Fa,-Array,+Flag)",
"Array will be instantiated to an UpdatableFsaArray (cf. module fsa_arrays)\
 \n indicating for each state in the transducer Fa the minimum cost from\
 \n that state to a final state. ").

fsa_minimum_path_array(Fa,PsArray,Flag) :-
    fsa_data:fsa_states_number(Fa,Size),
    fsa_semiring:zero(Flag,Zero),
    fsa_data:fsa_final_states(Fa,Finals),
    fsa_m_array:fsa_m_array_new(PsArray,Size),
    add_fs(Finals,Ag,PsArray,Zero), % add final states to agenda and array
    index_itj(Fa,ITJ,Zero),	    % index inverted transitions and jumps
    process_P(Ag,ITJ,PsArray,Flag).

add_fs([],[],_Ass,_).
add_fs([A|T],[g(A,Zero,[])|Ag],Ass,Zero) :-
    fsa_m_array:fsa_m_array_put(A,Zero/[],Ass),
    add_fs(T,Ag,Ass,Zero).


add(State0,Syms,His,Ag0,Ag,Arr,Flag) :-
    fsa_m_array:fsa_m_array_get(State0,V,Arr),
    (	var(V)
    ->	%% new state, add to Array and Ag
	Ag0 = [g(State0,Syms,His)|Ag],
	fsa_m_array:fsa_m_array_put(State0,Syms/His,Arr)
    ;	V=PrevPrefix/_,
	%% state was already there:
	fsa_semiring:minimum(Flag,Syms,PrevPrefix,Prefix),
	(   Prefix == PrevPrefix
	->  %% score is the same, Ag and Array remain the same
	    Ag0=Ag
	;   %% better score, add to Array and Ag
	    Ag0= [g(State0,Prefix,His)|Ag],
	    fsa_m_array:fsa_m_array_put(State0,Prefix/His,Arr)
	)
    ).

process_P([],_,_Arr,_).
process_P([H|T],ITJ,Arr,Flag) :-
    new_Ps(H,ITJ,Flag,Ag,T,Arr),
    process_P(Ag,ITJ,Arr,Flag).

new_Ps(g(State,Syms1,His0),ITJ,Flag,Ag0,Ag,Arr):-
    fsa_array:fsa_array_access(State,Keylist,[],ITJ),
    new_Ps0(Keylist,State,Syms1,His0,Flag,Ag0,Ag,Arr).

new_Ps0([],_,_,_,_,Ag,Ag,_Ar).
new_Ps0([Sym-Targets|Tail],State,Syms1,His0,Flag,Ag0,Ag,Ar):-
    new_Ps00(Targets,Sym,State,Syms1,His0,Flag,Ag0,Ag1,Ar),
    new_Ps0(Tail,State,Syms1,His0,Flag,Ag1,Ag,Ar).

new_Ps00([],_,_,_,_,_,Ag,Ag,_Ar).
new_Ps00([State0|Tail],In/Syms0,State,Syms1,His0,Flag,
	 Ag0,Ag,Ar):-
    his(His0,In,His),
    fsa_semiring:plus(Flag,Syms0,Syms1,Syms),
    add(State0,Syms,His,Ag0,Ag1,Ar,Flag),
    new_Ps00(Tail,In/Syms0,State,Syms1,His0,Flag,Ag1,Ag,Ar).

index_itj(Fa,ITJ,Zero):-
    fsa_data:fsa_transitions(Fa,Trans0),
    inv_trans(Trans0,ITrans0,ITrans1),
    fsa_data:fsa_jumps(Fa,Jumps),
    inv_jumps(Jumps,ITrans1,Zero),
    sort(ITrans0,ITrans),
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(ITJ,Size),
    fsa_data:transitions_to_thing(ITrans,ITJ).

inv_trans([],T,T).
inv_trans([trans(P,Sym,Q)|T0],[trans(Q,Sym,P)|I0],I):-
    inv_trans(T0,I0,I).

inv_jumps([],[],_Zero).
inv_jumps([jump(P,Q)|J0],[trans(Q,[]/Zero,P)|J],Zero):-
    inv_jumps(J0,J,Zero).

his(His0,In,His):-
    (	In==[]
    ->	His0=His
    ;	His=[In|His0]
    ).


%% for fsa_minimum_path/3:
%% find the state with the lowest score for a given list of states
%% assuming that PsArray is an array representing the score for
%% some of those states. Ignore states without a score.
get_minimum_array([St0|St],PsArray,Min,Flag) :-
    fsa_m_array:fsa_m_array_get(St0,Min0,PsArray),
    (   var(Min0)
    ->	get_minimum_array(St,PsArray,Min,Flag)
    ;   get_minimum_array(St,PsArray,Min0,Min,Flag)
    ).

get_minimum_array([],_,M,M,_).
get_minimum_array([St0|St],PsArray,Min0/His0,Min/His,Flag):-
    fsa_m_array:fsa_m_array_get(St0,V,PsArray),
    (	var(V)
    ->	get_minimum_array(St,PsArray,Min0/His0,Min/His,Flag)
    ;	V=Min1/His1,
	fsa_semiring:minimum(Flag,Min0,Min1,Min2),
	(   Min2 == Min0
	->  His2 = His0
	;   His2 = His1
	),
	get_minimum_array(St,PsArray,Min2/His2,Min/His,Flag)
    ).


